/******************************************************************
Module : DrvInit.cpp    - the main file of NT driver
             Written 1998,1999 by Dmitri Leman
Purpose: NT kernel mode tracer.
******************************************************************/

#include "Headers.h"
#include "Driver.h"
#include "Trace.h"
#include "PEHelper.h"
#include "KrnlTrc.h"
#include "KrnlTrcD.h"
#include "Stubmgr.h"
#include "Intrcpt.h"

//Forward definitions:
NTSTATUS TracerDispatch
    ( IN PDEVICE_OBJECT p_pDeviceObject, IN PIRP p_pIrp );
VOID     TracerUnload( IN PDRIVER_OBJECT p_pDriverObject );


PDEVICE_OBJECT      g_pTracerDevice = NULL;//This device provides 
                    //IOCTL interface with our launcher application

/*-----------------------------------------------------------------

   FUNCTION: TracerIoctl

   PURPOSE:  
      
   PARAMETERS:                   .
                                  
   IN PDEVICE_OBJECT   p_pDeviceObject        our device object
   IN PIRP             p_pIrp                 request packet
   IN PVOID            p_pInputBuffer         input buffer
   IN ULONG            p_ulInputBufferLength  length of data in the
                                              input buffer
   IN PVOID            p_pOutputBuffer        output buffer
   IN ULONG            p_ulOutputBufferLength size of the 
                                              output buffer
   IN ULONG            p_ulIoControlCode      IOCTL code 
   OUT     ULONG    *  p_pulBytesPlacedToOutput return the number
                                              of bytes written to
                                              the output buffer
   RETURN VALUE:
      NTSTATUS 0 on success, error code otherwise
_________________________________________________________________*/

NTSTATUS TracerIoctl
( 
    IN PDEVICE_OBJECT   p_pDeviceObject, 
    IN PIRP             p_pIrp,
    IN PVOID            p_pInputBuffer,
    IN ULONG            p_ulInputBufferLength,
    IN PVOID            p_pOutputBuffer,
    IN ULONG            p_ulOutputBufferLength,
    IN ULONG            p_ulIoControlCode,
    OUT ULONG    *      p_pulBytesPlacedToOutput //cannot be NULL
)
{
    NTSTATUS l_Status;

    *p_pulBytesPlacedToOutput = 0;
    l_Status = STATUS_NOT_IMPLEMENTED;

    switch(p_ulIoControlCode)
    {
    case IOCTL_TRACER_CHECK_ACTIVE:
        *((ULONG*)p_pOutputBuffer) = IOCTL_TRACER_CHECK_ACTIVE;
        *p_pulBytesPlacedToOutput = 4;
        l_Status = STATUS_SUCCESS;
        break;
    case IOCTL_TRACER_DUMP_TRACE:
        l_Status = STATUS_UNSUCCESSFUL;
        if(g_Data.m_pParams)
        {
            if(bDumpTraceBufferToFile(
                g_Data.m_pParams->m_szLogFileName))
            {
                l_Status = STATUS_SUCCESS;
            }
        }
        *p_pulBytesPlacedToOutput = 0;
        break;
    default:
        MYTRACE(TF_Error, 
            "ERROR: Unknown IOCTL (%x)", p_ulIoControlCode);

        *p_pulBytesPlacedToOutput = 0;
        l_Status = STATUS_NOT_IMPLEMENTED;
    }//switch(p_ulIoControlCode)
    return l_Status;
}//NTSTATUS TracerIoctl


/*-----------------------------------------------------------------

   FUNCTION: TracerDispatch

   PURPOSE:  Standard dispatch routine of the driver. Handles
    IRP_MJ_CREATE, IRP_MJ_CLOSE and IRP_MJ_DEVICE_CONTROL
      
   PARAMETERS:                   .
                                  
   IN PDEVICE_OBJECT   p_pDeviceObject - our device object
   IN PIRP             p_pIrp          - request packet

   RETURN VALUE:
      NTSTATUS 0 on success, error code otherwise
_________________________________________________________________*/

NTSTATUS TracerDispatch
( 
    IN PDEVICE_OBJECT   p_pDeviceObject,
    IN PIRP             p_pIrp 
)
{
    PIO_STACK_LOCATION  l_pIrpStack;
    PVOID               l_pInputBuffer;
    PVOID               l_pOutputBuffer;
    ULONG               l_ulInputBufferLength;
    ULONG               l_ulOutputBufferLength;
    ULONG               l_ulIoControlCode;

    //by default mark all requests as successful
    p_pIrp->IoStatus.Status      = STATUS_SUCCESS;
    p_pIrp->IoStatus.Information = 0;

    l_pIrpStack = IoGetCurrentIrpStackLocation (p_pIrp);

    l_pInputBuffer     = p_pIrp->AssociatedIrp.SystemBuffer;
    l_ulInputBufferLength   = 
        l_pIrpStack->Parameters.DeviceIoControl.InputBufferLength;
    l_pOutputBuffer    = p_pIrp->AssociatedIrp.SystemBuffer;
    l_ulOutputBufferLength  = 
        l_pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
    l_ulIoControlCode   = 
        l_pIrpStack->Parameters.DeviceIoControl.IoControlCode;

    switch (l_pIrpStack->MajorFunction) {
    case IRP_MJ_CREATE:

        MYTRACE(TF_Driver, "Tracer: IRP_MJ_CREATE");

        break;

    case IRP_MJ_CLOSE:

        MYTRACE(TF_Driver, "Tracer: IRP_MJ_CLOSE");

        break;

    case IRP_MJ_DEVICE_CONTROL:
        MYTRACE(TF_Driver, "Tracer: IRP_MJ_DEVICE_CONTROL");

        p_pIrp->IoStatus.Status = TracerIoctl(
            p_pDeviceObject, p_pIrp,
            l_pInputBuffer, l_ulInputBufferLength, 
            l_pOutputBuffer, l_ulOutputBufferLength,
            l_ulIoControlCode, &p_pIrp->IoStatus.Information);
        break;
    }

    // Complete the IRP
    IoCompleteRequest( p_pIrp, IO_NO_INCREMENT );
    return p_pIrp->IoStatus.Status;
}//NTSTATUS TracerDispatch

/*-----------------------------------------------------------------

   FUNCTION: TracerUnload

   PURPOSE:  Called by the OS before unloading our driver
      
   PARAMETERS:                   .
                                  
   IN PDRIVER_OBJECT p_pDriverObject - our driver object

_________________________________________________________________*/

VOID TracerUnload
( 
    IN PDRIVER_OBJECT p_pDriverObject 
)
{
    MYTRACE(TF_Driver, "Entering TracerUnload");

    WCHAR          l_szLinkName[]  = TRACER_DOS_DEVICE_NAME_W;
    UNICODE_STRING l_LinkName;

    // Delete the symbolic link for our device
    RtlInitUnicodeString( &l_LinkName, l_szLinkName );
    IoDeleteSymbolicLink( &l_LinkName );

    //Delete our device
    if ( g_pTracerDevice != p_pDriverObject->DeviceObject )
    {
        IoDeleteDevice( g_pTracerDevice ); 
        g_pTracerDevice = NULL;
    }
    IoDeleteDevice( p_pDriverObject->DeviceObject );
    __try
    {
        g_Data.ShutDownKrnlTrcDriver();
    }
    __except(1)
    {
        MYTRACE(TF_Error, "ERROR: Exception in ShutDownKrnlTrcDll."
            " Restart the PC");
    }

    MYTRACE(TF_Driver, "Leaving TracerUnload");

    //Deinitialize the trace
    bEnableTraceToOutputLog(NULL, TF_Error, (ULONG)-1);//disable it
    vTraceUnInitialize();
}//VOID TracerUnload

/*-----------------------------------------------------------------

   FUNCTION: NTSTATUS DriverEntry

   PURPOSE:  Standard entry point of NT kernel driver
      
   PARAMETERS:                   .
                                  
   IN PDRIVER_OBJECT  p_pDriverObject - our driver object
   IN PUNICODE_STRING p_pRegistryPath - registry path for our driver

   RETURN VALUE:
      NTSTATUS      0 on success, error code otherwise
_________________________________________________________________*/
extern "C" NTSTATUS DriverEntry
(
    IN PDRIVER_OBJECT  p_pDriverObject, 
    IN PUNICODE_STRING p_pRegistryPath 
)
{
    ULONG d;
    __asm push [d]
    __asm pop [d]

    //Initialize our trace immediately, so we will be able
    //to print error messages
    vTraceInitialize();
    vSetCurTraceMasks(TT_DebugMonitor, (ULONG)-1, (ULONG)-1);
    bEnableTraceToOutputLog(p_pDriverObject, TF_Error, (ULONG)-1);

    MYTRACE(TF_Driver, "Entering DriverEntry");

    NTSTATUS       l_Status;
    WCHAR          l_szDeviceName[]  = TRACER_KERNEL_DEVICE_NAME;
    UNICODE_STRING l_DeviceName;
    WCHAR          l_szLinkName[]  = TRACER_DOS_DEVICE_NAME_W;
    UNICODE_STRING l_LinkName;

    //Initialize the device name
    RtlInitUnicodeString (&l_DeviceName, l_szDeviceName);

    //Create the device for communications with our application
    l_Status = IoCreateDevice ( p_pDriverObject,
                                0,
                                &l_DeviceName,
                                FILE_DEVICE_TRACER,
                                0,
                                TRUE,
                                &g_pTracerDevice );

    if (!NT_SUCCESS(l_Status)) {
        MYTRACE(TF_Error, "ERROR: IoCreateDevice(%*S)",
            l_DeviceName.Length, l_DeviceName.Buffer);
    }

    // If successful, make a symbolic link that allows for the 
    // device object's access from Win32 programs
    if (NT_SUCCESS(l_Status)) {

        // Create a symbolic link to be used by the application
        // as a file name to create a handle to our device
        RtlInitUnicodeString (&l_LinkName, l_szLinkName);
        l_Status =IoCreateSymbolicLink(&l_LinkName, &l_DeviceName);
        if (!NT_SUCCESS(l_Status)) 
        {
            MYTRACE(TF_Error, "ERROR: IoCreateSymbolicLink(%*S)",
                l_LinkName.Length, l_LinkName.Buffer);
        }

        // Register dispatch routine
        p_pDriverObject->MajorFunction[IRP_MJ_CLOSE]      =
            p_pDriverObject->MajorFunction[IRP_MJ_CREATE] =
            p_pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] =
                TracerDispatch;
        p_pDriverObject->DriverUnload = TracerUnload;
    }

    if (NT_SUCCESS(l_Status)) 
    {
        //Initialize the tracer itself
        l_Status = g_Data.KrnlTrcDriverInit(p_pRegistryPath);
        if(!NT_SUCCESS(l_Status))
        {
            //If failed - deinitialize it
            g_Data.ShutDownKrnlTrcDriver();
        }
    }

    //In case of any error - deinitialize everything and return 
    //the error to the OS, which will unload us
    if (!NT_SUCCESS(l_Status)) 
    {

        MYTRACE(TF_Error, "ERROR: DriverEntry failed. %x ", 
            l_Status);

        RtlInitUnicodeString (&l_LinkName, l_szLinkName);
        IoDeleteSymbolicLink( &l_LinkName );

        if( g_pTracerDevice ) 
        {
            IoDeleteDevice( g_pTracerDevice );
            g_pTracerDevice = NULL;
        }
        vTraceUnInitialize();
        return l_Status;
    }
    MYTRACE(TF_Driver, "Leaving DriverEntry OK");
    return l_Status;
}//NTSTATUS DriverEntry
