/******************************************************************
Module : EnumMods.cpp Class EnumModules to enumerate loaded modules
                    in NT kernel mode
             Written 1998,1999 by Dmitri Leman
Purpose: NT kernel mode tracer.
******************************************************************/
#include "Headers.h"
#include "Trace.h"
#include "EnumMods.h"

#define DRIVER_INFORMATION 11 //the code to obtain the list of 
    //loaded modules from ZwQuerySystemInformation 

//declare the undocumented ZwQuerySystemInformation function
extern "C" ULONG __stdcall 
    ZwQuerySystemInformation(ULONG, PULONG, ULONG, PVOID);


/*-----------------------------------------------------------------

   FUNCTION: EnumModules::vCleanup

   PURPOSE:  Initialize all members of the class without deleting
    the previous values
_________________________________________________________________*/
void EnumModules::vCleanup()
{
    m_ulNumBytes = 0;
    m_ulNumEntries = 0;
    m_pulBuffer = NULL;

    m_pDriverInfo = NULL;
    m_lCurIndex = 0;
}//EnumModules::vCleanup

/*-----------------------------------------------------------------

   FUNCTION: EnumModules::vFree

   PURPOSE:  Free the allocated resources
_________________________________________________________________*/

void EnumModules::vFree()
{
    if(m_pulBuffer)
    {
        ExFreePool(m_pulBuffer);
    }
    vCleanup();
}//EnumModules::vFree

/*-----------------------------------------------------------------

   FUNCTION: EnumModules::bLoadInfo

   PURPOSE:  Obtain the list of all loaded NT kernel mode modules
      
   PARAMETERS:          
                                  
   bool p_bForceReload  - if the class already hold a data, delete
                          it and load new.

   RETURN VALUE:
      bool          true on success, false on failure
_________________________________________________________________*/
bool EnumModules::bLoadInfo
(
    bool p_bForceReload
)
{
    if(p_bForceReload)
    {
        vFree();
    }
    else
    {
        if(m_pulBuffer)
        {
            return true;
        }
    }

    //First call the Query with zero size to obtain the required 
    //size of the buffer

    ULONG l_ulTemp = 0;
    ZwQuerySystemInformation
        (DRIVER_INFORMATION, &l_ulTemp, 0, &m_ulNumBytes);

    //ZwQuerySystemInformation will return an array of DriverInfo
    //structures. But the reported m_ulNumBytes is less than the 
    //size of the array, because the last structure is cut short
    //depending on the actual size of the string.
    //We will allocate the full size to simplify our calculations.

    int l_iSizeHeader = 3*sizeof(ULONG);
    int l_iNumDrivers = 
        (m_ulNumBytes - l_iSizeHeader + sizeof(DriverInfo)-1) /
        sizeof(DriverInfo);
    m_ulNumBytes = 
        l_iSizeHeader + l_iNumDrivers * sizeof(DriverInfo);

    m_pulBuffer = (PULONG)ExAllocatePool(PagedPool, m_ulNumBytes);

    if (m_pulBuffer == NULL)
    {
        MYTRACE(TF_Error, 
            "ERROR: ExAllocatePool(%d) failed", m_ulNumBytes);
        return false;
    }

    //Now get the array of driver information structures into 
    //the buffer
    ULONG l_ulBytesReturned;
    ZwQuerySystemInformation(DRIVER_INFORMATION, m_pulBuffer, 
        m_ulNumBytes, &l_ulBytesReturned);

    m_ulNumEntries = m_pulBuffer[0];
    m_pDriverInfo = (DriverInfo *)
        (((char*)m_pulBuffer) + l_iSizeHeader);
    m_lCurIndex = 0;

    if((int)m_ulNumEntries != l_iNumDrivers)
    {
        MYTRACE(TF_Error, 
            "ERROR: ZwQuerySystemInformation returned %d drivers."
            "We expected %d", m_ulNumEntries, l_iNumDrivers);
    }
    return true;
}//bool EnumModules::bLoadInfo

/*-----------------------------------------------------------------

   FUNCTION: EnumModules::pGetFirst

   PURPOSE: Get a pointer to the information about the first module
      
   RETURN VALUE:
      DriverInfo *  - pointer to the structure, which keeps the
                      information about the module - name and base
_________________________________________________________________*/
const EnumModules::DriverInfo * EnumModules::pGetFirst()
{
    bLoadInfo(false);//make sure that the information is obtained

    m_pDriverInfo = (DriverInfo *)(m_pulBuffer + 3);

    //prepare for pGetNext(), which will increment these values
    m_lCurIndex = -1;
    m_pDriverInfo--;
    return pGetNext();
}//const EnumModules::DriverInfo * EnumModules::pGetFirst()

/*-----------------------------------------------------------------

   FUNCTION: EnumModules::pGetNext

   PURPOSE:  Get a pointer to the information about the next module
      
   RETURN VALUE:
      DriverInfo *  - pointer to the structure, which keeps the
                      information about the module - name and base
_________________________________________________________________*/

const EnumModules::DriverInfo * EnumModules::pGetNext()
{
    if(ULONG(m_lCurIndex + 1) >= m_ulNumEntries)
    {
        return NULL;
    }
    m_lCurIndex++;
    m_pDriverInfo++;

    int l_iSpaceLeft = 
        m_ulNumBytes - (ULONG(m_pDriverInfo) - ULONG(m_pulBuffer));
    if(l_iSpaceLeft < sizeof(DriverInfo))
    {
        MYTRACE(TF_Error, 
            "ERROR: m_lCurIndex %d "
            "l_iSpaceLeft(%d) < sizeof(DriverInfo)(%d)", 
            m_lCurIndex, l_iSpaceLeft, sizeof(DriverInfo));
        return NULL;
    }
    return m_pDriverInfo;
}//const EnumModules::DriverInfo * EnumModules::pGetNext()

/*-----------------------------------------------------------------

   FUNCTION: EnumModules::GetModuleHandle

   PURPOSE:  This is an replacement of GetModuleHandle Win32 API.
             It searches the list of loaded modules for the
             specified name and returns the base address of this
             module.
      
   PARAMETERS:                   .
                                  
   const char * p_pszModuleName   - name of the module
   RETURN VALUE:
      ULONG         The base address of the module or 0
_________________________________________________________________*/

ULONG  EnumModules::GetModuleHandle
(
    const char * p_pszModuleName
)
{
    if(!bLoadInfo(false))
    {
        return 0;
    }//if(!bLoadInfo(false))
    
    DriverInfo * l_pDriverInfo = (DriverInfo *)(m_pulBuffer + 3);
    for(ULONG i = 0; i < m_ulNumEntries; i++)
    {
        char * l_pszName = 
            strrchr(l_pDriverInfo[i].m_szName, '\\');
        if(l_pszName)
        {
            l_pszName++;
        }
        else
        {
            l_pszName = l_pDriverInfo[i].m_szName;
        }
        for(int j = 0; ;j++)
        {
            if(!p_pszModuleName[j])
            {
                if(!l_pszName[j] || l_pszName[j] == '.')
                {//The names are equal.
                    return l_pDriverInfo[i].m_ulAddress;
                }
                break;
            }
            if(toupper(p_pszModuleName[j])!=toupper(l_pszName[j]))
                break;
        }//for(int j = 0; j++)
    }
    return 0;
}//ULONG  EnumModules::GetModuleHandle

