/******************************************************************
Module : FileSupp.cpp    - wrapper around file API
             Written 1998,1999 by Dmitri Leman
Purpose: NT kernel mode tracer.
******************************************************************/

#include "Headers.h"
#include "FileSupp.h"
#include "Trace.h"

/*-----------------------------------------------------------------

   FUNCTION: MyFileImpl::Cleanup()

   PURPOSE: Initialize the file handle without freeing it
-----------------------------------------------------------------*/      
void MyFileImpl::Cleanup()
{
#if !defined(_DRIVER)
    m_hFileHandle = INVALID_HANDLE_VALUE;
#else
    m_hFileHandle = NULL;
#endif
}

/*-----------------------------------------------------------------

   FUNCTION: ULONG MyFileImpl::OpenFile

   PURPOSE:  To open a file under Win32 or NT kernel mode.
      
   PARAMETERS:                   .
                                  
   const char * p_pszFileName     name of the file (0 terminated)
   bool         p_bWrite          open for writing if true, 
                                  reading otherwise
   bool         p_bPrependQuestionMarks - if working in kernel mode
                                  prepend "\??\" to the file name
   PUNICODE_STRING p_FileName     (for drivers only) If != NULL
                                  this name will be used instead of
                                  p_pszFileName
   RETURN VALUE:
      ULONG     0 on success, in case of error return Win32 error
        (result of GetLastError) or NT kernel status.
-----------------------------------------------------------------*/
ULONG MyFileImpl::OpenFile
(
    const char * p_pszFileName,
    bool         p_bWrite,
    bool         p_bPrependQuestionMarks,
#if defined(_DRIVER)
    PUNICODE_STRING p_FileName
#else
    void *
#endif
)
{
    CloseFile();//if a file is opened, close it
#if !defined(_DRIVER)
    DWORD l_dwDesiredAccess = p_bWrite? GENERIC_WRITE:GENERIC_READ;
    DWORD l_dwCreationDisposition = p_bWrite? 
        CREATE_ALWAYS : OPEN_EXISTING;

    m_hFileHandle = CreateFile(p_pszFileName, l_dwDesiredAccess, 0,
        NULL, l_dwCreationDisposition, FILE_ATTRIBUTE_NORMAL,NULL);
    if(m_hFileHandle == INVALID_HANDLE_VALUE)
    {
        DWORD l_dwError = GetLastError();

        MYTRACE(TF_Error,
            "ERROR: Cannot open file \"%s\" for %s. %s", 
            p_pszFileName, p_bWrite? "writing" : "reading", 
            ERR_EXPL(l_dwError));
        return l_dwError;
    }
    return 0;
#else
    if(KeGetCurrentIrql() > PASSIVE_LEVEL)
        return STATUS_UNSUCCESSFUL;
    ANSI_STRING l_AnsiName;
    UNICODE_STRING l_UnicodeName;
    NTSTATUS l_Status;

    if(!p_FileName)
    {
        RtlInitAnsiString(&l_AnsiName, p_pszFileName);

        l_UnicodeName.Length = 0;
        l_UnicodeName.MaximumLength = 
                    (l_AnsiName.Length + 5) * sizeof(WCHAR);
                    
        if(NULL == ( l_UnicodeName.Buffer = (PWCHAR)
            ExAllocatePool(PagedPool,l_UnicodeName.MaximumLength)))
        {
            MYTRACE(TF_Error, "ERROR: ExAllocatePool(%d) failed",
                l_UnicodeName.MaximumLength);
            return STATUS_INSUFFICIENT_RESOURCES;
        }

        if(p_bPrependQuestionMarks)
        {
            l_UnicodeName.Buffer[0] = '\\';
            l_UnicodeName.Buffer[1] = '?';
            l_UnicodeName.Buffer[2] = '?';
            l_UnicodeName.Buffer[3] = '\\';

            l_UnicodeName.Buffer += 4;
            l_UnicodeName.MaximumLength -= 4 * sizeof(WCHAR);
        }
        RtlAnsiStringToUnicodeString
            (&l_UnicodeName, &l_AnsiName, false);
        if(p_bPrependQuestionMarks)
        {
            l_UnicodeName.Buffer -= 4;
            l_UnicodeName.Length += 4 * sizeof(WCHAR);
            l_UnicodeName.MaximumLength += 4 * sizeof(WCHAR);
        }
        p_FileName = &l_UnicodeName;
    }//if(!p_FileName)
    OBJECT_ATTRIBUTES l_Attr;
    InitializeObjectAttributes(&l_Attr, p_FileName, 
        OBJ_CASE_INSENSITIVE, NULL, NULL);

    IO_STATUS_BLOCK l_IoStatus;
    
    l_IoStatus.Status = 0;
    l_IoStatus.Information = 0;

    ACCESS_MASK l_AccessMask = p_bWrite? 
        GENERIC_READ | GENERIC_WRITE | SYNCHRONIZE :
        GENERIC_READ | SYNCHRONIZE;
    ULONG l_dwShare = p_bWrite? 0 : FILE_SHARE_READ;
    ULONG l_dwCreateDisposition = p_bWrite?
        FILE_OVERWRITE_IF : FILE_OPEN;

    l_Status = ZwCreateFile(&m_hFileHandle, l_AccessMask,
        &l_Attr, &l_IoStatus, 0, 
        FILE_ATTRIBUTE_NORMAL, l_dwShare, l_dwCreateDisposition,
        FILE_NON_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT,
        NULL, 0);
    if(!NT_SUCCESS(l_Status))
    {
        MYTRACE(TF_Error, 
            "ERROR : failed to open file %S for %s. Status %x",
            p_FileName->Buffer, 
            p_bWrite? "writing" : "reading", l_Status);
    }
    if(p_FileName == &l_UnicodeName)
        RtlFreeUnicodeString(p_FileName);

    if(!NT_SUCCESS(l_Status))
    {
        m_hFileHandle = NULL;
        return l_Status;
    }
    return STATUS_SUCCESS;
#endif
}//ULONG MyFileImpl::OpenFile

/*-----------------------------------------------------------------

   FUNCTION: ULONG MyFileImpl::CloseFile()

   PURPOSE:  Close the file opened by OpenFile
   
   RETURN VALUE:
      ULONG     0 on success, in case of error return Win32 error
        (result of GetLastError) or NT kernel status.
-----------------------------------------------------------------*/
ULONG MyFileImpl::CloseFile()
{
#if !defined(_DRIVER)
    DWORD l_dwError = 0;
    if(m_hFileHandle != INVALID_HANDLE_VALUE)
    {
        if(!CloseHandle(m_hFileHandle))
        {
            l_dwError = GetLastError();
            MYTRACE(TF_Error, "ERROR: CloseHandle failed. %s",
                ERR_EXPL(l_dwError));
        }
    }//if(m_hFileHandle != INVALID_HANDLE_VALUE)
    m_hFileHandle = INVALID_HANDLE_VALUE;
    return l_dwError;
#else
    NTSTATUS l_Status = STATUS_SUCCESS;
    if(m_hFileHandle)
    {
        l_Status = ZwClose(m_hFileHandle);
        if(!NT_SUCCESS(l_Status))
        {
            MYTRACE(TF_Error, 
                "ERROR : ZwClose failed. Status %x", l_Status);
        }
    }//if(m_hFileHandle)
    m_hFileHandle = NULL;
    return l_Status;
#endif
}//ULONG MyFileImpl::CloseFile()

/*-----------------------------------------------------------------

   FUNCTION: ULONG MyFileImpl::GetFileSize

   PURPOSE:  Get size of the file
   
   PARAMETERS:
    PLARGE_INTEGER  p_pReturnSize - return 64 bit size of the file
   RETURN VALUE:
      ULONG     0 on success, in case of error return Win32 error
        (result of GetLastError) or NT kernel status.
-----------------------------------------------------------------*/
ULONG MyFileImpl::GetFileSize
(
    OUT PLARGE_INTEGER  p_pReturnSize
)
{
#if !defined(_DRIVER)
    DWORD l_dwSizeLow = 0, l_dwSizeHigh = 0;
    DWORD l_dwError = 0;
    l_dwSizeLow = ::GetFileSize(m_hFileHandle, &l_dwSizeHigh);
    if(l_dwSizeLow == (DWORD)-1)
    {
        l_dwError = GetLastError();
        if(l_dwError)
        {
            MYTRACE(TF_Error, "ERROR: GetFileSize failed. %s",
                ERR_EXPL(l_dwError));
        }
    }
    if(p_pReturnSize)
    {
        p_pReturnSize->LowPart = l_dwSizeLow;
        p_pReturnSize->HighPart = l_dwSizeHigh;
    }
    return l_dwError;
#else
    IO_STATUS_BLOCK l_IOStatus;
    NTSTATUS        l_Status;
    FILE_STANDARD_INFORMATION   l_Info;

    l_Status = ZwQueryInformationFile(m_hFileHandle, &l_IOStatus,
        &l_Info, sizeof(l_Info), FileStandardInformation);
    if(!NT_SUCCESS(l_Status))
    {
        MYTRACE(TF_Error, "ERROR: ZwQueryInformationFile ret %x",
            l_Status);
        if(p_pReturnSize)
           p_pReturnSize->QuadPart = 0;
        return l_Status;
    }
    if(p_pReturnSize)
        *p_pReturnSize = l_Info.EndOfFile;
    return STATUS_SUCCESS;
#endif
}//ULONG MyFileImpl::GetFileSize

/*-----------------------------------------------------------------

   FUNCTION: ULONG MyFileImpl::ReadWriteFile

   PURPOSE:  Read or write from / to file
   
   PARAMETERS:
    IN  bool    p_bWrite         - true for write, false for read
    OUT void * p_lpBuffer  - buffer to read/write
    IN  ULONG  p_dwNumberOfBytes - number of bytes to read or write
    IN  OUT PLARGE_INTEGER p_pByteOffset   - if != NULL, read/write
        starting at this offset, which will be incremented.
    OUT ULONG * p_pdwNumberOfBytesProcessed - if != NULL, returns
        the number of bytes returned in p_lpBuffer. No error is
        returned if this number is smaller than requested.
        If == NULL and the number of bytes returned is smaller
        than requested, than an error will be returned.

   RETURN VALUE:
      ULONG     0 on success, in case of error return Win32 error
        (result of GetLastError) or NT kernel status.
-----------------------------------------------------------------*/
ULONG MyFileImpl::ReadWriteFile
(
    IN      bool    p_bWrite,
    OUT     void *  p_lpBuffer,
    IN      ULONG   p_dwNumberOfBytes,
    IN  OUT PLARGE_INTEGER p_pByteOffset,
    OUT     ULONG*  p_pdwNumberOfBytesProcessed
)
{
    if(p_pdwNumberOfBytesProcessed)
      *p_pdwNumberOfBytesProcessed = 0;
#if !defined(_DRIVER)
    DWORD l_dwBytesProcessed = 0;
    DWORD l_dwError = 0;

    if(p_pByteOffset)
    {
        DWORD l_dwOffsetLow = p_pByteOffset->LowPart;
        LONG  l_dwOffsetHigh = p_pByteOffset->HighPart;
        l_dwOffsetLow = SetFilePointer(m_hFileHandle, 
            l_dwOffsetLow, &l_dwOffsetHigh, FILE_BEGIN);
        if(l_dwOffsetLow == (DWORD)-1)
        {
            l_dwError = GetLastError();
            if(l_dwError)
            {
                MYTRACE(TF_Error, "ERROR in SetFilePointer. %s",
                    ERR_EXPL(l_dwError));
                return l_dwError;
            }
        }//if(l_dwOffsetLow == (DWORD)-1)
        if(l_dwOffsetLow != p_pByteOffset->LowPart ||
           l_dwOffsetHigh != p_pByteOffset->HighPart)
        {
            MYTRACE(TF_Error, 
                "ERROR SetFilePointer(%8x%8x) ret %8x%8x",
                p_pByteOffset->HighPart, p_pByteOffset->LowPart,
                l_dwOffsetHigh, l_dwOffsetLow);
            return ERROR_SEEK;
        }
    }//if(p_pByteOffset)

    if(!p_bWrite?
        WriteFile(m_hFileHandle, p_lpBuffer, 
        p_dwNumberOfBytes, &l_dwBytesProcessed, NULL):
        ReadFile(m_hFileHandle, p_lpBuffer, 
        p_dwNumberOfBytes, &l_dwBytesProcessed, NULL))
    {
        l_dwError = GetLastError();

        MYTRACE(TF_Error, 
            "ERROR: %s with size %d failed (%s)",
            p_bWrite? "WriteFile" : "ReadFile",
            p_dwNumberOfBytes, 
            ERR_EXPL(l_dwError));
        return l_dwError;
    }
    if(p_pByteOffset)
        p_pByteOffset->QuadPart += l_dwBytesProcessed;

    if(p_pdwNumberOfBytesProcessed)
      *p_pdwNumberOfBytesProcessed = l_dwBytesProcessed;
    else if(l_dwBytesProcessed < p_dwNumberOfBytes)
    {
        MYTRACE(TF_Error, 
            "ERROR: %s with size %d returned only %d bytes",
            p_bWrite? "WriteFile" : "ReadFile",
            p_dwNumberOfBytes, l_dwBytesProcessed);
        return ERROR_END_OF_MEDIA;
    }
    return l_dwError;
#else
    if(KeGetCurrentIrql() > PASSIVE_LEVEL)
        return STATUS_UNSUCCESSFUL;

    IO_STATUS_BLOCK l_IoStatus;

    l_IoStatus.Status = 0;
    l_IoStatus.Information = 0;

    NTSTATUS l_Status = 
        p_bWrite? 
        ZwWriteFile(m_hFileHandle, NULL, NULL, NULL,
        &l_IoStatus, (void*)p_lpBuffer, p_dwNumberOfBytes, 
        p_pByteOffset, NULL):
        ZwReadFile(m_hFileHandle, NULL, NULL, NULL,
        &l_IoStatus, (void*)p_lpBuffer, p_dwNumberOfBytes, 
        p_pByteOffset, NULL);
    if(!NT_SUCCESS(l_Status))
    {
        MYTRACE(TF_Error, "ERROR: %s with size %d failed (%x)",
            p_bWrite? "ZwWriteFile" : "ZwReadFile",
            p_dwNumberOfBytes, l_Status);
        return l_Status;
    }
    if(p_pByteOffset)
        p_pByteOffset->QuadPart += l_IoStatus.Information;

    if(p_pdwNumberOfBytesProcessed)
      *p_pdwNumberOfBytesProcessed = l_IoStatus.Information;
    else if(l_IoStatus.Information < p_dwNumberOfBytes)
    {
        MYTRACE(TF_Error, 
            "%s with size %d returned only %d bytes",
            p_bWrite? "ZwWriteFile" : "ZwReadFile",
            p_dwNumberOfBytes, l_IoStatus.Information);
        return STATUS_END_OF_FILE;
    }
    return STATUS_SUCCESS;
#endif
}//ULONG MyFileImpl::ReadWriteFile

ULONG MyFileImpl::FlushBuffers()
{
#if !defined(_DRIVER)
    DWORD l_dwError = 0;
    if(!FlushFileBuffers(m_hFileHandle))
    {
        l_dwError = GetLastError();
        MYTRACE(TF_Error, "ERROR: FlushFileBuffers failed. %s",
            ERR_EXPL(l_dwError));
    }
    return l_dwError;
#else
        return STATUS_UNSUCCESSFUL;
#endif
}//NTSTATUS MyFileImpl::FlushBuffers
