/******************************************************************
Module : InterMap.cpp - collections of modules, interfaces, 
          functions and parameters used during config file parsing
             Written 1998,1999 by Dmitri Leman
Purpose: COM/OLE tracer and NT kernel mode tracer.
*****************************************************************/
#include "Headers.h"
#include "Trace.h"
#include "parmtype.h"
#include "CfgMgr.h"


#include "InterMap.h"

/*-----------------------------------------------------------------

   FUNCTION: NamedArray<T>::iGetSize

   PURPOSE:  Get size of collection (derived classes will used it 
        to know how many interfaces, or functions they have)
      
   RETURN VALUE:
        int         Number of elements
-----------------------------------------------------------------*/
template<class T>
int NamedArray<T>::iGetSize()
{
    return m_Array.size();
}//int NamedArray<T>::iGetSize

/*-----------------------------------------------------------------

   FUNCTION: NamedArray<T>::iFind

   PURPOSE:  Find an object(interface, DLL or function) by the name.
             Return index of the object in the table and, 
             if asked, the pointer to the object.
      
   PARAMETERS:                   .
                                  
   const string & p_rstrName         name to look for
   T **           p_ppReturnObject   return pointer to the object

   RETURN VALUE:
        int         Index of the object or -1 if not found
-----------------------------------------------------------------*/
template<class T>
int NamedArray<T>::iFind
(
    const string & p_rstrName,
    T **           p_ppReturnObject/* = NULL*/
)
{
    MapNameToIndex::iterator l_Iter = m_Map.find(p_rstrName);
    if(l_Iter == m_Map.end())
    {
        if(p_ppReturnObject)
          *p_ppReturnObject = NULL;
        return -1;
    }
    if(p_ppReturnObject)
      *p_ppReturnObject = pGet(l_Iter->second);
    return l_Iter->second;
}//int NamedArray<T>::iFind

/*-----------------------------------------------------------------

   FUNCTION: NamedArray<T>::pGet

   PURPOSE:  Get an object from the collection by the index
      
   PARAMETERS:                   .
                                  
   int p_iIndex                   Index of the object

   RETURN VALUE:
    T * Pointet to the object or -1 if no such index exists
-----------------------------------------------------------------*/
template<class T>
T * NamedArray<T>::pGet(int p_iIndex)
{
    if(DWORD(p_iIndex) >= m_Array.size())
        return NULL;
    return m_Array.at(p_iIndex);
}//T * NamedArray<T>::pGet

/*-----------------------------------------------------------------

   FUNCTION: NamedArray<T>::bAdopt

   PURPOSE:  Store the object in the collection. The object
    must be created by new operator and the caller looses
    the ownership of the object if this function returns TRUE.
    In this case the collection will own the object and will
    delete it at the end.
      
   PARAMETERS:                   .
                                  
   const string & p_rstrName      Name of the object
   T            * p_pObject       Object created by new operator
   int          * p_piReturnIndex return index in the table where
                                  the object was placed
   RETURN VALUE:
    BOOL        TRUE if the object was successfully adopted
-----------------------------------------------------------------*/
template<class T>
BOOL NamedArray<T>::bAdopt
(
    const string & p_rstrName,
    T            * p_pObject,
    int          * p_piReturnIndex
)
{
    int l_iNewIndex = m_Array.size();
    pair<MapNameToIndex::iterator,bool> l_Res = 
        m_Map.insert(pair<string,int>(p_rstrName, l_iNewIndex));

    BOOL l_bInsertOK = l_Res.second? TRUE : FALSE;
    //if false - insert failed because the element already exists

    int l_iIndex = l_Res.first->second;
    if(p_piReturnIndex)
      *p_piReturnIndex = l_iIndex;
    
    if(l_bInsertOK)
        m_Array.push_back(p_pObject);

    return l_bInsertOK;
}//bool NamedArray<T>::bAdopt

/*-----------------------------------------------------------------

   FUNCTION: NamedArray<T>::vFree

   PURPOSE:  Delete all objects in the collection
      
-----------------------------------------------------------------*/
template<class T>
void NamedArray<T>::vFree()
{
    Array::iterator l_Iter = m_Array.begin();
    Array::iterator l_IterEnd = m_Array.end();
    while(l_Iter != l_IterEnd)
    {
        T * l_pObject = *l_Iter;
        delete l_pObject;
        l_Iter++;
    }
    m_Array.clear();
    m_Map.clear();
}//void NamedArray<T>::vFree()


/*-----------------------------------------------------------------

   FUNCTION: AllInterfacesUnderConstruction::
                AllInterfacesUnderConstruction

   PURPOSE:  constructor
-----------------------------------------------------------------*/
AllInterfacesUnderConstruction::AllInterfacesUnderConstruction()
:m_iIUnknownIndex(-1), m_iIDispatchIndex(-1)
{}//AllInterfacesUnderConstruction::AllInterfacesUnderConstruction

/*-----------------------------------------------------------------

   FUNCTION: AllInterfacesUnderConstruction::iAddInterface

   PURPOSE:  Add a new interface object
      
   PARAMETERS:                   .
                                  
   const string & p_rstrName       name of the interface
   const IID    & p_rID            IID of the interface
   InterfaceUnderConstruction ** p_ppReturnInterface   - return
                                    pointer to the new object
   RETURN VALUE:
      int       Index of the interface in a table or -1 on error
-----------------------------------------------------------------*/
int AllInterfacesUnderConstruction::iAddInterface
(
    const string & p_rstrName,
    const string & p_rstrBase,
    const IID    & p_rID,
    InterfaceUnderConstruction ** p_ppReturnInterface
)
{
    InterfaceUnderConstruction * l_pBase = NULL;
    int l_iBaseIndex = -1;
    if(p_rstrBase.length())
    {
        l_iBaseIndex = m_Interfaces.iFind(p_rstrBase, &l_pBase);
        if(l_iBaseIndex < 0 || !l_pBase)
        {
            MYTRACE(TF_Error, 
                "ERROR: cannot find base interface %s",
                p_rstrBase.c_str());
            return -1;
        }
    }//if(p_rstrBase.length())

    int l_iIndex = -1;
    InterfaceUnderConstruction * l_pInt = 
        new InterfaceUnderConstruction(
        InterfaceUnderConstruction::TI_Interface, 
        p_rstrName, p_rID);
    if(!m_Interfaces.bAdopt(p_rstrName, l_pInt, &l_iIndex))
    {
        //It may happen if the interface was already declared
        //by a forward reference.
        InterfaceUnderConstruction * l_pOrigInt = 
            m_Interfaces.pGet(l_iIndex);
        if(!l_pOrigInt || 
           !l_pOrigInt->bReplaceForwardRefByInterface(l_pInt))
        {
            delete l_pInt;
            MYTRACE(TF_Error, "ERROR: interface %s duplicated", 
                p_rstrName.c_str());
            if(p_ppReturnInterface)
                *p_ppReturnInterface = NULL;
            return -1;
        }
        delete l_pInt;
        l_pInt = l_pOrigInt;
    }
    if(l_pBase)
        l_pInt->vAcceptBase(l_iBaseIndex, *l_pBase);

    if(p_ppReturnInterface)
        *p_ppReturnInterface = l_pInt;

    if(!memcmp(&p_rID, &IID_IUnknown, sizeof(IID_IUnknown)))
        m_iIUnknownIndex = l_iIndex;
    else if(!memcmp(&p_rID, &IID_IDispatch, sizeof(IID_IDispatch)))
        m_iIDispatchIndex = l_iIndex;
    return l_iIndex;
}//int AllInterfacesUnderConstruction::iAddInterface

/*-----------------------------------------------------------------

   FUNCTION: AllInterfacesUnderConstruction::iAddOrFindDLL

   PURPOSE:  Add new DLL object or locate existing one
      
   PARAMETERS:                   .
                                  
   const string & p_rstrName      DLL name 
   InterfaceUnderConstruction ** p_ppReturnInterface - return
                                    pointer to a new object
   RETURN VALUE:
      int       Index of the interface in a table or -1 on error
-----------------------------------------------------------------*/
int AllInterfacesUnderConstruction::iAddOrFindDLL
(
    const string & p_rstrName,
    InterfaceUnderConstruction ** p_ppReturnInterface
)
{
    InterfaceUnderConstruction * l_pInt = NULL;
    int l_iIndex = m_DLLs.iFind(p_rstrName, &l_pInt);
    if(l_iIndex < 0 || !l_pInt)
    {
        l_pInt = new InterfaceUnderConstruction(
            InterfaceUnderConstruction::TI_DLL, p_rstrName);
        if(!m_DLLs.bAdopt(p_rstrName, l_pInt, &l_iIndex))
        {
            delete l_pInt;
            MYTRACE(TF_Error, "ERROR: cannot add DLL %s", 
                p_rstrName.c_str());
            if(p_ppReturnInterface)
                *p_ppReturnInterface = NULL;
            return -1;
        }
    }
    if(p_ppReturnInterface)
        *p_ppReturnInterface = l_pInt;
    return l_iIndex;
}//int AllInterfacesUnderConstruction::iAddOrFindDLL

/*-----------------------------------------------------------------

   FUNCTION: AllInterfacesUnderConstruction::iGetInterfaceReference

   PURPOSE:  Some functions may return a specific interface. To 
    declare such function in the config file it may be necessary to 
    specify the name of the interface. If the interface was already 
    defined in the config file, we may simply get it's index. If it 
    was not defined, we should create a "forward reference"-a dummy
    interface object which has only a name.
        This function looks for a defined interface or creates a
    reference. When later the interface declaration will be found,
    InterfaceUnderConstruction::bReplaceForwardRefByInterface
    function will convert existing reference object to 
    a complete interface object.
      
   PARAMETERS:                   .
                                  
   const string & p_rstrName      Name of the reference.

   RETURN VALUE:
      int           index of the interface
-----------------------------------------------------------------*/
int AllInterfacesUnderConstruction::iGetInterfaceReference
(
    const string & p_rstrName
)
{
    int l_iIndex = m_Interfaces.iFind(p_rstrName);
    if(l_iIndex >= 0)
        return l_iIndex;
    //Try to declare a forward reference
    InterfaceUnderConstruction * l_pInt = 
        new InterfaceUnderConstruction(
        InterfaceUnderConstruction::TI_ForwardReference, 
        p_rstrName);
    if(!m_Interfaces.bAdopt(p_rstrName, l_pInt, &l_iIndex))
    {
        //Should not happen
        delete l_pInt;
        MYTRACE(TF_MessageBox, 
            "ERROR : cannot add new interface");
        return -1;
    }
    return l_iIndex;
}//int AllInterfacesUnderConstruction::iGetInterfaceReference

/*-----------------------------------------------------------------

   FUNCTION: AllInterfacesUnderConstruction::iAddIUnknown

   PURPOSE:  Creates a IUnknown interface object if it was not
    already defined.
      
   RETURN VALUE:
      int    Index or IUnknown interface in a table or -1 on error
-----------------------------------------------------------------*/
int AllInterfacesUnderConstruction::iAddIUnknown()
{
    if(m_iIUnknownIndex >= 0)
        return m_iIUnknownIndex;

    InterfaceUnderConstruction * l_pInt;
    m_iIUnknownIndex = iAddInterface("IUnknown", "", 
            IID_IUnknown, &l_pInt);
    if(m_iIUnknownIndex < 0 || !l_pInt)
        return -1;
    
    FunctionUnderConstruction * l_pFunc = NULL;

    if(!l_pInt->bAddFunction("QueryInterface", 0, &l_pFunc))
        return -1;

    l_pFunc->Params().bAddParam(PARAM_THIS, PF_In, FALSE, 0);
    l_pFunc->Params().bAddParam(PARAM_LPGUID, PF_In, FALSE, 0);
    l_pFunc->Params().bAddParam(PARAM_LPIUNKNOWN, PF_Out, TRUE, 
        1 //1 means the index of parameter holding the IID
        );
        

    if(!l_pInt->bAddFunction("AddRef", 1, NULL))
        return -1;
    if(!l_pInt->bAddFunction("Release", 2, NULL))
        return -1;
    return m_iIUnknownIndex;
}//int AllInterfacesUnderConstruction::iAddIUnknown

/*-----------------------------------------------------------------

   FUNCTION: InterfaceUnderConstruction::bAddFunction

   PURPOSE:  Add a function object to the interface
      
   PARAMETERS:                   .
                                  
   const string &  p_rstrFunctionName,   name of the function
   int    p_iFunctionIndex      index in the vTable (non-negative)
   FunctionUnderConstruction ** p_ppReturnFunction - return
                                    pointer to the new object

   RETURN VALUE:
      BOOL          TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL InterfaceUnderConstruction::bAddFunction
(
    const string &  p_rstrFunctionName,
    int             p_iFunctionIndex,
    FunctionUnderConstruction ** p_ppReturnFunction
)
{
    if(p_ppReturnFunction)
       *p_ppReturnFunction = NULL;
    if(m_Type == TI_ForwardReference)
    {
        MYTRACE(TF_Error, 
            "ERROR: cannot add function to dummy interface");
        return FALSE;
    }

    if(m_iMaxFunctionIndex < p_iFunctionIndex)
       m_iMaxFunctionIndex = p_iFunctionIndex;

    int l_iIndex = -1;
    FunctionUnderConstruction * l_pFunc = 
        new FunctionUnderConstruction
            (p_rstrFunctionName, p_iFunctionIndex);

    if(!bAdopt(p_rstrFunctionName, l_pFunc, &l_iIndex))
    {
        delete l_pFunc;
        MYTRACE(TF_Error, 
            "ERROR: function %s in interface %s is duplicated", 
            p_rstrFunctionName.c_str(), m_strName.c_str());
        return FALSE;
    }
    if(!l_pFunc)
        return FALSE;//should not happen
    m_iNumFunctions++;
    if(p_ppReturnFunction)
       *p_ppReturnFunction = l_pFunc;
    return TRUE;
}//BOOL InterfaceUnderConstruction::bAddFunction

/*-----------------------------------------------------------------

   FUNCTION: InterfaceUnderConstruction::
    bReplaceForwardRefByInterface

   PURPOSE: Convert existing reference object to a complete 
    interface object. See 
    AllInterfacesUnderConstruction::iGetInterfaceReference.

   PARAMETERS:                   .
   const InterfaceUnderConstruction * l_pInt A new interface which
                                             was just created.
                                           We will copy it's
                                           information to the
                                           former reference object

   RETURN VALUE:
      BOOL         TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL InterfaceUnderConstruction::bReplaceForwardRefByInterface
(
    const InterfaceUnderConstruction * l_pInt
)
{
    if(m_Type != TI_ForwardReference || 
            l_pInt->m_Type != TI_Interface)
        return FALSE;
    if(m_strName.compare(l_pInt->m_strName))
        return FALSE;//different names
    m_Type = TI_Interface;
    m_ID = l_pInt->m_ID;
    m_iMaxFunctionIndex = -1;
    //We do not copy the map
    return TRUE;
}//BOOL InterfaceUnderConstruction::bReplaceForwardRefByInterface

/*-----------------------------------------------------------------

   FUNCTION: ParamsUnderConstr::bAddParam

   PURPOSE:  Add function parameter
      
   PARAMETERS:                   .
                                  
    ParamType   p_Type              type of parameter
    BYTE        p_byFlags           flags from enum ParamFlags
    BOOL        p_bExtraBytePresent TRUE if this parameter has 
                                    the second byte
    BYTE        p_byExtra           the second byte

   RETURN VALUE:
      BOOL        TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL ParamsUnderConstr::bAddParam
(
    ParamType   p_Type, 
    BYTE        p_byFlags,
    BOOL        p_bExtraBytePresent,
    BYTE        p_byExtra
)
{
    if(bIs2ByteParam(p_Type) != (p_bExtraBytePresent? true:false))
    {
        MYTRACE(TF_Error, "ERROR : wrong type of param");
        return FALSE;
    }
    int l_iNeededSize = m_EncodedParams.size() + 
        p_bExtraBytePresent? 2 : 1;
    if(l_iNeededSize > MAX_NUM_BYTES_PARAMS)
    {
        MYTRACE(TF_Error, 
            "ERROR : too many parameters bytes (%d > %d)",
            l_iNeededSize, MAX_NUM_BYTES_PARAMS);
        return FALSE;
    }
    if(m_iNumParameters + 1 > MAX_NUM_PARAMS)
    {
        MYTRACE(TF_Error, "ERROR : too many parameters (%d > %d)",
            m_iNumParameters + 1, MAX_NUM_PARAMS);
        return FALSE;
    }
    m_iNumParameters++;
    m_byCommonFlags |= p_byFlags;
    m_EncodedParams.push_back(p_Type | p_byFlags);
    if(p_bExtraBytePresent)
        m_EncodedParams.push_back(p_byExtra);
    return TRUE;
}//BOOL ParamsUnderConstr::bAddParam

//Memory layout of configuration data in the binary config file:
//class ConfigManager
//  array of MyInterface[ConfigManager::m_iTotalNumInts]
//  array of ASM_APIFunctionStub[ConfigManager::m_iNumStubs]
//  variable size collection of names

/*-----------------------------------------------------------------

   FUNCTION: iExportData

   PURPOSE:  A function for exporting a simple byte array.
      
   PARAMETERS:                   .
                                  
   PBYTE        p_pMem          beginning of the whole config data
   int          p_iTotalSize    total size of the whole config data
   int          p_iCurOffset    number of bytes already filled       
   const char * p_pszData       data to be exported  
   int          p_iSize         size of data to be exported  
   const char * p_pszDescr      description of data for debugging  

   RETURN VALUE:
      int        return number of bytes added to the exported area
-----------------------------------------------------------------*/
int iExportData
(
    PBYTE   p_pMem,
    int     p_iTotalSize,
    int     p_iCurOffset,
    const char * p_pszData,
    int          p_iSize,
    const char * p_pszDescr
)
{
    if(p_iTotalSize - p_iCurOffset < p_iSize)
    {
        MYTRACE(TF_Error, 
            "ERROR : not enough space for %s", p_pszDescr);
        return -1;
    }
    if(p_pMem)
        memcpy(p_pMem + p_iCurOffset, p_pszData, p_iSize);
    return p_iSize;
}//int iExportData

/*-----------------------------------------------------------------

   FUNCTION: ParamsUnderConstr::iExport

   PURPOSE:  Export encoded function parameters
      
   PARAMETERS:                   .
                                  
   PBYTE        p_pMem          beginning of the whole config data
   int          p_iTotalSize    total size of the whole config data
   int          p_iCurOffset    number of bytes already filled       

   RETURN VALUE:
      int       return number of bytes added to the exported area
-----------------------------------------------------------------*/
int ParamsUnderConstr::iExport
(
    PBYTE   p_pMem,
    int     p_iTotalSize,
    int     p_iCurOffset
)
{
    int l_iNumBytes = m_EncodedParams.size();
    int l_iSize = sizeof(FunctionParameters) + 
        l_iNumBytes - 1;//1 byte is already there
    if(p_iTotalSize - p_iCurOffset < l_iSize)
    {
        MYTRACE(TF_Error, 
            "ERROR : not enough space for parameters");
        return -1;
    }
    if(m_iNumParameters > MAX_NUM_PARAMS || l_iNumBytes > 255)
    {
        MYTRACE(TF_Error, "ERROR : too many parametes");
        return -1;
    }
    
    if(p_pMem)
    {
        FunctionParameters * l_pParams = (FunctionParameters * )
            (p_pMem + p_iCurOffset);
        l_pParams->m_byNumParameters = 
            BYTE(m_iNumParameters) |  
            (m_byCommonFlags & PF_AllFlags);
        l_pParams->m_byNumBytes = BYTE(l_iNumBytes);
        for(int i = 0; i < l_iNumBytes; i++)
            l_pParams->m_byParams[i] = m_EncodedParams.at(i);
    }//if(p_pMem)
    return l_iSize;
}//int ParamsUnderConstr::iExport

/*-----------------------------------------------------------------

   FUNCTION: FunctionUnderConstruction::iExport

   PURPOSE:  export config data for the function
      
   PARAMETERS:                   .

   PBYTE        p_pMem          beginning of the whole config data
   int          p_iTotalSize    total size of the whole config data
   int          p_iCurOffset    number of bytes already filled       
   ConfigManager * p_pCfg       config data structure  
   DWORD   p_dwModuleNameOffset offset of the DLL or interface name
                                to which this function belongs  

   RETURN VALUE:
      int       return number of bytes added to the exported area
-----------------------------------------------------------------*/
int FunctionUnderConstruction::iExport
(
    PBYTE   p_pMem,
    int     p_iTotalSize,
    int     p_iCurOffset,
    ConfigManager * p_pCfg,
    DWORD   p_dwModuleNameOffset
)
{
    if(p_pCfg->m_iStubsFilled >= p_pCfg->m_iNumStubs)
    {
        //should never happen
        MYTRACE(TF_Error, "ERROR: too many functions");
        return -1;
    }
    DWORD l_dwOffsetStub = p_pCfg->m_dwOffsetStubs + 
        p_pCfg->m_iStubsFilled * sizeof(ASM_APIFunctionStub);
    p_pCfg->m_iStubsFilled++;

    int l_iTotalAddSize = 0;
    int l_iSize = iExportData(p_pMem, p_iTotalSize, p_iCurOffset, 
        m_strName.c_str(), m_strName.length()+1, "Function name");
    if(l_iSize < 0)
        return -1;
    DWORD l_dwFunctionNameOffset = p_iCurOffset;
    p_iCurOffset += l_iSize;
    l_iTotalAddSize += l_iSize;

    l_iSize = m_Params.iExport(p_pMem, p_iTotalSize, p_iCurOffset);
    if(l_iSize < 0)
        return -1;
    DWORD l_dwParamsOffset = p_iCurOffset;
    p_iCurOffset += l_iSize;
    l_iTotalAddSize += l_iSize;

    if(p_pMem)
    {
        ASM_APIFunctionStub * l_pStub = (ASM_APIFunctionStub * )
            (p_pMem + l_dwOffsetStub);
        l_pStub->vInitForFile(p_dwModuleNameOffset,
            l_dwFunctionNameOffset, m_iIndex, l_dwParamsOffset);
    }
    return l_iTotalAddSize;
}//int FunctionUnderConstruction::iExport

/*-----------------------------------------------------------------

   FUNCTION: InterfaceUnderConstruction::iExportFunctions

   PURPOSE: Export configuration data for one fucntion 
      
   PARAMETERS:                   .
                                  
   NamedArray<InterfaceUnderConstruction> & p_rAllIntr - collection
                                to which the interface belongs
   PBYTE   p_pMem               beginning of the whole config data    
   int     p_iTotalSize         total size of the whole config data  
   int     p_iCurOffset         number of bytes already filled       
   ConfigManager * p_pCfg       config data structure    
   DWORD   p_dwModuleNameOffset offset of the DLL or interface name
                                to which this function belongs

   RETURN VALUE:
      int          return number of bytes added to the exported area
-----------------------------------------------------------------*/
int InterfaceUnderConstruction::iExportFunctions
(
    NamedArray<InterfaceUnderConstruction> & p_rAllIntr,
    PBYTE   p_pMem,
    int     p_iTotalSize,
    int     p_iCurOffset,
    ConfigManager * p_pCfg,
    DWORD   p_dwModuleNameOffset
)
{
    int l_iTotalAddSize = 0;
    if(m_iBaseIndex >= 0)
    {
        InterfaceUnderConstruction * l_pBase = 
            p_rAllIntr.pGet(m_iBaseIndex);
        if(!l_pBase)
        {
            MYTRACE(TF_Error, "ERROR: cannot get base interface");
            return -1;
        }
        l_iTotalAddSize = l_pBase->iExportFunctions(p_rAllIntr, 
            p_pMem, p_iTotalSize, p_iCurOffset, p_pCfg,
            p_dwModuleNameOffset);
        if(l_iTotalAddSize < 0)
            return -1;
        p_iCurOffset += l_iTotalAddSize;
    }
    int l_iNumOurFunctions = iGetSize();
    for(int i = 0; i < l_iNumOurFunctions; i++)
    {
        FunctionUnderConstruction * l_pFun = pGet(i);
        if(l_pFun)
        {
            int l_iSize = l_pFun->iExport
                (p_pMem, p_iTotalSize, p_iCurOffset, p_pCfg,
                p_dwModuleNameOffset);
            if(l_iSize < 0)
                return -1;
            p_iCurOffset += l_iSize;
            l_iTotalAddSize += l_iSize;
        }
    }//for(int i = 0; i < l_iNumFunctions; i++)
    return l_iTotalAddSize;
}//int InterfaceUnderConstruction::iExportFunctions

/*-----------------------------------------------------------------

   FUNCTION: InterfaceUnderConstruction::iExport

   PURPOSE:  Export configuration data for one interface
      
   PARAMETERS:                   .
                                  
   NamedArray<InterfaceUnderConstruction> & p_rAllIntr - collection
                              to which this interface belongs
   PBYTE   p_pMem             beginning of the whole config data        
   int     p_iTotalSize       total size of the whole config data      
   int     p_iCurOffset       number of bytes already filled           
   ConfigManager * p_pCfg     config data structure       

   RETURN VALUE:
      int  return number of bytes added to the exported area
-----------------------------------------------------------------*/
int InterfaceUnderConstruction::iExport
(
    NamedArray<InterfaceUnderConstruction> & p_rAllIntr,
    PBYTE   p_pMem,
    int     p_iTotalSize,
    int     p_iCurOffset,
    ConfigManager * p_pCfg
)
{
    if(p_pCfg->m_iInterfacesFilled >= p_pCfg->m_iTotalNumInts)
    {
        //should never happen
        MYTRACE(TF_Error, "ERROR: too many interfaces");
        return -1;
    }
    int l_iNumFunctions = iGetNumFunctions();
    if(p_pCfg->m_iStubsFilled + l_iNumFunctions > 
        p_pCfg->m_iNumStubs)
    {
        //should never happen
        MYTRACE(TF_Error, "ERROR: too many functions");
        return -1;
    }

    MyInterface * l_pMyInt = p_pMem?
        p_pCfg->m_ArrayInterfaces + 
        p_pCfg->m_iInterfacesFilled : NULL;
    p_pCfg->m_iInterfacesFilled++;

    DWORD l_dwOffsetStubs = p_pCfg->m_dwOffsetStubs + 
        p_pCfg->m_iStubsFilled * sizeof(ASM_APIFunctionStub);

    int l_iTotalAddSize = 0;
    int l_iSize = iExportData(p_pMem, p_iTotalSize, p_iCurOffset, 
        m_strName.c_str(), m_strName.length()+1, "Interface name");
    if(l_iSize < 0)
        return -1;
    DWORD l_dwOffsetName = p_iCurOffset;
    p_iCurOffset += l_iSize;
    l_iTotalAddSize += l_iSize;

    if(l_pMyInt)
    {
        l_pMyInt->m_iNumFunctions = l_iNumFunctions;
        l_pMyInt->m_iMaxFunctionIndex = m_iMaxFunctionIndex;
        l_pMyInt->m_ID = m_ID;
        l_pMyInt->m_dwOffsetFunctions = l_dwOffsetStubs;
        l_pMyInt->m_dwOffsetName  = l_dwOffsetName;
        l_pMyInt->m_dwModuleBase = NULL;
        l_pMyInt->m_dwModuleBasePrevious = NULL;
    }    

    l_iSize = iExportFunctions(p_rAllIntr, p_pMem, p_iTotalSize, 
        p_iCurOffset, p_pCfg, l_dwOffsetName);
    if(l_iSize < 0)
        return -1;
    p_iCurOffset += l_iSize;
    l_iTotalAddSize += l_iSize;

    return l_iTotalAddSize;
}//int InterfaceUnderConstruction::iExport

/*-----------------------------------------------------------------

   FUNCTION: AllInterfacesUnderConstruction::iExportInterfaces

   PURPOSE:  Export configuration data for the specified
    collection of interfaces or DLLs
      
   PARAMETERS:                   .
                                  
   NamedArray<InterfaceUnderConstruction> & p_rInts - collection
                                        to export
   PBYTE   p_pMem             beginning of the whole config data        
   int     p_iTotalSize       total size of the whole config data      
   int     p_iCurOffset       number of bytes already filled           
   ConfigManager * p_pCfg     config data structure    

   RETURN VALUE:
      int       return number of bytes added to the exported area
-----------------------------------------------------------------*/
int AllInterfacesUnderConstruction::iExportInterfaces
(
    NamedArray<InterfaceUnderConstruction> & p_rInts,
    PBYTE   p_pMem,
    int     p_iTotalSize,
    int     p_iCurOffset,
    ConfigManager * p_pCfg
)
{
    int l_iNum = p_rInts.iGetSize();
    int l_iFilledSize = 0;
    for(int i = 0; i < l_iNum; i++)
    {
        InterfaceUnderConstruction * l_pInt = p_rInts.pGet(i);
        if(l_pInt)
        {
            int l_iSizeInt = l_pInt->iExport
                (p_rInts, p_pMem, p_iTotalSize, 
                p_iCurOffset, p_pCfg);
                
            if(l_iSizeInt < 0)
                return -1;

            l_iFilledSize += l_iSizeInt;
            p_iCurOffset += l_iSizeInt;
        }
    }//for(i = 0; i < l_iNum; i++)
    return l_iFilledSize;
}//AllInterfacesUnderConstruction::iExportInterfaces

/*-----------------------------------------------------------------

   FUNCTION: AllInterfacesUnderConstruction::iExport

   PURPOSE:  Export the entire configuration data
      
   PARAMETERS:                   .
                                  
   PBYTE   p_pMem            beginning of the whole config data         
   int     p_iTotalSize      total size of the whole config data       
   int     p_iCurOffset      number of bytes already filled            

   RETURN VALUE:
      int       return number of bytes added to the exported area
-----------------------------------------------------------------*/
int AllInterfacesUnderConstruction::iExport
(
    PBYTE   p_pMem,
    int     p_iTotalSize,
    int     p_iCurOffset
)
{
    int l_iNumDLLs = m_DLLs.iGetSize();
    int l_iNumInterfaces = m_Interfaces.iGetSize();
    int l_iNumDLLExports = 0;
    int l_iNumInterfaceMethods = 0;

    int i;
    for(i = 0; i < l_iNumDLLs; i++)
    {
        InterfaceUnderConstruction * l_pInt = m_DLLs.pGet(i);
        if(l_pInt)
            l_iNumDLLExports += l_pInt->iGetNumFunctions();
    }//for(i = 0; i < l_iNumDLLs; i++)
    for(i = 0; i < l_iNumInterfaces; i++)
    {
        InterfaceUnderConstruction * l_pInt = m_Interfaces.pGet(i);
        if(l_pInt)
            l_iNumInterfaceMethods += l_pInt->iGetNumFunctions();
    }//for(i = 0; i < l_iNumInterfaces; i++)

    int l_iTotalInts = l_iNumInterfaces + l_iNumDLLs;
    int l_iTotalStubs = l_iNumDLLExports + l_iNumInterfaceMethods;
    
    //Calculate and reserve space for config manager and
    //array of MyInterface structures
    int l_iSizeForConfigManager = sizeof(ConfigManager) +
        sizeof(MyInterface) * (l_iTotalInts - 1);
                    //1 element is inside ConfigManager
    if(p_iTotalSize - p_iCurOffset < l_iSizeForConfigManager)
    {
        MYTRACE(TF_Error, 
            "ERROR : not enough space for ConfigManager");
        return -1;
    }

    ConfigManager l_Cfg;//dummy structure in case if p_pMem == NULL
    ConfigManager * l_pCfg = p_pMem ?
        (ConfigManager * )(p_pMem + p_iCurOffset) : &l_Cfg;
    p_iCurOffset += l_iSizeForConfigManager;

    //Now reserve space for array of ASM_APIFunctionStub
    int l_iSizeArrayStubs = 
        sizeof(ASM_APIFunctionStub)*l_iTotalStubs;
    if(p_iTotalSize - p_iCurOffset < l_iSizeArrayStubs)
    {
        MYTRACE(TF_Error, 
            "ERROR : not enough space for array of stubs");
        return -1;
    }
    DWORD l_dwOffsetStubs = p_iCurOffset;
    p_iCurOffset += l_iSizeArrayStubs;

    l_pCfg->m_iNumInterfaces = l_iNumInterfaces;
    l_pCfg->m_iNumModules    = l_iNumDLLs;
    l_pCfg->m_iTotalNumInts = l_iTotalInts;
    l_pCfg->m_iIUnknownIndex = m_iIUnknownIndex;
    l_pCfg->m_iIDispatchIndex = m_iIDispatchIndex;

    l_pCfg->m_iNumStubs = l_iNumDLLExports + l_iNumInterfaceMethods; 
    l_pCfg->m_iNumInterfStubs = l_iNumInterfaceMethods;
    l_pCfg->m_iNumExpStubs = l_iNumDLLExports;
    l_pCfg->m_dwOffsetStubs = l_dwOffsetStubs;

    l_pCfg->m_iInterfacesFilled = 0;
    l_pCfg->m_iStubsFilled = 0;

    int l_iSize;

    l_iSize = iExportInterfaces(m_Interfaces, p_pMem, p_iTotalSize, 
        p_iCurOffset, l_pCfg);
    if(l_iSize < 0)
        return -1;
    
    p_iCurOffset += l_iSize;
    
    l_iSize = iExportInterfaces(m_DLLs, p_pMem, p_iTotalSize, 
        p_iCurOffset,l_pCfg);
    if(l_iSize < 0)
        return -1;

    p_iCurOffset += l_iSize;
    return p_iCurOffset;
}//int AllInterfacesUnderConstruction::iExport
