/******************************************************************
Module : Intrcpt.cpp    - all business of interception
             Written 1998,1999 by Dmitri Leman
Purpose: Purpose: NT kernel mode tracer.
******************************************************************/
#include "Headers.h"
#include "Driver.h"
#include "NtImage.h"
#include "KrnlTrcD.h"
#include "StubMgr.h"
#include "log.h"
#include "PEHelper.h"
#include "parmtype.h"
#include "Trace.h"
#include "return.h"
#include "Intrcpt.h"
#include "FileSupp.h"
#include "EnumMods.h"

/*-----------------------------------------------------------------

   FUNCTION: Interceptor::IsItInterceptedFunction

   PURPOSE:  Determine whether the specified function was 
    intercepted by us. We can do it because all our stubs has the
    same structure and points to our IntrerceptEntry function
      
   PARAMETERS:                   .
                                  
   void * p_pAddress  Address of the function (or our stub)

   RETURN VALUE:
      ASM_APIFunctionStub * pointer to the stub or NULL
-----------------------------------------------------------------*/
ASM_APIFunctionStub * Interceptor::IsItInterceptedFunction
(
    void * p_pAddress
)
{
    ASM_APIFunctionStub * l_pStub = (ASM_APIFunctionStub * )
        (ULONG(p_pAddress) - 
        offsetof(ASM_APIFunctionStub, m_Code));

    if(!MmIsAddressValid(l_pStub))
        return NULL;
    if(l_pStub->m_dwSignature != STUB_SIGNATURE)
        return NULL;
    if(l_pStub->m_Code.instr_PUSH != PUSH_INSTR)
        return NULL;
    if(l_pStub->m_Code.instr_RET != RET_INSTR)
        return NULL;
    if(l_pStub->m_Code.instr_CALL == RET_INSTR)
        return l_pStub;//this stub was shortcut
    if(l_pStub->m_Code.instr_CALL != CALL_INSTR)
        return NULL;
    if(l_pStub->m_Code.offset_LogFunction !=
        (ULONG)Logger::IntrerceptEntryNaked - 
        ((ULONG)(&l_pStub->m_Code.instr_CALL)+5))
        return NULL;
    return l_pStub;
}//ASM_APIFunctionStub * Interceptor::IsItInterceptedFunction

/*-----------------------------------------------------------------

   FUNCTION: GetEnclosingSectionHeader

   PURPOSE:  Given a relative virtual address in a module (RVA), 
             look up the section header that encloses it 
             and return a pointer to its IMAGE_SECTION_HEADER
   
   PARAMETERS:                   .
                                  
   ULONG             p_dwAddress  - RVA(relative virtual addreess)
   PIMAGE_NT_HEADERS p_pNTHeader  - pointer to the NT header of the
                                    module

   RETURN VALUE:
      PIMAGE_SECTION_HEADER   - ponter to the section hader or NULL
_________________________________________________________________*/

PIMAGE_SECTION_HEADER GetEnclosingSectionHeader
(
    ULONG             p_dwAddress,
    PIMAGE_NT_HEADERS p_pNTHeader
)
{
    PIMAGE_SECTION_HEADER l_pSection = (PIMAGE_SECTION_HEADER)
        ((UINT_PTR)p_pNTHeader +
         FIELD_OFFSET( IMAGE_NT_HEADERS, OptionalHeader ) +
         ((PIMAGE_NT_HEADERS)(p_pNTHeader))->
         FileHeader.SizeOfOptionalHeader);
    for(int i=0; i < p_pNTHeader->FileHeader.NumberOfSections; 
        i++, l_pSection++ )
    {
        // Is the address within this section?
        if ( (p_dwAddress >= l_pSection->VirtualAddress) && 
             (p_dwAddress < (l_pSection->VirtualAddress + 
                             l_pSection->Misc.VirtualSize)))
            return l_pSection;
    }
    return NULL;
}//GetEnclosingSectionHeader

/*-----------------------------------------------------------------

   FUNCTION: Interceptor::iInterceptImportedFunctionsInModule

   PURPOSE:  Intercept functions imported by the specified
    module as we were instructed by the config file
      
   PARAMETERS:                   .
                                  
    PVOID p_pModuleBaseAddress  Base address of the module
    PCSTR p_pszFileName         Name of the file for this module  
    bool  p_bRestore            Undo the interception
    bool  p_bUseFile - If true -read the import info from the file.
                        NT drivers may have import descriptors in
                        "INIT" section, which is discardable.
                        Therefore, we will try to read it from file

   RETURN VALUE:
      int       Number of intercepted functions
                0 if no functions of interest were found
                -1 on error
-----------------------------------------------------------------*/
int Interceptor::iInterceptImportedFunctionsInModule
(
    PVOID p_pModuleBaseAddress,
    PCSTR p_pszFileName,
    bool  p_bRestore,
    bool  p_bUseFile
)
{
    if(!p_pModuleBaseAddress || 
        ULONG(p_pModuleBaseAddress) < 0x80000000)
        return 0; // we don't work with user addresses
    PIMAGE_NT_HEADERS l_pNTHeader = (PIMAGE_NT_HEADERS)
        pCheckModuleHeader(p_pModuleBaseAddress);
    if(!l_pNTHeader)
    {
        MYTRACE(TF_Error, 
            "ERROR : cannot recognize module header %s at %x",
            p_pszFileName, p_pModuleBaseAddress);
        return -1;
    }

    IMAGE_IMPORT_DESCRIPTOR* l_pImportDesc; 

    ULONG l_dwImportsStartRVA = 
        l_pNTHeader->OptionalHeader.DataDirectory
        [IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress;
    if ( !l_dwImportsStartRVA )
    {
        MYTRACE(TF_Error, 
            "ERROR : cannot find import table in module %x",
            p_pModuleBaseAddress);
        return -1;
    }

    l_pImportDesc = (IMAGE_IMPORT_DESCRIPTOR*)
        (ULONG(p_pModuleBaseAddress) + ULONG(l_dwImportsStartRVA));

    // Find the section (usually .idata) that contains the imports.
    PIMAGE_SECTION_HEADER l_pSection = 
       GetEnclosingSectionHeader(l_dwImportsStartRVA, l_pNTHeader);
    if ( !l_pSection )
    {
        MYTRACE(TF_Error, 
            "ERROR : cannot find a section containing the "
            "import table in module %x",
            p_pModuleBaseAddress);
        return -1;
    }

    ULONG l_dwImportFileOffset = 0;
    NTSTATUS l_Status;
    MyFileLocal l_File;
    if(p_bUseFile)
    {
        long l_lDelta = (long)(l_pSection->VirtualAddress - 
                              l_pSection->PointerToRawData);
        //This delta is the difference between the location of this
        //section in memory and in the file. In order to find the 
        //offset in the file of some object, which belongs to this 
        //section, we should subtract the delta from this 
        //object's RVA.

        l_dwImportFileOffset = l_dwImportsStartRVA - l_lDelta;

        //Some file names returned by ZwQuerySystemInformation
        //does not have any directory information. They seems to
        //always be in Winnt\system\drivers. Other have
        //the directory in the form "Winnt\System\". In both cases
        //we should make correction in the directory name to
        //be able to open the file.
        char l_szFilePath[2*_MAX_PATH];
        if(!strchr(p_pszFileName, '\\'))
        {
            //this file name has no directory. 
            //May be it is in Windows\System32\Drivers
            //Try to contruct a full path
            _snprintf(l_szFilePath, sizeof(l_szFilePath),
                "\\SystemRoot\\System32\\Drivers\\%s",
                p_pszFileName);
            p_pszFileName = l_szFilePath;
        }//if(!strchr(p_pszFileName, '\\'))
        else if(!_strnicmp(p_pszFileName, 
            g_Data.m_pParams->m_szWindowsDir+2,
            g_Data.m_pParams->m_iLengthWindowsDir-2))
        {
            //In this case the module name is in form 
            //"\WinNT\some_dir\module_name"
            //Try to contruct a full path
            _snprintf(l_szFilePath, sizeof(l_szFilePath),
                "\\SystemRoot\\%s",
                p_pszFileName + 
                g_Data.m_pParams->m_iLengthWindowsDir-1);

            p_pszFileName = l_szFilePath;
        }
                        
        l_Status = l_File.OpenFile(p_pszFileName,
            false,false, NULL);

        if(!NT_SUCCESS(l_Status))
        {
            return -1;
        }
    }//if(p_bUseFile)

    int l_iNumFunctions = 0;
    while ( 1 )
    {
        IMAGE_IMPORT_DESCRIPTOR l_ImportDescr;
        if(p_bUseFile)
        {
            //Read the next IMAGE_IMPORT_DESCRIPTOR structure
            //from the file
            LARGE_INTEGER l_liOffset;
            l_liOffset.QuadPart = l_dwImportFileOffset;

            l_Status = l_File.ReadWriteFile(false, 
                &l_ImportDescr, sizeof(l_ImportDescr),
                &l_liOffset, NULL);
            if(!NT_SUCCESS(l_Status))
            {
                break;
            }
            l_pImportDesc = &l_ImportDescr;
        }
        //If p_bUseFile == false, we already have l_pImportDesc
        //pointing to the location in memory, where the import
        //descriptor is, unless it is discarded.

        // See if we've reached an empty IMAGE_IMPORT_DESCRIPTOR
        if ( (l_pImportDesc->TimeDateStamp==0 ) && 
             (l_pImportDesc->Name==0) )
        {
            break;
        }
        for (IMAGE_THUNK_DATA * l_pThunk = (IMAGE_THUNK_DATA*)
                (ULONG(p_pModuleBaseAddress) + 
                 ULONG(l_pImportDesc->FirstThunk));
            l_pThunk->u1.Function; l_pThunk++)
        {
            if(p_bRestore)
            {
                ASM_APIFunctionStub * l_pStub = 
                    IsItInterceptedFunction(l_pThunk->u1.Function);
                if(l_pStub)
                {
                    InterlockedExchange
                        ((long*)&l_pThunk->u1.Function, 
                        l_pStub->dwGetOriginalFunctionAddress());

                    MYTRACE(TF_Intercepted, 
                        "Import %s restored in %s",
                        l_pStub->m_dwFunctionNameOffset,
                        p_pszFileName);

                    l_iNumFunctions++;
                }
            }
            else
            {
                ASM_APIFunctionStub * l_pStub = 
                    g_Data.m_pConfigMgr->pFindStubByOrigAddress
                        ((void*)l_pThunk->u1.Function);

                if (l_pStub)
                {
                    ULONG l_dwHookAddress = 
                        (ULONG)&l_pStub->m_Code;

                    InterlockedExchange
                        ((long*)&l_pThunk->u1.Function, 
                        l_dwHookAddress);

                    MYTRACE(TF_Intercepted, 
                        "Import %s intercepted in %s",
                        l_pStub->m_dwFunctionNameOffset,
                        p_pszFileName);

                    l_iNumFunctions++;

                }//if(l_dwHookAddress)
            }
        }//for (l_pThunk...

        if(p_bUseFile)
            l_dwImportFileOffset += sizeof(l_ImportDescr);
        else
            l_pImportDesc++;
    }//while (1)
    l_File.CloseFile();
    return l_iNumFunctions;
}//int Interceptor::iInterceptImportedFunctionsInModule

/*-----------------------------------------------------------------

   FUNCTION: Interceptor::iProcessExportForOneModuleAndOneStub

   PURPOSE:  For the given module and stub update this stub's
    original function address and intercept the exported function

   RETURN VALUE: 
     0 if no function was intercepted
     1 of a function was intercepted
     -1 in case of an error
_________________________________________________________________*/
int Interceptor::iProcessExportForOneModuleAndOneStub
(
    PVOID                   p_pNTHeader,
    MyInterface         &   p_rInt,
    ASM_APIFunctionStub &   p_rStub,
    bool                    p_bUnhook
)
{
    if(p_bUnhook && !p_rStub.dwGetOriginalFunctionAddress())
    {
        //Nothing to do - the function was not hooked
        return 0;
    }

    if(!p_pNTHeader)
    {
        //Module is not loaded
        p_rStub.vSetOriginalFunctionAddress(0);
        return 0;//it is not an error
    }
    //Module is loaded. Get the address of 
    //the exported function and store it in our stub

    ULONG * l_pdwSlot = pdwGetSlotInExportTable(
        (void*)p_rInt.m_dwModuleBase, 
        (LPCSTR)p_rStub.m_dwFunctionNameOffset,
        p_pNTHeader);

    if (!l_pdwSlot)
    {
        //May be the config file simply misspelled the name.
        //In this case it is not a big error.
        //We will report it as an error to warn the User.
        MYTRACE(TF_Error,
            "ERROR : cannot find exported %s in %s",
            p_rStub.m_dwFunctionNameOffset,
            p_rInt.m_dwOffsetName);
        if(p_bUnhook && p_rStub.dwGetOriginalFunctionAddress())
        {
            //We cannot remove the hook. It is a serions error,
            //which should prevent us from terminating
            return -1;
        }
        p_rStub.vSetOriginalFunctionAddress(0);
        return 0;
    }//if (!l_pdwSlot)

    ULONG l_dwAddrOurStub = (ULONG)(&p_rStub.m_Code);
    ULONG l_dwOurStubRelativeToModule = 
        l_dwAddrOurStub - p_rInt.m_dwModuleBase;
    if(p_bUnhook)
    {
        //If the slot in the export table still points
        //to our stub, replace it by the original address
        ULONG l_dwOrigFunctionRelativeToModule = 
            p_rStub.dwGetOriginalFunctionAddress() -
            p_rInt.m_dwModuleBase;

        ULONG l_dwPrevValue = (ULONG)
            InterlockedCompareExchange((void**)l_pdwSlot,
                (void*)l_dwOrigFunctionRelativeToModule,
                (void*)l_dwOurStubRelativeToModule);
        if(l_dwPrevValue == l_dwOurStubRelativeToModule)
        {
            //it is the expected situation. We just restored
            //the pointer.
            MYTRACE(TF_Intercepted, 
                "Export %s restored in %s",
                p_rStub.m_dwFunctionNameOffset,
                p_rInt.pszGetName());

            return 1;
        }
        if(l_dwPrevValue == l_dwOrigFunctionRelativeToModule)
        {
            //Looks like it already points to the original address.
            //May be the module was reloaded.
            return 0;
        }
        MYTRACE(TF_Error, 
            "ERROR: export %s in %s points to %x instead of "
            "our stub at %x",
            p_rStub.m_dwFunctionNameOffset,
            p_rInt.m_dwOffsetName,
            l_dwPrevValue + p_rInt.m_dwModuleBase,
            l_dwAddrOurStub);
        //It may be a serious problem. May be somebody else
        //plays the same trick of interception.
        //It is better to prevent us from terminating
        return -1;
    }//if(p_bUnhook)
    ULONG l_dwExportedAddress = *l_pdwSlot + p_rInt.m_dwModuleBase;
    //We assume that we have no more than one 
    //stub per an exported function.
    //Therefore, compare the function against our
    //stub to determine whether we already intercepted it
    if(l_dwExportedAddress != l_dwAddrOurStub)
    {
        p_rStub.vSetOriginalFunctionAddress(l_dwExportedAddress);

        //Intercept exported function
        InterlockedExchange((long*)l_pdwSlot, 
            l_dwOurStubRelativeToModule);

        MYTRACE(TF_Intercepted, 
            "Export %s intercepted in %s",
            p_rStub.m_dwFunctionNameOffset,
            p_rInt.pszGetName());

        return 1;
    }
    return 0;
}//int Interceptor::iProcessExportForOneModuleAndOneStub

/*-----------------------------------------------------------------

   FUNCTION: Interceptor::bProcessExportForOneModul

   PURPOSE:  For all stubs in the given module update
    original function address and intercept the exported function

   RETURN VALUE: 
     false in case of a serious error
_________________________________________________________________*/
bool Interceptor::bProcessExportsForOneModule
(
    MyInterface         &   p_rInt,
    bool                    p_bUnhook
)
{
    PVOID l_pNTHeader = NULL;
    bool  l_bResult = true;
    if(p_rInt.m_dwModuleBase)
    {
        l_pNTHeader = 
            pCheckModuleHeader((void *)p_rInt.m_dwModuleBase);
        if(!l_pNTHeader)
        {
            MYTRACE(TF_Error, 
                "ERROR: cannot get NT header on module %x %s",
                p_rInt.m_dwModuleBase, 
                (LPCSTR)p_rInt.m_dwOffsetName);
        }
    }

    //Module loaded or unloaded. It is time to refresh stubs
    ASM_APIFunctionStub * l_pStubs = 
        (ASM_APIFunctionStub *)p_rInt.m_dwOffsetFunctions;

    int l_iNumFunctionsIntercepted = 0;
    for(int j = 0; j < p_rInt.m_iNumFunctions; j++)
    {
        int l_iResult = iProcessExportForOneModuleAndOneStub(
            l_pNTHeader, p_rInt, l_pStubs[j], p_bUnhook);
        if(l_iResult < 0)
            l_bResult = false;
        else
            l_iNumFunctionsIntercepted += l_iResult;
    }//for(int j = 0; j < l_rInt.m_iNumFunctions; j++)

    if(l_iNumFunctionsIntercepted > 0)
    {
        MYTRACE(TF_ApiHooked, 
        "%d exported functions %s for module %x %s",
            l_iNumFunctionsIntercepted,
            p_bUnhook? "restored" : "hooked",
            p_rInt.m_dwModuleBase, 
            p_rInt.pszGetName());
    }
    return l_bResult;
}//bool Interceptor::bProcessExportsForOneModule

/*-----------------------------------------------------------------

   FUNCTION: Interceptor::bHookAllModules

   PURPOSE:  Enumerate all modules in the process and intercept
    all their imported functions as instructed by the config file
      
   RETURN VALUE:
      bool       true on success, false on failure

-----------------------------------------------------------------*/
bool Interceptor::bHookAllModulesImpl
(
    EnumModules & p_rEnum,
    bool  p_bRestore
)
{
    __try
    {
        const EnumModules::DriverInfo * l_pDriverInfo = 
            p_rEnum.pGetFirst();
        while(l_pDriverInfo)
        {
            MYTRACE(TF_ListAllModules, 
                "%x %s",
                l_pDriverInfo->m_ulAddress, 
                l_pDriverInfo->m_szName); 

            int l_iNumFuns = 
                iInterceptImportedFunctionsInModule(
                (void*)l_pDriverInfo->m_ulAddress,
                l_pDriverInfo->m_szName,
                p_bRestore, true);
            if(l_iNumFuns > 0)
            {
                MYTRACE(TF_ApiHooked, 
                "%d Imported functions %s for module %x %s",
                    l_iNumFuns,
                    p_bRestore? "restored" : "hooked",
                    l_pDriverInfo->m_ulAddress, 
                    l_pDriverInfo->m_szName);
            }
            l_pDriverInfo = p_rEnum.pGetNext();
        }//while(l_pDriverInfo)
        return true;
    }
    __except(1)
    {
        MYTRACE(TF_Error, "Exception in bHookAllModulesImpl");
        return false;
    }
}//bool Interceptor::bHookAllModulesImpl

/*-----------------------------------------------------------------

   FUNCTION: Interceptor::bHookAllModules

   PURPOSE:  Enumerate all modules in the process and intercept
    all their imported functions as instructed by the config file
      
   RETURN VALUE:
      bool       true on success, false on failure

-----------------------------------------------------------------*/
bool Interceptor::bHookAllModules
(
    bool  p_bRestore
)
{
    EnumModules l_Enum;    
    return bHookAllModulesImpl(l_Enum, p_bRestore);
}//bool Interceptor::bHookAllModules(void)
