/******************************************************************
Module : KrnlTrc.cpp      - main module of KrnlTrc.exe. 
                            Handles initialization and a UI.
             Written 1998 by Dmitri Leman
Purpose: NT kernel mode tracer.
******************************************************************/
#include "Headers.h"
#include "Trace.h"
#include "KrnlTrc.h"
#include <winioctl.h>
#include "InstDrv.h"
//#include <windows.h>
//#include <stddef.h>
#include "Driver.h"
#include "ParseCfg.h"

static const char s_szINISection[] = "Options";
static const char s_szINICmdLineKey[] = "CommandLine";
static const char s_szINIConfigFileKey[] = "ConfigFile";

static const char s_szCaption[] = "KrnlTrc";

/*-----------------------------------------------------------------

   FUNCTION: bParseConfig

   PURPOSE: Load and parse the config file, calculate it's size,
      create allocate memory and export the config to this memory
      
   PARAMETERS:                   .
    KrnlTrcParams * p_pTempParams,            temporary parameters
                                              will be copied to
                                              the allocated memory
    KrnlTrcParams ** p_ppReturnParams,        return pointer to
                                              the allocated memory
                                  
   RETURN VALUE:
      BOOL  TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL bParseConfig
(
    KrnlTrcParams * p_pTempParams,
    KrnlTrcParams ** p_ppReturnParams
)
{
    Parser l_Parser;
    int l_iCfgSize = -1;
    if(!l_Parser.bLoadConfigFile
        (p_pTempParams->m_szConfigFileName))
    {
        MYTRACE(TF_MessageBox, "ERROR in parsing config file %s", 
            p_pTempParams->m_szConfigFileName);
        return FALSE;
    }
    l_iCfgSize = l_Parser.iGetExportedSize();
    if(l_iCfgSize < 0)
    {
        MYTRACE(TF_MessageBox, 
            "ERROR: cannot get size of config data");
        return FALSE;
    }
    int l_iSizeMemory = sizeof(KrnlTrcParams) + l_iCfgSize;

    KrnlTrcParams * l_pParams = 
        (KrnlTrcParams *)malloc(l_iSizeMemory);
    if(!l_pParams)
    {
        MYTRACE(TF_MessageBox, "malloc(%d) failed. %s", 
            l_iSizeMemory, ERR_EXPL(GetLastError()));
        return FALSE;
    }
    memset(l_pParams,0,l_iSizeMemory);

    memcpy(l_pParams, p_pTempParams, sizeof(KrnlTrcParams));

    l_pParams->m_dwParamSize = l_iSizeMemory;
    l_pParams->m_bPerformInterception = true;

    PBYTE l_pCfgData = (PBYTE)(l_pParams+1);

    l_iCfgSize = l_Parser.iExport(l_pCfgData, l_iCfgSize, 0);
    if(l_iCfgSize < 0)
    {
        MYTRACE(TF_MessageBox,"ERROR in exporting config file %s",
            p_pTempParams->m_szConfigFileName);

        free(l_pParams);
        return FALSE;
    }
    l_pParams->m_dwOffsetCfgData = sizeof(KrnlTrcParams);

    if(p_ppReturnParams)
      *p_ppReturnParams = l_pParams;
    else
        free(l_pParams);
    return TRUE;
}//BOOL bParseConfig

/*-----------------------------------------------------------------

   FUNCTION: bWriteConfigToFile

   PURPOSE:  Write the configuration file translated to a binary
   form to a file. This file will be loaded and used by our driver.
      
   PARAMETERS:                   .
                                  
   KrnlTrcParams * p_pParams      The configuration data

   RETURN VALUE:
      BOOL          TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL bWriteConfigToFile
(
    KrnlTrcParams * p_pParams
)
{
    HANDLE l_hFile = CreateFile(
        p_pParams->m_szConfigOutFileName, GENERIC_WRITE, 0,
        NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if(l_hFile == INVALID_HANDLE_VALUE)
    {
        MYTRACE(TF_MessageBox,  
            "ERROR: CreateFile(%s, GENERIC_WRITE) failed %s",
            p_pParams->m_szConfigOutFileName,
            ERR_EXPL(GetLastError()));
        return FALSE;
    }
    DWORD l_dwBytesWritten = 0;
    if(!WriteFile(l_hFile, p_pParams, 
        p_pParams->m_dwParamSize, &l_dwBytesWritten, NULL))
    {
        MYTRACE(TF_MessageBox,  "ERROR: WriteFile failed %s",
            ERR_EXPL(GetLastError()));
        CloseHandle(l_hFile);
        return FALSE;
    }
    if(l_dwBytesWritten != p_pParams->m_dwParamSize)
    {
        MYTRACE(TF_MessageBox,  
            "ERROR: WriteFile written %d instead of %d",
            l_dwBytesWritten != p_pParams->m_dwParamSize);
        CloseHandle(l_hFile);
        return FALSE;
    }
    CloseHandle(l_hFile);
    return TRUE;
}//BOOL bWriteConfigToFile


/*-----------------------------------------------------------------

   FUNCTION: bWriteDriverParameters

   PURPOSE:  Write the parameters (such as the name of the binary
    config file) to the driver's entry in the registry
      
   PARAMETERS:                   .
                                  
   KrnlTrcParams * p_pParams      Config structure

   RETURN VALUE:
      BOOL TRUE on success, FALSE on failure

-----------------------------------------------------------------*/
BOOL bWriteDriverParameters
(
    KrnlTrcParams * p_pParams
)
{
    char l_szSubkey[256];
    wsprintf(l_szSubkey, 
        "System\\CurrentControlSet\\Services\\%s\\Parameters",
        p_pParams->m_szDriverName);
    const char l_szValue[] = TRACER_REG_CONFIG_PATH_A;

    HKEY l_hKey = NULL;
    DWORD l_dwDisposition = 0;
    long l_Result = RegCreateKeyEx(HKEY_LOCAL_MACHINE, l_szSubkey, 
        0, "", REG_OPTION_NON_VOLATILE,
        KEY_SET_VALUE, NULL, &l_hKey, &l_dwDisposition);
    if(l_Result != ERROR_SUCCESS)
    {
        MYTRACE(TF_MessageBox, 
            "ERROR: RegCreateKeyEx(%s) failed. %s",
            l_szSubkey, ERR_EXPL(l_Result));
        return FALSE;
    }

    l_Result = RegSetValueEx(l_hKey, l_szValue, 0, REG_SZ, 
        (CONST BYTE *)p_pParams->m_szConfigOutFileNameDriver,
        strlen(p_pParams->m_szConfigOutFileNameDriver)+1);
    if(l_Result != ERROR_SUCCESS)
    {
        MYTRACE(TF_MessageBox, 
            "ERROR: RegSetValueEx %s %s failed. %s",
            l_szSubkey, l_szValue, ERR_EXPL(l_Result));
        RegCloseKey(l_hKey);
        return FALSE;
    }

    RegCloseKey(l_hKey);
    return TRUE;
}//BOOL bWriteDriverParameters


/*-----------------------------------------------------------------

   FUNCTION: bDoTrace

   PURPOSE:  Start the driver and show a message box. When the
    message box is dismissed, stop the driver.
      
   PARAMETERS:                   .
                                  
   const KrnlTrcParams * p_pParams - config structure

   RETURN VALUE:
      BOOL TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL bDoTrace
(
    const KrnlTrcParams * p_pParams
)
{
    BOOL l_bResult = FALSE;
    TCHAR l_szDriverName[] = _T("KrnlTrc1");
    TCHAR l_szDeviceName[] = TRACER_WIN32_FILE_NAME_A;

    DeleteFile(p_pParams->m_szLogFileName);

    HANDLE l_hDevice = NULL;
    if(!InstallDriver::bStartDriverAndOpenDevice(
        p_pParams->m_szDriverName,
        p_pParams->m_szDeviceName, &l_hDevice))
    {
        MYTRACE(TF_MessageBox, "ERROR : StartDeviceDriver failed");
        return FALSE;
    }

    DWORD l_dwReturnValue = 0;

    DWORD l_dwBytesReturned = 0;
    if(!DeviceIoControl(l_hDevice, IOCTL_TRACER_CHECK_ACTIVE, 
        NULL, 0, &l_dwReturnValue, sizeof(l_dwReturnValue), 
        &l_dwBytesReturned, NULL))
    {
        MYTRACE(TF_Error, 
            _T("ERROR: DeviceIoControl(IOCTL_TRACER_CHECK_ACTIVE)"
            " failed - %s"), 
            ERR_EXPL(GetLastError()));
    }
    if(l_dwBytesReturned != sizeof(DWORD))
    {
        MYTRACE(TF_Error, 
            _T("ERROR: DeviceIoControl(IOCTL_TRACER_CHECK_ACTIVE)"
            "returned %d bytes instead of %d"), 
            l_dwBytesReturned, sizeof(DWORD));
    }
    if(l_dwReturnValue != IOCTL_TRACER_CHECK_ACTIVE)
    {
        MYTRACE(TF_Error, 
            _T("ERROR: DeviceIoControl(IOCTL_TRACER_CHECK_ACTIVE) "
            "returned value %d instead of %d"),
            l_dwReturnValue, IOCTL_TRACER_CHECK_ACTIVE);
    }
    MessageBox(0, _T("Working"), s_szCaption, 0);

    CloseHandle(l_hDevice);

    bool l_bUnregister = false;
    if(IDYES == MessageBox(0, 
        _T("Do you want to unregister the driver?"), _T(""), 
        MB_YESNO))
    {
        l_bUnregister = true;
    }

    InstallDriver::bUnloadDeviceDriver(
        p_pParams->m_szDriverName, l_bUnregister );

    if(GetFileAttributes(p_pParams->m_szLogFileName) != (DWORD)-1)
    {
        ShellExecute(NULL, "open", "Notepad.exe", 
            p_pParams->m_szLogFileName, NULL, SW_SHOW);
    }
    return TRUE;
}//BOOL bDoTrace

/*----------------------------------------------------------------

   FUNCTION: bProgramBody

   PURPOSE: Called from WinMain.
      
   PARAMETERS:                   .
                                  
   HINSTANCE p_hInstance        handle to current instance
   HINSTANCE p_hPrevInstance    not used
   LPSTR  p_lpszCmdLine         pointer to command line
   int    p_nCmdShow            show state of window

   RETURN VALUE:
      BOOL  TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL bProgramBody
( 
    HINSTANCE p_hInstance, 
    HINSTANCE p_hPrevInstance,
    LPSTR  p_lpszCmdLine, 
    int    p_nCmdShow 
)
{
    BOOL l_bResult = FALSE;
    //Prepare path names for INI file, driver, etc.
    char l_szCurExePath[_MAX_PATH];
    char l_szCurExeName[_MAX_PATH];
    char l_szINIFile   [_MAX_PATH];

    GetModuleFileName(p_hInstance, 
        l_szCurExePath, sizeof(l_szCurExePath));
    char * l_pExt = strrchr(l_szCurExePath, '.');
    char * l_pName = strrchr(l_szCurExePath, '\\');
    if(!l_pExt || !l_pName) 
    {
        MYTRACE(TF_MessageBox, 
         "Error : cannot find an extension or name in the path %s",
            l_szCurExePath);
        return FALSE;
    }
    l_pName++;
    int l_iExt = l_pExt - l_szCurExePath;
    int l_iName = l_pName - l_szCurExePath;
    int l_iNameLen = l_iExt - l_iName;
    memcpy(l_szCurExeName, l_szCurExePath + l_iName, l_iNameLen);
    l_szCurExeName[l_iNameLen] = 0;

    memcpy(l_szINIFile, l_szCurExeName, l_iNameLen);
    strcpy(l_szINIFile + l_iNameLen, ".ini");

    //We accept only one (optional) parameter - name of INI file
    strupr(p_lpszCmdLine);
    char * l_pszParam = strstr(p_lpszCmdLine, "-INI=");
    if(l_pszParam)
    {
        l_pszParam += 5;
        int l_iLen = strlen(l_pszParam);
        if(*l_pszParam == '"')
        {
            char * p = strchr(l_pszParam+1, '"');
            if(p)
            {
                l_pszParam++;
                l_iLen = p - l_pszParam;
            }
        }
        memcpy(l_szINIFile, l_pszParam, l_iLen);
        l_szINIFile[l_iLen] = 0;
    }//if(l_pszParam)
    char l_szFullINIPath   [_MAX_PATH];
    GetFullPathName(l_szINIFile, 
        sizeof(l_szFullINIPath), l_szFullINIPath, NULL);


    //Enable trace
    DWORD l_dwOutputDebugMask = 
        GetPrivateProfileInt(s_szINISection, "OutputDebugMask", 
        TF_Error, l_szFullINIPath);
   
    DWORD l_dwMessageBoxMask = 
        GetPrivateProfileInt(s_szINISection, "MessageBoxMask",
        TF_MessageBox, l_szFullINIPath);

    DWORD l_dwTraceBufferMask = 
        GetPrivateProfileInt(s_szINISection, "TraceBufferMask",
        TF_MainLog, l_szFullINIPath);
    int l_iTraceBufferSize = 
        GetPrivateProfileInt(s_szINISection, "TraceBufferSize",
        1024*1024, l_szFullINIPath);

    l_dwTraceBufferMask |= TF_Error;   //always trace  errors
    l_dwOutputDebugMask |= TF_Error;   //always trace  errors
    
    vSetCurTraceMasks(TT_DebugMonitor,
        l_dwOutputDebugMask, (DWORD)-1);
    
    l_dwMessageBoxMask |= TF_MessageBox;
    vSetCurTraceMasks(l_dwMessageBoxMask,
        l_dwMessageBoxMask, (DWORD)-1);


    KrnlTrcParams l_TempParams;
    KrnlTrcParams * l_pParams = &l_TempParams;

    GetPrivateProfileString(s_szINISection, "DriverName", 
        "KrnlTrc.sys", 
        l_pParams->m_szDriverName, sizeof(l_pParams->m_szDriverName),
        l_szFullINIPath);

    char l_szDriverPath[_MAX_PATH];
    GetPrivateProfileString(s_szINISection, "DriverPath", 
        "KrnlTrc.sys", l_szDriverPath, sizeof(l_szDriverPath),
        l_szFullINIPath);
    GetFullPathName(l_szDriverPath, 
        sizeof(l_pParams->m_szDriverPath), 
        l_pParams->m_szDriverPath, NULL);

    strcpy(l_pParams->m_szDeviceName, TRACER_WIN32_FILE_NAME_A);

    
    char l_szTempConfigPath[_MAX_PATH];
    memcpy(l_szTempConfigPath, l_szCurExeName, l_iNameLen);
    strcpy(l_szTempConfigPath + l_iNameLen, ".cfg");
    
    GetPrivateProfileString(s_szINISection, "ConfigFile", 
        l_szTempConfigPath, l_szTempConfigPath, 
        sizeof(l_szTempConfigPath), l_szFullINIPath);
    GetFullPathName(l_szTempConfigPath,
        sizeof(l_pParams->m_szConfigFileName), 
        l_pParams->m_szConfigFileName, NULL);

    char l_szTempConfigOutPath[_MAX_PATH];
    memcpy(l_szTempConfigOutPath, l_szCurExeName, l_iNameLen);
    strcpy(l_szTempConfigOutPath + l_iNameLen, ".cfo");
    
    GetPrivateProfileString(s_szINISection, "ConfigOutFile", 
        l_szTempConfigOutPath, l_szTempConfigOutPath, 
        sizeof(l_szTempConfigOutPath), l_szFullINIPath);
    GetFullPathName(l_szTempConfigOutPath,
        sizeof(l_pParams->m_szConfigOutFileName), 
        l_pParams->m_szConfigOutFileName, NULL);

    GetPrivateProfileString(s_szINISection, "ConfigOutFileDriver", 
        "", l_szTempConfigOutPath, sizeof(l_szTempConfigOutPath), 
        l_szFullINIPath);
    if(*l_szTempConfigOutPath)
    {
        //Take what the User entered as is
        strcpy(l_pParams->m_szConfigOutFileNameDriver,
            l_szTempConfigOutPath);
    }
    else
    {
        //Try to prepend "\??\" to the name
        memcpy(l_pParams->m_szConfigOutFileNameDriver, 
            "\\??\\",4);
        strcpy(l_pParams->m_szConfigOutFileNameDriver+4, 
            l_pParams->m_szConfigOutFileName);
    }
    
    char l_szTempLogFileName[_MAX_PATH];
    memcpy(l_szTempLogFileName, l_szCurExeName, l_iNameLen);
    strcpy(l_szTempLogFileName + l_iNameLen, ".log");
    GetPrivateProfileString(s_szINISection, "DllLogFileName", 
        l_szTempLogFileName, 
        l_szTempLogFileName, 
        sizeof(l_szTempLogFileName), l_szFullINIPath);
    GetFullPathName(l_szTempLogFileName, 
        sizeof(l_pParams->m_szLogFileName), 
        l_pParams->m_szLogFileName, NULL);

    GetWindowsDirectory(l_pParams->m_szWindowsDir, 
        sizeof(l_pParams->m_szWindowsDir));
    l_pParams->m_iLengthWindowsDir = 
        strlen(l_pParams->m_szWindowsDir);
    GetSystemDirectory(l_pParams->m_szDriversDir, 
        sizeof(l_pParams->m_szDriversDir));
    strcat(l_pParams->m_szDriversDir, "\\Drivers");
    l_pParams->m_iLengthDriversDir = 
        strlen(l_pParams->m_szDriversDir);

    l_pParams->m_dwOutputDebugStringMask = l_dwOutputDebugMask;
    l_pParams->m_dwTraceBufferMask = l_dwTraceBufferMask;
    l_pParams->m_iTraceBufferSize = l_iTraceBufferSize;

    l_pParams->m_bPrintThreadId = 
        GetPrivateProfileInt(s_szINISection, "PrintThreadID",
        true, l_szFullINIPath)? true : false;
    l_pParams->m_bPrintProcessorNumber = 
        GetPrivateProfileInt(s_szINISection, "PrintProcessorNumber",
        false, l_szFullINIPath)? true : false;
    l_pParams->m_bPrintIRQL = 
        GetPrivateProfileInt(s_szINISection, "PrintIRQL",
        false, l_szFullINIPath)? true : false;

    l_pParams->m_iLimitStringOutput = 
        GetPrivateProfileInt(s_szINISection, "MaxStrLen",
            30, l_szFullINIPath);
    l_pParams->m_dwOffsetCfgData = 0;

    l_pParams->m_bPrintParams = 
        GetPrivateProfileInt(s_szINISection, "PrintParams",
        true, l_szFullINIPath)? true : false;
    l_pParams->m_bInterceptReturns = 
        GetPrivateProfileInt(s_szINISection, "InterceptReturns",
        true, l_szFullINIPath)? true : false;
    l_pParams->m_iNumReturnStubs = 
        GetPrivateProfileInt(s_szINISection, "NumReturnStubs",
            30, l_szFullINIPath);
    l_pParams->m_bDontFreeStubMemory = 
        GetPrivateProfileInt(s_szINISection, "DontFreeStubMemory",
            false, l_szFullINIPath)? true : false;

    l_bResult = TRUE;
    if(!bParseConfig(l_pParams, &l_pParams))
        l_bResult = FALSE;
        
    if(l_bResult && !bWriteConfigToFile(l_pParams))
        l_bResult = FALSE;

    if(l_bResult && !InstallDriver::bRegisterDeviceDriver(
        l_pParams->m_szDriverName, l_pParams->m_szDriverPath))
    {
        MYTRACE(TF_MessageBox,
            "RegisterDeviceDriver(%s,%s) failed",
            l_pParams->m_szDriverName, 
            l_pParams->m_szDriverPath);
        l_bResult = FALSE;
    }
    if(l_bResult && !bWriteDriverParameters(l_pParams))
        l_bResult = FALSE;
    if(l_bResult)
    {
        char l_szMessage[256];
        wsprintf(l_szMessage, 
            "Config file %s was parsed and stored in %s\n"
            "Driver %s at %s is registered\n"
            "Do you want to start the tracer now?",
            l_pParams->m_szConfigFileName,
            l_pParams->m_szConfigOutFileName,
            l_pParams->m_szDriverName, 
            l_pParams->m_szDriverPath
            );
        if(IDYES == MessageBox
            (0, l_szMessage, s_szCaption, MB_YESNO))
        {
            if(!bDoTrace(l_pParams))
                l_bResult = FALSE;
        }
    }
    if(l_pParams && l_pParams != &l_TempParams)
        free(l_pParams);
    return l_bResult;
}//BOOL bProgramBody

/*----------------------------------------------------------------

   FUNCTION: WinMain

   PURPOSE: Initial entry point for the application. See SDK help.
      
   PARAMETERS:                   .
                                  
   HINSTANCE p_hInstance            handle to current instance
   HINSTANCE p_hPrevInstance        not used
   LPSTR  p_lpszCmdLine          pointer to command line
   int    p_nCmdShow             show state of window

   RETURN VALUE:
      int                        Exit value
-----------------------------------------------------------------*/
int APIENTRY WinMain
( 
    HINSTANCE p_hInstance, 
    HINSTANCE p_hPrevInstance,
    LPSTR     p_lpszCmdLine, 
    int       p_nCmdShow 
)
{
    vTraceInitialize();
    BOOL l_bResult = bProgramBody(
        p_hInstance, p_hPrevInstance, p_lpszCmdLine, p_nCmdShow);
    vTraceUnInitialize();
    return l_bResult? 0 : 1;
}//int APIENTRY WinMain

