/******************************************************************
Module : KrnlTrcD.cpp     - initialization / deinitialization of
             the global structures of the interceptor driver
             Written 1998,1999 by Dmitri Leman
Purpose: NT kernel mode tracer.
******************************************************************/
#include "Headers.h"
#include "Driver.h"
#include "log.h"
#include "return.h"
#include "KrnlTrc.h"
#include "KrnlTrcD.h"
#include "StubMgr.h"
#include "Trace.h"
#include "PEHelper.h"
#include "Intrcpt.h"
#include "EnumMods.h"
#include "FileSupp.h"

const char s_szGetProcAddressName[] = "GetProcAddress";
SpyGlobalData g_Data = 
    {false, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
ReturnMgr s_RetMgr;

/*-----------------------------------------------------------------

   FUNCTION: SpyGlobalData::KrnlTrcDriverInit

   PURPOSE: Called from DriverEntry

   PARAMETERS:
     IN  PUNICODE_STRING p_pRegistryPath - path to the registry 
        location for this driver
      
   RETURN VALUE:
      NTSTATUS   0 on success, error code otherwise
-----------------------------------------------------------------*/
NTSTATUS SpyGlobalData::KrnlTrcDriverInit
(
    IN  PUNICODE_STRING p_pRegistryPath
)
{
    if(m_bSpyInitialized)
        return STATUS_SUCCESS;

    UNICODE_STRING  l_ConfigFilePath;
    l_ConfigFilePath.Length = l_ConfigFilePath.MaximumLength = 0;
    l_ConfigFilePath.Buffer = NULL;

    NTSTATUS l_Status = ReadRegistry
        (p_pRegistryPath, &l_ConfigFilePath);

    if(    !NT_SUCCESS( l_Status ))
    {
        RtlFreeUnicodeString(&l_ConfigFilePath);
        return l_Status;
    }

    //Open and read the binary configuration file
    MyFileLocal l_File;

    l_Status = l_File.OpenFile(NULL, false, false, 
        &l_ConfigFilePath);
    RtlFreeUnicodeString(&l_ConfigFilePath);

    if(!NT_SUCCESS(l_Status))
        return l_Status;

    LARGE_INTEGER  l_FileSize;
    l_Status = l_File.GetFileSize(&l_FileSize);
    if(!NT_SUCCESS(l_Status))
    {
        l_File.CloseFile();
        return l_Status;
    }
    if(l_FileSize.HighPart || 
       l_FileSize.LowPart < sizeof(KrnlTrcParams))
    {
        MYTRACE(TF_Error, "ERROR: file size (%x%x) too strange",
            l_FileSize.HighPart, l_FileSize.LowPart);
        l_File.CloseFile();
        return STATUS_INVALID_PARAMETER;
    }
    
    m_pParams = (KrnlTrcParams *)ExAllocatePool(
        NonPagedPool, l_FileSize.LowPart);
    if(!m_pParams)
    {
        MYTRACE(TF_Error, "ERROR: ExAllocatePool(%d) failed",
            l_FileSize.LowPart);
        l_File.CloseFile();
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    l_Status =l_File.ReadWriteFile(false, m_pParams, 
        l_FileSize.LowPart, NULL, NULL);
    l_File.CloseFile();

    if(!NT_SUCCESS(l_Status))
    {
        ExFreePool(m_pParams);
        m_pParams = NULL;
        return l_Status;
    }

    //Initialize trace

    vSetCurTraceMasks(TT_DebugMonitor,
        m_pParams->m_dwOutputDebugStringMask, (ULONG)-1);
    
    vSetCurTraceMasks(TT_MessageBox, 0, (ULONG)-1);
    //Don't use message box in the hook

    bAllocateTraceBuffer(m_pParams->m_iTraceBufferSize,
        m_pParams->m_dwTraceBufferMask, (ULONG)-1);
    
    MYTRACE(TF_SpyInit, "KrnlTrcDriverInit %s %s %x %x %d",
      m_pParams->m_szConfigFileName,
      m_pParams->m_szLogFileName,
      m_pParams->m_dwOutputDebugStringMask,
      m_pParams->m_dwTraceBufferMask,
      m_pParams->m_iTraceBufferSize);

    int l_iSizeMap = sizeof(KrnlTrcParams);

    l_Status = s_RetMgr.Init(m_pParams->m_iNumReturnStubs);
    if(!NT_SUCCESS(l_Status))
    {
        //We don't need to free previously allocated resource 
        //here, because ShutDownKrnlTrcDriver will be called
        return l_Status;
    }

    if(!m_pParams->m_bPerformInterception)
        return STATUS_SUCCESS;//this is a test call. 
                              //Don't do anything

    if(!m_pParams->m_dwOffsetCfgData)
    {
        MYTRACE(TF_Error, "ERROR : wrong offset of config data");
        return STATUS_INVALID_PARAMETER;
    }

    m_pConfigMgr = (StubManager * )
        (((PUCHAR)m_pParams) + m_pParams->m_dwOffsetCfgData);
    m_pConfigMgr->vDriverInit
        ((PUCHAR)m_pConfigMgr, Logger::IntrerceptEntryNaked);

    m_pRetMgr = &s_RetMgr;

    m_iLimitStringOutput = m_pParams->m_iLimitStringOutput;
    m_bPrintThreadId = m_pParams->m_bPrintThreadId;
    m_bPrintProcessorNumber = m_pParams->m_bPrintProcessorNumber;
    m_bPrintIRQL = m_pParams->m_bPrintIRQL;

    g_Data.m_pConfigMgr->bProcessAllModulesExports(false);
    Interceptor::bHookAllModules(false);

    m_bSpyInitialized = true;
    return STATUS_SUCCESS;
}//NTSTATUS SpyGlobalData::KrnlTrcDriverInit

/*-----------------------------------------------------------------

   FUNCTION: SpyGlobalData::ReadRegistry

   PURPOSE: Read driver's parameters from the registry

   PARAMETERS:
     IN  PUNICODE_STRING p_pRegistryPath - path to the registry 
        location for this driver
    OUT PUNICODE_STRING p_pConfigFilePath - return path to
        the configuration file, obtained from the registry
      
   RETURN VALUE:
      NTSTATUS   0 on success, error code otherwise
-----------------------------------------------------------------*/
NTSTATUS SpyGlobalData::ReadRegistry
(
    IN  PUNICODE_STRING p_pRegistryPath,
    OUT PUNICODE_STRING p_pConfigFilePath
)
{
    //Construct a 0-terminated registry path with the subkey
    static const WCHAR l_wszSubkey[] = L"\\Parameters";
    PWCHAR p_pwszFullPath;

    int l_iFullPathSize = p_pRegistryPath->Length + 
        sizeof(l_wszSubkey) + sizeof( UNICODE_NULL );
                
    if(NULL == (p_pwszFullPath = (PWCHAR)
        ExAllocatePool(PagedPool, l_iFullPathSize)))
    {
        MYTRACE(TF_Error, "ERROR: ExAllocatePool(%d) failed",
            l_iFullPathSize);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    RtlMoveMemory( p_pwszFullPath, p_pRegistryPath->Buffer, 
                    p_pRegistryPath->Length);
    RtlMoveMemory( p_pwszFullPath + p_pRegistryPath->Length/2, 
        l_wszSubkey, sizeof(l_wszSubkey));
    p_pwszFullPath[ l_iFullPathSize/2-1 ] = UNICODE_NULL;
    

    RTL_QUERY_REGISTRY_TABLE l_Table[10];
    
    RtlZeroMemory( l_Table, sizeof(l_Table));

    l_Table[0].Name    = TRACER_REG_CONFIG_PATH_W;
    l_Table[0].Flags= RTL_QUERY_REGISTRY_DIRECT | 
                      RTL_QUERY_REGISTRY_REQUIRED;
    l_Table[0].EntryContext = p_pConfigFilePath;

    NTSTATUS l_Status = RtlQueryRegistryValues(
                RTL_REGISTRY_ABSOLUTE, p_pwszFullPath,
                l_Table, NULL, NULL );
    ExFreePool( p_pwszFullPath );

    if(    !NT_SUCCESS( l_Status ))
    {
        MYTRACE(TF_Error, 
            "ERROR: RtlQueryRegistryValues failed %x", l_Status);
        return l_Status;
    }
    return l_Status;
}//NTSTATUS SpyGlobalData::ReadRegistry

/*-----------------------------------------------------------------

   FUNCTION: ShutDownKrnlTrcDriver

   PURPOSE:  Uninitialize the DLL
-----------------------------------------------------------------*/
void SpyGlobalData::ShutDownKrnlTrcDriver(void)
{
    MYTRACE(TF_SpyInit, "ShutDownKrnlTrcDriver");

    bool l_bFree = true;

    if(g_Data.m_pConfigMgr)
    {
        g_Data.m_pConfigMgr->vShortcutAllStubs();

        //call hook functions with parameter "true" to undo
        //the interception.
        if(!Interceptor::bHookAllModules(true))
            l_bFree = false;
        if(!g_Data.m_pConfigMgr->bProcessAllModulesExports(true))
            l_bFree = false;
        if(!l_bFree)
        {
            //Our driver will be unloaded, but we will not free
            //our memory with stubs. This should prevent crashes.
            MYTRACE(TF_Error, 
                "ERROR: bProcessAllModulesExports returned false. "
                "Termination may be unsafe. Restart the PC");
        }
        if(m_pParams->m_bDontFreeStubMemory)
            l_bFree = false;
    }//if(g_Data.m_pConfigMgr)
        
    if(m_pParams && m_pParams->m_bDontFreeStubMemory)
        l_bFree = false;

    if(!l_bFree)
    {
        MYTRACE(TF_Error, 
            "ERROR: Configuration data left in the non-paged memory");
    }

    Logger::PrintStatistics();
    if(m_pParams)
        bDumpTraceBufferToFile(m_pParams->m_szLogFileName);

    if(l_bFree)
    {
        s_RetMgr.Shutdown();

        m_pRetMgr = NULL;

        m_pConfigMgr = NULL;

        if(m_pParams)
        {
            #ifdef _DEBUG
            memset(m_pParams, 0xEE, m_pParams->m_dwParamSize);
            #endif  
            ExFreePool(m_pParams);
        }
        m_pParams = NULL;

        m_bSpyInitialized = false;
    }
}//void ShutDownKrnlTrcDriver
