/******************************************************************
Module : Log.cpp      - module which performs decoding intercepted
                    functions names and parameters and prints them
             Written 1998,1999 by Dmitri Leman
Purpose: Purpose: NT kernel mode tracer.
*****************************************************************/
#include "Headers.h"
#include "Driver.h"
#include "parmtype.h"
#include "return.h"
#include "RetStub.h"
#include "Trace.h"
#include "KrnlTrcD.h"
#include "StubMgr.h"
#include "Log.h"
#include "intrcpt.h"
#include "Stat.h"

#define _STATISTICS

#ifdef _STATISTICS
KeepStat g_StatEntry = {0,(ULONG)-1,0,0}, g_StatReturn = {0,(ULONG)-1,0,0};
#endif

/*-----------------------------------------------------------------
   FUNCTION: OneParamInfo::OneParamInfo

   PURPOSE:  Constructor - initialize information about parameter.
      
   PARAMETERS:                   .
    ParamType    p_ParamType    - type of parameters
    UCHAR        p_byParamFlags - flags describing the parameter
    UCHAR        p_byExtraByte  - second byte (for two-byte params)
        previous three parameters a taken from the encoded
        parameter block in the function stub

    ULONG    *    p_pdwAddrCurParameter - place where the parameter
        is located now (usually on the application's stack, but
        it may point in the data field of VARIANT structure if we
        want to process a VARIANT).
        This pointer may be NULL if we processing
        parameters during the function return
-----------------------------------------------------------------*/
OneParamInfo::OneParamInfo
(
    ParamType    p_ParamType,
    UCHAR        p_byParamFlags,
    UCHAR        p_byExtraByte,
    ULONG    *    p_pdwAddrCurParameter
)
{
    m_ParamType      = p_ParamType;
    m_byParamFlags   = p_byParamFlags;
    m_byExtraByte    = p_byExtraByte;
    m_pdwAddrCurParameter = p_pdwAddrCurParameter;
    m_dwCurParamValue = m_pdwAddrCurParameter?
        *m_pdwAddrCurParameter : 0;
    m_iParamSize = 0;
}//OneParamInfo::OneParamInfo

/*----------------------------------------------------------------
   FUNCTION: ParamProcContext::vInit

   PURPOSE:  Initializes the structure by
    information about function's parameters
      
   PARAMETERS:                   .

    const FunctionParameters * p_pParams - encoded params from stub
    ReturnStub               * p_pReturnStub, - pointer return stub
    ULONG                    * p_pdwCallStackFrame - pointer to the
                                   actual stack of the application
                                   which holds the parameters.
                                   May be NULL during processing
                                   of function return
    bool                       p_bOut    true if we processing
                                        function return. false -
                                        during function call.
-----------------------------------------------------------------*/
void ParamProcContext::vInit
(
    const FunctionParameters * p_pParams,
    ReturnStub               * p_pReturnStub,
    ULONG                    * p_pdwCallStackFrame,//may be NULL
    bool                       p_bOut    
)
{
    m_pParams = p_pParams;
    m_pReturnStub = p_pReturnStub;
    m_dwThisAddress = 0;
    m_pCallStackFrame = p_pdwCallStackFrame;
    m_bOut = p_bOut;
}//ParamProcContext::vInit

 /*----------------------------------------------------------------

   FUNCTION: ParamProcContext::bStoreRetreiveOutParameter

   PURPOSE: In order to process function's results returned through
    parameters (for example returned pointers to interfaces),
    we need to preserve stack entries which points to these 
    parameters because the original application stack frame will 
    be gone when the function returns. bStoreRetreiveOutParameter 
   function stores and extracts these pointers in our private stack
      
   PARAMETERS:                   .
                                  
   bool p_bStore             Store or retrieve
   ULONG & p_rdwParameter    if p_bStore == true, this parameter 
                             will be passed in, otherwise it will
                             be used to return the stored value.
   ReturnStub               * p_pReturnStub - our return stub
   int * p_piOutputParamIndex     - index of this parameter in 
                                the array inside our private stack

   RETURN VALUE:
      bool          true on success, false on failure
-----------------------------------------------------------------*/
bool ParamProcContext::bStoreRetreiveOutParameter
(
    bool          p_bStore,
    ULONG       & p_rdwParameter,
    ReturnStub  * p_pReturnStub,
    int         * p_piOutputParamIndex
)
{
    if(*p_piOutputParamIndex >= MAX_OUT_PARAMETERS)
    {
        MYTRACE(TF_Error, "Function has too many [OUT] "
            "parameters (%d). Maximum allowed is %d",
            *p_piOutputParamIndex, MAX_OUT_PARAMETERS);
        return false;
    }
    if(p_bStore)
    {
        //Store the pointer to the output data to the
        //storage on our private stack.
        //We will use this pointer to print the returned
        //data when the function will return
        p_pReturnStub->m_dwOutParameter[*p_piOutputParamIndex] = 
            p_rdwParameter;
    }
    else
    {
        //Now (when the function returned) we can
        //extract the pointer to the output data
        //from our private stack and can print the data
        p_rdwParameter = p_pReturnStub->
            m_dwOutParameter[*p_piOutputParamIndex];
    }
    (*p_piOutputParamIndex)++;
    return true;
}//bool ParamProcContext::bStoreRetreiveOutParameter

/*-----------------------------------------------------------------

   FUNCTION: ParamProcContext::iAddStringToOutputBuff

   PURPOSE:  Simple helper to copy a string to a buffer.
    Unlike strncpy copies only entire string if it fits.
    Returns number of bytes copied without terminating '\0'.
      
   PARAMETERS:                   .
                                  
   PSTR  p_pszOutputBuff - destination buffer
   int   p_iBuffSize     - size of the buffer
   PCSTR p_pszString     - input string

   RETURN VALUE:
      int       number of character placed in the output buffer
-----------------------------------------------------------------*/
int iAddStringToOutputBuff
(
    PSTR    p_pszOutputBuff,
    int        p_iBuffSize,
    PCSTR    p_pszString
)
{
    int l_iLen = strlen(p_pszString);
    
    if(l_iLen + 1 >= p_iBuffSize)
        return 0;
    memcpy(p_pszOutputBuff, p_pszString, l_iLen+1);
    return l_iLen;
}//int iAddStringToOutputBuff

int MyI2A
(
    PSTR       p_pszReturnPrintedParams,
    int        p_iParamsBuffSize,
    int        p_iPrintIntegerBytes,
    ULONG      p_ulValue
)
{
    static char s_HexadecimalDigits[17] = "0123456789ABCDEF";
    for(int i = p_iPrintIntegerBytes-1 ; i >= 0; i--)
    {
        p_pszReturnPrintedParams[2*i+1] = 
            s_HexadecimalDigits[(p_ulValue & 0xf)];
        p_ulValue >>= 4;
        p_pszReturnPrintedParams[2*i] = 
            s_HexadecimalDigits[(p_ulValue & 0xf)];
        p_ulValue >>= 4;
    }
    p_pszReturnPrintedParams[2*p_iPrintIntegerBytes] = 0;
    return p_iPrintIntegerBytes*2;
}//int MyI2A

/*-----------------------------------------------------------------

   FUNCTION: iPrintInteger

   PURPOSE:  Format integers of various length to an output buffer
      
   PARAMETERS:                   .
                                  
   PSTR p_pszReturnPrintedParams  Output buffer
   int  p_iParamsBuffSize         Size of the buffer
   int  p_iPrintIntegerBytes      Size of the integer
   ULONG * p_pdwAddrCurParam      Pointer to the integer
   ULONG   p_dwValue              Value of the integer

   RETURN VALUE:
      int       number of character placed in the output buffer 
-----------------------------------------------------------------*/
int iPrintInteger
(
    PSTR       p_pszReturnPrintedParams,
    int        p_iParamsBuffSize,
    int        p_iPrintIntegerBytes,
    ULONG *    p_pdwAddrCurParam,
    ULONG      p_dwValue
)
{
    if(!p_iPrintIntegerBytes)
        return 0;

    if(20 >= p_iParamsBuffSize)
        return 0;

    int l_iFilledBytes = 0;
    ULONG l_dwValue = 0;
    switch(p_iPrintIntegerBytes)
    {
    case 1: l_dwValue = (UCHAR)p_dwValue; break;
    case 2: l_dwValue = (USHORT)p_dwValue; break;
    case 4: l_dwValue = (ULONG)p_dwValue; break;
    case 8: 
        if(!p_pdwAddrCurParam)
            return 0;
        l_iFilledBytes = MyI2A(
            p_pszReturnPrintedParams, p_iParamsBuffSize, 4, 
            p_pdwAddrCurParam[1]);
        l_dwValue = p_pdwAddrCurParam[0];
        p_iPrintIntegerBytes = 4;
        break;
    default:
        return 0;
    }
    l_iFilledBytes += MyI2A(
        p_pszReturnPrintedParams + l_iFilledBytes, 
        p_iParamsBuffSize - l_iFilledBytes, p_iPrintIntegerBytes, 
        l_dwValue);
    return l_iFilledBytes;
}//int ParamProcContext::iPrintInteger

/*-----------------------------------------------------------------

   FUNCTION: ParamProcContext::iGetLPSTR

   PURPOSE:  Output string (or UNICODE string) to a buffer
      
   PARAMETERS:                   .
                                  
   PSTR  p_pStr                  Source string 
   int   p_iCharSize             Size of a character (1 or 2)
   int   p_iNumCharsToPrint      How many character to print 
                                 (may be -1 to print up to '\0')
   PSTR  p_pszOutputBuffer       Output buffer
   int   p_iSizeOutBuffer        Size of the buffer 

   RETURN VALUE:
      int       number of character placed in the output buffer
-----------------------------------------------------------------*/
int ParamProcContext::iGetLPSTR
( 
   PSTR  p_pStr,
   int   p_iCharSize,
   int   p_iNumCharsToPrint,
   PSTR  p_pszOutputBuffer,
   int   p_iSizeOutBuffer
)
{
    int i;
    bool l_bStopAtNull = false;
    if(p_iNumCharsToPrint < 0)
    {
        l_bStopAtNull = true;
        p_iNumCharsToPrint = g_Data.m_iLimitStringOutput;
    }
    else if(p_iNumCharsToPrint > g_Data.m_iLimitStringOutput)
    {
       p_iNumCharsToPrint = g_Data.m_iLimitStringOutput;
    }

    if(p_iSizeOutBuffer < 4)
       return 0;
    if(!p_pStr)
    {
        if(6 < p_iSizeOutBuffer)
        {
            strcpy(p_pszOutputBuffer, "\"null\"");
            return 6;
        }
        return 0;
    }
    
    int l_iFilled = 0;
    p_pszOutputBuffer[l_iFilled++] = ':'; 
    p_pszOutputBuffer[l_iFilled++] = '\"';
        
    for ( i=0; i < p_iNumCharsToPrint; i++ )
    {
        if(!MmIsAddressValid(p_pStr))
            break;
        if(l_bStopAtNull && !*p_pStr)
            break;

        //Make sure that we have space for 2 characters now
        //and the closing quote with NULL at the end
        if(l_iFilled + 4 >= p_iSizeOutBuffer)
            break;
        
        char c = *p_pStr;
        p_pszOutputBuffer[l_iFilled] = c;
        if ( c == '\0' ) 
        { 
           p_pszOutputBuffer[l_iFilled++] = '\\'; 
           p_pszOutputBuffer[l_iFilled] = '0'; 
        }
        else if ( c == '\r' ) 
        { 
           p_pszOutputBuffer[l_iFilled++] = '\\'; 
           p_pszOutputBuffer[l_iFilled] = 'r'; 
        }
        else if ( c == '\n' ) 
        { 
           p_pszOutputBuffer[l_iFilled++] = '\\'; 
           p_pszOutputBuffer[l_iFilled] = 'n'; 
        }
        else if ( c == '\t' ) 
        { 
           p_pszOutputBuffer[l_iFilled++] = '\\'; 
           p_pszOutputBuffer[l_iFilled] = 't'; 
        }
        l_iFilled++;
        p_pStr += p_iCharSize;
    }//for ( i=0; i < l_iLimitSize; i++ )

    if ( i == 0 )   // Not a valid string
        return 0;
    
    p_pszOutputBuffer[l_iFilled++] = '\"';
        // Valid string ptr - end quote and null
    p_pszOutputBuffer[l_iFilled] = 0;   
        // terminate the string
    
    return l_iFilled;
}//int ParamProcContext::iGetLPSTR

/*-----------------------------------------------------------------

   FUNCTION: ParamProcContext::iGetDDKString

   PURPOSE:  Print content of UNICODE_STRING or 
    ANSI_STRING to the output buffer
      
   PARAMETERS:                   .
                                  
   UNICODE_STRING * p_pUStr      UNICODE_STRING to print 
   PSTR p_pszOutputBuffer        pointer to the output buffer 
   int  p_iBuffSize              size of  the output buffer 
   int  p_iCharacterSize         1 for ANSI, 2 for Unicode

   RETURN VALUE:
      int       number of character placed in the output buffer
-----------------------------------------------------------------*/
int ParamProcContext::iGetDDKString
(
    UNICODE_STRING * p_pUStr,
    PSTR p_pszOutputBuffer,
    int  p_iSizeOutBuffer,
    int  p_iCharacterSize
)
{
    if(!p_pszOutputBuffer || p_iSizeOutBuffer < 4)
    {
        MYTRACE(TF_Error, "ERROR : bad params of iGetDDKString");
        return 0;
    }

    if(!p_pUStr)
    {
        if(6 < p_iSizeOutBuffer)
        {
            strcpy(p_pszOutputBuffer, "\"null\"");
            return 6;
        }
        return 0;
    }
    return iGetLPSTR((char*)p_pUStr->Buffer, 
        p_iCharacterSize, p_pUStr->Length/p_iCharacterSize,
        p_pszOutputBuffer, p_iSizeOutBuffer);
}//int ParamProcContext::iGetDDKString

/*-----------------------------------------------------------------

   FUNCTION: ParamProcContext::iPrintParameter

   PURPOSE:  Format parameter of various types to an output buffer
      
   PARAMETERS:                   .
                                  
   OneParamInfo & p_rParam        Description of the parameter
   PSTR p_pszReturnPrintedParams  Output buffer
   int  p_iParamsBuffSize         Size of the buffer
   int  p_iPrintIntegerBytes      Length of an integer parameter
   PCSTR p_pszParamType           Type of the parameter (string)
   
        The following 3 values a valid only if the parameter is
        as interface pointer

   RETURN VALUE:
      int       number of character placed in the output buffer
-----------------------------------------------------------------*/
int ParamProcContext::iPrintParameter
(
    OneParamInfo & p_rParam,
    PSTR    p_pszReturnPrintedParams,
    int        p_iParamsBuffSize,
    int        p_iPrintIntegerBytes,
    PCSTR    p_pszParamType
)
{
    int l_iFilledBytes = 0;

    //Output parameter's type
    l_iFilledBytes += iAddStringToOutputBuff(
        p_pszReturnPrintedParams + l_iFilledBytes, 
        p_iParamsBuffSize - l_iFilledBytes,
        p_pszParamType);
    
    if(l_iFilledBytes + 1 < p_iParamsBuffSize) 
        p_pszReturnPrintedParams[l_iFilledBytes++] = ':';
    p_pszReturnPrintedParams[l_iFilledBytes] = 0;

    ULONG l_dwCurParamValue = p_rParam.m_dwCurParamValue;
    ULONG * l_pdwCurParamAddr  = p_rParam.m_pdwAddrCurParameter;

    //If it is a pointer - try to dereference it
    if(p_rParam.m_byParamFlags & PF_Pointer)
    {
        if(!l_dwCurParamValue)
            return l_iFilledBytes;//NULL pointer
        if(!MmIsAddressValid((PVOID)l_dwCurParamValue))
        {
            if(30 < p_iParamsBuffSize - l_iFilledBytes)
            {
                l_iFilledBytes += iAddStringToOutputBuff(
                    p_pszReturnPrintedParams + l_iFilledBytes,
                    p_iParamsBuffSize - l_iFilledBytes,
                    "Bad pointer ");
                    
                l_iFilledBytes += iPrintInteger(
                    p_pszReturnPrintedParams + l_iFilledBytes,
                    p_iParamsBuffSize - l_iFilledBytes,
                    4, NULL, l_dwCurParamValue);
            }
            return l_iFilledBytes;
        }//if(!MmIsAddressValid((PVOID)l_dwCurParamValue))

        if(10 >= p_iParamsBuffSize - l_iFilledBytes)
            return l_iFilledBytes;

        l_iFilledBytes += iPrintInteger(
            p_pszReturnPrintedParams + l_iFilledBytes,
            p_iParamsBuffSize - l_iFilledBytes,
            4, NULL, l_dwCurParamValue);
        l_iFilledBytes += iAddStringToOutputBuff(
            p_pszReturnPrintedParams + l_iFilledBytes,
            p_iParamsBuffSize - l_iFilledBytes,
            "->");

        l_pdwCurParamAddr = (ULONG*)l_dwCurParamValue;
        l_dwCurParamValue = *((PULONG)l_dwCurParamValue);
    }//if(p_byFlags & PF_Pointer)

    //Print some parameters as integers
    if(p_iPrintIntegerBytes)
    {
        l_iFilledBytes += iPrintInteger(
            p_pszReturnPrintedParams + l_iFilledBytes,
            p_iParamsBuffSize - l_iFilledBytes,
            p_iPrintIntegerBytes, 
            l_pdwCurParamAddr,
            l_dwCurParamValue);
    }//if(p_iPrintIntegerBytes)

    ParamType l_ParamType = p_rParam.m_ParamType;

    //Special treatment for some parameter types
    if (l_ParamType == PARAM_LPSTR ||
        l_ParamType == PARAM_LPWSTR)
    {
        if(g_Data.m_iLimitStringOutput + 5 >= 
            p_iParamsBuffSize - l_iFilledBytes)
            return l_iFilledBytes;

        //Output string (or UNICODE string)
        int l_iWritten = iGetLPSTR( (PSTR)l_dwCurParamValue, 
           (l_ParamType == PARAM_LPWSTR)? 2 : 1, -1,
          p_pszReturnPrintedParams + l_iFilledBytes,
          p_iParamsBuffSize - l_iFilledBytes);
        l_iFilledBytes += l_iWritten;
    }
    else if(l_ParamType == PARAM_U_STR)
    {   //print UNICODE_STRING
        l_iFilledBytes += iGetDDKString(
            (UNICODE_STRING *)l_pdwCurParamAddr,
            p_pszReturnPrintedParams + l_iFilledBytes,
            p_iParamsBuffSize - l_iFilledBytes, 2);
    }
    else if(l_ParamType == PARAM_A_STR)
    {   //print ANSI_STRING
        l_iFilledBytes += iGetDDKString(
            (UNICODE_STRING*)l_pdwCurParamAddr,
            p_pszReturnPrintedParams + l_iFilledBytes,
            p_iParamsBuffSize - l_iFilledBytes, 1);
    }
    p_pszReturnPrintedParams[l_iFilledBytes] = 0;
    return l_iFilledBytes;
}//int ParamProcContext::iPrintParameter

/*-----------------------------------------------------------------

   FUNCTION: ParamProcContext::iProcParameter

   PURPOSE:  This function processes one parameter passes in or out
    of the intercepted function. It prints parameters to the
    output buffer for later logging. Also it detects pointers
    to interfaces and intercepts them. It preserves pointers
    to output parameters during the function call in order to
    be able to process them later during the function return.
      
   PARAMETERS:                   .
                                  
   OneParamInfo &  p_rParam         - description of the parameters
   bool            p_bParamFromStack - is it on a function stack or
                                      elsewhere (for example inside
                                       a VARIANT structure)
   PSTR  p_pszReturnPrintedParams    - output buffer
   int   p_iParamsBuffSize           - size of the buffer

   RETURN VALUE:
      int       number of character placed in the output buffer
-----------------------------------------------------------------*/
int ParamProcContext::iProcParameter
(
    OneParamInfo &  p_rParam,
    bool            p_bParamFromStack,
    PSTR        p_pszReturnPrintedParams,//must be a valid pointer
    int            p_iParamsBuffSize
)
{
    int  l_iReturnParamsFilled = 0;
    bool l_bPointer = false;
    bool l_bPreserveUntillExit = false;
    int     l_iPrintIntegerBytes = 0; 
    PSTR l_pszParamType = NULL;

    const UCHAR l_byParamFlags = p_rParam.m_byParamFlags;
    const ParamType    l_ParamType = p_rParam.m_ParamType;

    //first - switch by parameter type
    switch (l_ParamType)
    {
        case PARAM_DWORD:
            l_pszParamType = "ULONG"; 
            l_iPrintIntegerBytes = 4; 
            break;
        case PARAM_WORD:
            l_pszParamType = "WORD"; 
            l_iPrintIntegerBytes = 2; 
            break;
        case PARAM_BYTE:
            l_pszParamType = "UCHAR"; 
            l_iPrintIntegerBytes = 1; 
            break;
        case PARAM_FLOAT:
            l_pszParamType = "FLOAT";
            l_iPrintIntegerBytes = 0;//We do not support float
            break;
        case PARAM_DOUBLE:
            l_pszParamType = "DOUBLE";
            l_iPrintIntegerBytes = 0;//We do not support double
            if(!(l_byParamFlags & PF_Pointer))
                p_rParam.m_iParamSize = sizeof(double);
            break;
        case PARAM_DWORD64:
            l_pszParamType = "DWORD64";
            l_iPrintIntegerBytes = 8;
            if(!(l_byParamFlags & PF_Pointer))
                p_rParam.m_iParamSize = sizeof(__int64);
            break;
        case PARAM_LPSTR:
            l_pszParamType = "LPSTR"; 
            l_iPrintIntegerBytes = 4; 
            l_bPointer = true;
            break;
        case PARAM_LPWSTR:
            l_pszParamType = "LPWSTR"; 
            l_iPrintIntegerBytes = 4;
            l_bPointer = true;
            break;
        case PARAM_LPDATA:
            l_pszParamType = "LPDATA"; 
            l_iPrintIntegerBytes = 4; 
            l_bPointer = true;
            break;
        case PARAM_HANDLE:
            l_pszParamType = "HANDLE"; 
            l_iPrintIntegerBytes = 4; 
            break;
        case PARAM_HWND:
            l_pszParamType = "HWND"; 
            l_iPrintIntegerBytes = 4; 
            break;
        case PARAM_BOOL:
            l_pszParamType = "bool"; 
            l_iPrintIntegerBytes = 1; 
            break;
        case PARAM_LPCODE:
            l_pszParamType = "LPCODE"; 
            l_iPrintIntegerBytes = 4; 
            l_bPointer = true;
            break;
        case PARAM_U_STR:
            l_pszParamType = "UNICODE_STRING";
            l_iPrintIntegerBytes = 0;
            if(!(l_byParamFlags & PF_Pointer))
                p_rParam.m_iParamSize = sizeof(UNICODE_STRING);
            break;
        case PARAM_A_STR:
            l_pszParamType = "ANSI_STRING";
            l_iPrintIntegerBytes = 0;
            if(!(l_byParamFlags & PF_Pointer))
                p_rParam.m_iParamSize = sizeof(ANSI_STRING);
            break;
        default:
            l_pszParamType = "<unknown>"; 
            l_iPrintIntegerBytes = 0;
    }//switch (m_ParamType)

    if(l_byParamFlags & PF_Pointer)
       l_bPointer = true;

    //If the parameters is a pointer to output value, we will need
    //to preserve the pointer in order to process the value
    //during the function output (because the application stack
    //frame is not available then).
    if(l_bPointer && (l_byParamFlags & PF_Out))
        l_bPreserveUntillExit = true;
    if(l_bPreserveUntillExit && p_bParamFromStack)
    {
        //parameter must be 4 bytes long ( a pointer)
        bStoreRetreiveOutParameter(
            m_bOut? false : true, p_rParam.m_dwCurParamValue,
            m_pReturnStub, &m_iOutputParamIndex);
    }

    //Print [IN] parameters during function call and
    //[OUT] parameters during the return:
    bool l_bPrintNow = 
        ((m_bOut  && (l_byParamFlags & PF_Out)) ||
        (!m_bOut && (l_byParamFlags & PF_In)))? true : false;

    if(l_ParamType == PARAM_THIS)
    {
        //The first parameter in COM interface calls is a pointer
        //to the COM object itself - "this" pointer.
        m_dwThisAddress = p_rParam.m_dwCurParamValue;
        l_bPrintNow = false; //This's are hidded - don't print them
    }

    if(l_bPrintNow)
    {
        l_iReturnParamsFilled += iPrintParameter(p_rParam,
            p_pszReturnPrintedParams + l_iReturnParamsFilled,
            p_iParamsBuffSize - l_iReturnParamsFilled,
            l_iPrintIntegerBytes,
            l_pszParamType
            );
    }//if(l_bPrintNow)
    return l_iReturnParamsFilled;
}//int iProcParameter

/*-----------------------------------------------------------------

   FUNCTION: ParamProcContext::iProcAllParams

   PURPOSE:  Process all parameters. This includes printing
    parameters to the output buffer and intercepting interfaces
      
   PARAMETERS:                   .
                                  
   PSTR p_pszReturnPrintedParams  - output buffer
   int  p_iParamsBuffSize         - size of the buffer

   RETURN VALUE:
      int       number of character placed in the output buffer
-----------------------------------------------------------------*/
int ParamProcContext::iProcAllParams
(
    PSTR    p_pszReturnPrintedParams,
    int        p_iParamsBuffSize
)
{
    unsigned i;
    m_iOutputParamIndex = 0;
    
    int    l_iReturnParamsFilled = 0;
    p_pszReturnPrintedParams[0] = 0;

    m_dwThisAddress = 0;

    ULONG l_uiNumParams = m_pParams->byGetNumParams();

    int l_iParamArrayIndex, l_iParamArrayNextIndex;
    ULONG * l_pdwPlaceInStack = m_pCallStackFrame;//may be NULL

    l_iParamArrayIndex = 0;

    for ( i=0; i < l_uiNumParams; i++ )
    {
        ParamType l_ParamType = m_pParams->GetParam(
            l_iParamArrayIndex, l_iParamArrayNextIndex);
        
        UCHAR l_byExtraByte = 
            m_pParams->byGetParamExtraByte(l_iParamArrayIndex);
            //get the extra byte or 0

        UCHAR l_byParamFlags = 
            m_pParams->byGetParamFlags(l_iParamArrayIndex);
        
        OneParamInfo l_Param
            (l_ParamType, l_byParamFlags, 
            l_byExtraByte, l_pdwPlaceInStack);

        //Do the job for one paramter:
        l_iReturnParamsFilled += iProcParameter(l_Param, true,
            p_pszReturnPrintedParams + l_iReturnParamsFilled,
            p_iParamsBuffSize - l_iReturnParamsFilled);

        //Advance a pointer to the current parameter in the 
        //application stack
        if(l_pdwPlaceInStack)
        {
            int l_iParamSize = l_Param.iGetParamSize();
            if(!l_iParamSize)
                l_iParamSize = sizeof(ULONG);
            l_pdwPlaceInStack = (PULONG)
                (PUCHAR(l_pdwPlaceInStack) + l_iParamSize);
        }//if(l_pdwPlaceInStack)

        if (l_ParamType != PARAM_THIS &&
            i + 1 < l_uiNumParams)
        {
            //Add a comma between parameters
            //(even if the parameter itself was not printed)
            if(l_iReturnParamsFilled < p_iParamsBuffSize)
            {
                p_pszReturnPrintedParams
                    [l_iReturnParamsFilled++] = ',';
                p_pszReturnPrintedParams
                    [l_iReturnParamsFilled] = 0;
            }
        }
        l_iParamArrayIndex = l_iParamArrayNextIndex;
    }//for ( i=0; i < l_uiNumParams; i++ )
    p_pszReturnPrintedParams[l_iReturnParamsFilled] = 0;
    return l_iReturnParamsFilled;
}//int ParamProcContext::iProcAllParams


/*-----------------------------------------------------------------

   FUNCTION: Logger::iMakePrefixString

   PURPOSE:  Helper function which fills the output buffer by
    some number of blank spaces proportional to the current
    level of recursion. It is useful to make indents in the trace.
      
   PARAMETERS:                   .
                                  
   PSTR p_pszBuffer              Output buffer 
   UINT p_uiLevel                Current recursion level 
-----------------------------------------------------------------*/
int Logger::iMakePrefixString
(
    PSTR    p_pszBuffer,
    int     p_iBuffSize
)
{
    int l_iFilledBytes = 0;
    if(g_Data.m_bPrintProcessorNumber)
    {
        l_iFilledBytes += iPrintInteger(
            p_pszBuffer + l_iFilledBytes,
            p_iBuffSize - l_iFilledBytes,
            1, NULL, KeGetCurrentProcessorNumber());
        if(2 < p_iBuffSize - l_iFilledBytes)
        {
            p_pszBuffer[l_iFilledBytes++] = ' ';
            p_pszBuffer[l_iFilledBytes] = 0;
        }
    }
    if(g_Data.m_bPrintIRQL)
    {
        l_iFilledBytes += iPrintInteger(
            p_pszBuffer + l_iFilledBytes,
            p_iBuffSize - l_iFilledBytes,
            2, NULL, KeGetCurrentIrql());
        if(2 < p_iBuffSize - l_iFilledBytes)
        {
            p_pszBuffer[l_iFilledBytes++] = ' ';
            p_pszBuffer[l_iFilledBytes] = 0;
        }
    }
    if(g_Data.m_bPrintThreadId)
    {
        if(KeGetCurrentIrql()<= DISPATCH_LEVEL)
        {
            l_iFilledBytes += iPrintInteger(
                p_pszBuffer + l_iFilledBytes,
                p_iBuffSize - l_iFilledBytes,
                4, NULL, (ULONG)PsGetCurrentThreadId());
            if(2 < p_iBuffSize - l_iFilledBytes)
            {
                p_pszBuffer[l_iFilledBytes++] = ' ';
                p_pszBuffer[l_iFilledBytes] = 0;
            }
        }
        else
        {
            if(10 < p_iBuffSize - l_iFilledBytes)
            {
                memset(p_pszBuffer + l_iFilledBytes, ' ', 9);
                l_iFilledBytes += 9;
                p_pszBuffer[l_iFilledBytes] = 0;
            }
        }
    }
    p_pszBuffer[l_iFilledBytes] = 0;
    return l_iFilledBytes;
}//int Logger::iMakePrefixString

/*-----------------------------------------------------------------
   FUNCTION: Logger::vLogFunction

   PURPOSE:  Prints to the output log information about function
    call - name and input parameters. This information
    should be preprocessed by the caller.
      
   PARAMETERS:                   .
   
   const char * p_pszFunctionName - function name.
   ParamProcContext    * p_pParamContext - object, which prints
        parameters. May be NULL to avoid printing parameters
   char       *  p_pszBuffer - memory provided by our caller for
                    our internal use
   int           p_iSizeBuffer - size of p_pszBuffer
   const char * p_pszFunctionParams - formatted parameters
-----------------------------------------------------------------*/
void Logger::vLogFunction
(
    const char       * p_pszFunctionName,
    ParamProcContext * p_pParamContext,
    bool               p_bNotEnoughStubs,
    char       *       p_pszBuffer,
    int                p_iSizeBuffer
)
{
    int l_iBytesFilled = iMakePrefixString
        (p_pszBuffer, p_iSizeBuffer);
    
    //We are trying to generate string in the following format:
    //"%s{ %s(%s)", Prefix, Function Name, Parameters

    if(2 < p_iSizeBuffer - l_iBytesFilled)
    {
        p_pszBuffer[l_iBytesFilled++] = '{';
        p_pszBuffer[l_iBytesFilled++] = ' ';
        p_pszBuffer[l_iBytesFilled] = 0;
        
        if(p_bNotEnoughStubs)
        {
            l_iBytesFilled += iAddStringToOutputBuff(
                p_pszBuffer + l_iBytesFilled,
                p_iSizeBuffer - l_iBytesFilled,
                " !No rtn stubs! ");
        }

        l_iBytesFilled += iAddStringToOutputBuff(
            p_pszBuffer + l_iBytesFilled,
            p_iSizeBuffer - l_iBytesFilled,
            p_pszFunctionName);
        if(1 < p_iSizeBuffer - l_iBytesFilled)
        {
            p_pszBuffer[l_iBytesFilled++] = '(';
            p_pszBuffer[l_iBytesFilled] = 0;

            if(p_pParamContext)
            {
                l_iBytesFilled +=
                    p_pParamContext->iProcAllParams(
                        p_pszBuffer + l_iBytesFilled,
                        p_iSizeBuffer - l_iBytesFilled);
            }
            else
            {
                if(1 < p_iSizeBuffer - l_iBytesFilled)
                {
                    p_pszBuffer[l_iBytesFilled++] = '?';
                    p_pszBuffer[l_iBytesFilled] = 0;
                }
            }

            if(1 < p_iSizeBuffer - l_iBytesFilled)
            {
                p_pszBuffer[l_iBytesFilled++] = ')';
                p_pszBuffer[l_iBytesFilled] = 0;
            }
        }//if(1 < p_iSizeBuffer - l_iBytesFilled)
    }//if(2 < p_iSizeBuffer - l_iBytesFilled)

    if(2 < p_iSizeBuffer - l_iBytesFilled)
    {
        p_pszBuffer[l_iBytesFilled++] = '\r';
        p_pszBuffer[l_iBytesFilled++] = '\n';
        p_pszBuffer[l_iBytesFilled] = 0;
    }
    else
    {
        p_pszBuffer[p_iSizeBuffer-3] = '\r';
        p_pszBuffer[p_iSizeBuffer-2] = '\n';
        p_pszBuffer[p_iSizeBuffer-1] = 0;
    }
    OutputTraceString(TF_MainLog, p_pszBuffer, l_iBytesFilled);
}//void Logger::vLogFunction

/*----------------------------------------------------------------

   FUNCTION: Logger::LogReturn

   PURPOSE:  Prints to the output log information about function
    return - return value and output parameters. This information
    should be preprocessed by the caller.
      
   PARAMETERS:                   .
                                  
   const char * p_pszFunctionName   - function name
   ParamProcContext    * p_pParamContext - object, which prints
        parameters. May be NULL to avoid printing parameters
   bool         p_bOutputParamsExist - true if some output params
        were declared in the config file
   ULONG        p_dwReturnValue     - function's return value
   char       *  p_pszBuffer - memory provided by our caller for
                    our internal use
   int           p_iSizeBuffer - size of p_pszBuffer
-----------------------------------------------------------------*/
void Logger::LogReturn
(
    const char    *    p_pszFunctionName,
    ParamProcContext * p_pParamContext,
    bool               p_bOutputParamsExist,
    ULONG              p_dwReturnValue,
    char       *       p_pszBuffer,
    int                p_iSizeBuffer
)
{
    int l_iBytesFilled = iMakePrefixString
        (p_pszBuffer, p_iSizeBuffer);
    
    //We are trying to generate string in the following format:
    //"%s} %s = %X.%s", 
    //    l_szPrefix, p_pszFunctionName, p_dwReturnValue,
    //    p_pszFunctionParams);

    if(2 < p_iSizeBuffer - l_iBytesFilled)
    {
        p_pszBuffer[l_iBytesFilled++] = '}';
        p_pszBuffer[l_iBytesFilled++] = ' ';
        p_pszBuffer[l_iBytesFilled] = 0;
        
        l_iBytesFilled += iAddStringToOutputBuff(
            p_pszBuffer + l_iBytesFilled,
            p_iSizeBuffer - l_iBytesFilled,
            p_pszFunctionName);
        if(3 < p_iSizeBuffer - l_iBytesFilled)
        {
            p_pszBuffer[l_iBytesFilled++] = ' ';
            p_pszBuffer[l_iBytesFilled++] = '=';
            p_pszBuffer[l_iBytesFilled++] = ' ';
            p_pszBuffer[l_iBytesFilled] = 0;
            l_iBytesFilled += iPrintInteger(
                p_pszBuffer + l_iBytesFilled,
                p_iSizeBuffer - l_iBytesFilled,
                4, NULL, p_dwReturnValue);
            if(1 < p_iSizeBuffer - l_iBytesFilled)
            {
                p_pszBuffer[l_iBytesFilled++] = '.';
                p_pszBuffer[l_iBytesFilled] = 0;

                if(p_pParamContext)
                {
                    l_iBytesFilled += iAddStringToOutputBuff(
                        p_pszBuffer + l_iBytesFilled,
                        p_iSizeBuffer - l_iBytesFilled,
                        " Out values = (");
                    l_iBytesFilled +=
                        p_pParamContext->iProcAllParams(
                            p_pszBuffer + l_iBytesFilled,
                            p_iSizeBuffer - l_iBytesFilled);
                    if(1 < p_iSizeBuffer - l_iBytesFilled)
                    {
                        p_pszBuffer[l_iBytesFilled++] = ')';
                        p_pszBuffer[l_iBytesFilled] = 0;
                    }
                }
                else if(p_bOutputParamsExist)
                {
                    if(1 < p_iSizeBuffer - l_iBytesFilled)
                    {
                        p_pszBuffer[l_iBytesFilled++] = '?';
                        p_pszBuffer[l_iBytesFilled] = 0;
                    }
                }
            }//if(1 < p_iSizeBuffer - l_iBytesFilled)
        }//if(3 < p_iSizeBuffer - l_iBytesFilled)
    }//if(2 < p_iSizeBuffer - l_iBytesFilled)

    if(2 < p_iSizeBuffer - l_iBytesFilled)
    {
        p_pszBuffer[l_iBytesFilled++] = '\r';
        p_pszBuffer[l_iBytesFilled++] = '\n';
        p_pszBuffer[l_iBytesFilled] = 0;
    }
    else
    {
        p_pszBuffer[p_iSizeBuffer-3] = '\r';
        p_pszBuffer[p_iSizeBuffer-2] = '\n';
        p_pszBuffer[p_iSizeBuffer-1] = 0;
    }

    OutputTraceString(TF_MainLog, p_pszBuffer, l_iBytesFilled);
}//void Logger::LogReturn

/*-----------------------------------------------------------------

   FUNCTION: Logger::IntrerceptEntry

   PURPOSE:  Main entry of the interceptor of exported functions.
              It will be called by IntrerceptEntryNaked.
      
   PARAMETERS:                   .
                                  
   ASM_APIFunctionStub * p_pStub  Pointer to the calling stub
   PULONG  p_pdwFrame             function call stack frame
-----------------------------------------------------------------*/
/*static*/
void Logger::IntrerceptEntry
(
    ASM_APIFunctionStub * p_pStub,
    PULONG  p_pdwFrame
)
{
#ifdef _STATISTICS
    GetStat l_Stat(g_StatEntry);
#endif

    __try
    {
        ReturnStub * l_pReturnStub = NULL;
        bool l_bReleaseStub = false;
        bool l_bNotEnoughStubs = false;
        
        if(g_Data.m_pRetMgr &&
            (g_Data.m_pParams->m_bPrintParams ||
             g_Data.m_pParams->m_bInterceptReturns))
        {
            //We need to get a stub is we are required to print
            //the function parameters or to intercept the return.

            ULONG l_ulReturnAddress = p_pdwFrame[0];
            l_pReturnStub = 
                g_Data.m_pRetMgr->pFindStub(l_ulReturnAddress);
            if(l_pReturnStub)
            {
                if(g_Data.m_pParams->m_bInterceptReturns)
                {
                    //Intercept the return address
                    //Fill the rest of the stub's members
                    l_pReturnStub->m_pStub = p_pStub;
    
                    //Replace the return address on the stack
                    //by the address of our stub's call instruction
                    p_pdwFrame[0] = (ULONG)
                        (&l_pReturnStub->m_byCallInstruction);
                }
                else
                {
                    //If we haven't intercepted the return,
                    //we will need to release the stub here
                    l_bReleaseStub = true;
                }
            }
            else
            {
                l_bNotEnoughStubs = true;
            }
        }//if(g_Data.m_pRetMgr &&

        if(l_pReturnStub)
        {
            //If there no return stub avaliable,
            //we will simply print functions name without parameters
            //and will not intercept the return.
            
            ParamProcContext    * l_pParamContext = NULL;
            if(g_Data.m_pParams->m_bPrintParams)
            {
                if(p_pdwFrame)
                    p_pdwFrame++;//Skip ULONG return address

                l_pReturnStub->m_Context.vInit(
                    ((FunctionParameters *)
                    p_pStub->m_dwParamsOffset),
                    l_pReturnStub, p_pdwFrame, false);

                l_pParamContext = &l_pReturnStub->m_Context;
            }//if(g_Data.m_pParams->m_bPrintParams)
            vLogFunction(
                (const char *)p_pStub->m_dwFunctionNameOffset,
                l_pParamContext, l_bNotEnoughStubs,
                l_pReturnStub->m_szFormattedString, 
                sizeof(l_pReturnStub->m_szFormattedString));

            if(l_bReleaseStub)
            {
                //Release the stub by erasing the return address member
                InterlockedExchange(
                    (long*)&l_pReturnStub->m_dwReturnAddress, 0);
            }
        }//if(l_pReturnStub)
        else
        {
            //We cannot use a large buffer embedded in the stub.
            //Therefore, allocate a small buffer of the stack.
            char l_szBuffer[64];
            vLogFunction(
                (const char *)p_pStub->m_dwFunctionNameOffset,
                NULL, l_bNotEnoughStubs,
                l_szBuffer, sizeof(l_szBuffer));
        }
    }
    __except(1)
    {
        MYTRACE(TF_Error, 
            "ERROR : exception inside IntrerceptEntry");
    }
#ifdef _STATISTICS
    l_Stat.vRecord();
#endif
}//void __stdcall IntrerceptEntry

/*-----------------------------------------------------------------

   FUNCTION: Logger::IntrerceptEntryNaked

   PURPOSE:  Main entry of the interceptor of exported functions
             It will be called by our stubs.
      
-----------------------------------------------------------------*/
/*static*/
void __declspec(naked) Logger::IntrerceptEntryNaked()
{
    PULONG l_pReturnAddress;
    PULONG l_pdwFrame;
    ASM_APIFunctionStub * l_pStub;
    __asm 
    {
        PUSHAD      ; Push EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI
        PUSHFD      ; Store flags
        MOV    EBP, ESP   ; allow access to local variables
        SUB    ESP, __LOCAL_SIZE ;reserve space for local variables
    
        mov l_pReturnAddress, EBP
    }
    l_pReturnAddress += 8+1;//This address is in our stub right
                            //after the call instruction
    
    l_pStub = (ASM_APIFunctionStub *)( (*l_pReturnAddress) - 5 - 
        offsetof(ASM_APIFunctionStub, m_Code.instr_CALL));
    
    l_pdwFrame = l_pReturnAddress + 2;
    //points to the return address of original call to our stub

    IntrerceptEntry(l_pStub, l_pdwFrame);

    __asm
    {
        ADD    ESP, __LOCAL_SIZE ;remove local variables from stack
        POPFD                          ;restore flags
        POPAD                          ;restore registers
        RET
    }
}//void  IntrerceptEntryNaked

/*-----------------------------------------------------------------

   FUNCTION: Logger::ProcessReturn

   PURPOSE:  Perform necessary processing of function return -
    format output parameters and log it to the output file.
      
   PARAMETERS:                   .
                                  
   ASM_APIFunctionStub * p_pStub  - stub describing the function
   ULONG      p_dwReturnValue     - return value
   ULONG      p_iRecursionLevel   - current recursion level                  
   ReturnStub * p_pReturnStub     - pointer to a return stub
-----------------------------------------------------------------*/
void Logger::ProcessReturn
(
    ASM_APIFunctionStub *  p_pStub,
    ULONG                  p_dwReturnValue,
    ReturnStub          *  p_pReturnStub
)
{
    ParamProcContext * l_pParamContext = NULL;
    
    FunctionParameters & l_rParams = 
        *((FunctionParameters *)p_pStub->m_dwParamsOffset);
    
    bool l_bOutputParamsExist = 
        (l_rParams.byGetFlags() & PF_Out)?true:false;

    if(g_Data.m_pParams->m_bPrintParams)
    {
        if(l_bOutputParamsExist)
        {
            p_pReturnStub->m_Context.vInit(
                ((FunctionParameters *)p_pStub->m_dwParamsOffset),
                p_pReturnStub, NULL, true);
            l_pParamContext = &p_pReturnStub->m_Context;
        }//if((l_rParams.byGetFlags() & PF_Out))
    }

    LogReturn((const char *)p_pStub->m_dwFunctionNameOffset, 
        l_pParamContext, l_bOutputParamsExist, p_dwReturnValue,
        p_pReturnStub->m_szFormattedString, 
        sizeof(p_pReturnStub->m_szFormattedString));
}//void Logger::ProcessReturn

void Logger::PrintStatistics()
{
#ifdef _STATISTICS
    MYTRACE(TF_MainLog, "Statistics: "
        "IntrerceptEntry: %u calls, %I64u ticks, min %u max %u\n"
        "CommonReturnPoint: %u calls, %I64u ticks, min %u max %u ",
        g_StatEntry.m_ulNumberOfCalls, g_StatEntry.m_ui64TotalTimeIn,
        g_StatEntry.m_ulMinTimeIn, g_StatEntry.m_ulMaxTimeIn,
        g_StatReturn.m_ulNumberOfCalls, g_StatReturn.m_ui64TotalTimeIn,
        g_StatReturn.m_ulMinTimeIn, g_StatReturn.m_ulMaxTimeIn
        );
#endif
}//void Logger::PrintStatistics()
