/*****************************************************************
Module : ParseCfg.cpp  - parser of configuration file
             Written 1998,1999 by Dmitri Leman
Purpose: COM/OLE tracer and NT kernel mode tracer.
******************************************************************/
#include "Headers.h"
#include "Trace.h"
#include <ctype.h>
#include <limits.h>
#include "intermap.h"
#include "parmtype.h"
#include "ParseCfg.h"
#include "Trace.h"


/*-----------------------------------------------------------------

   FUNCTION: Parser::Parser

   PURPOSE:  Constructor
-----------------------------------------------------------------*/
Parser::Parser()
{
    m_pAllIntr = new AllInterfacesUnderConstruction;

    m_iCurInterface = -1;
    m_pCurInterface = NULL;

    m_pCurFunction = NULL;
    m_CurrentFunction = FT_None;
    *m_szFunctionName = 0;
    m_iFunctionIndex = -1;

    *m_szDLLorInterfaceName = 0;
    *m_szBaseInterfaceName = 0;

    m_iLineNumber = 0;

    m_bLogGetProcAddress = FALSE;
    //IID  m_InterfaceID;
}//Parser::Parser()

/*-----------------------------------------------------------------

   FUNCTION: Parser::~Parser

   PURPOSE:  Destructor
-----------------------------------------------------------------*/
Parser::~Parser()
{
    delete m_pAllIntr;
}
 
/*-----------------------------------------------------------------

   FUNCTION: Parser::bAddInterface

   PURPOSE:  Declare a new interface
      
   RETURN VALUE:
      BOOL       TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bAddInterface()
{
    if (m_CurrentFunction != FT_InterfaceName)
        return FALSE;

    m_iCurInterface = 
        m_pAllIntr->iAddInterface(
        m_szDLLorInterfaceName, m_szBaseInterfaceName,
        m_InterfaceID, &m_pCurInterface);
    if(m_iCurInterface < 0)
        return FALSE;
    return TRUE;
}//BOOL Parser::bAddInterface

/*-----------------------------------------------------------------

   FUNCTION: Parser::bStartFunctionConstruction

   PURPOSE:  Start parsing and construction of a new function
      
   RETURN VALUE:
      BOOL       TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bStartFunctionConstruction()
{
    BOOL l_bResult = FALSE;

    m_pCurFunction = NULL;
    if (m_CurrentFunction == FT_Export)
    {
        InterfaceUnderConstruction * l_pIntr = NULL;
        int l_iIndex = m_pAllIntr->iAddOrFindDLL
            (m_szDLLorInterfaceName, &l_pIntr);
        if(l_iIndex < 0 || !l_pIntr)
        {
            MYTRACE(TF_Error, 
                "ERROR : cannot add DLL %s to interfaces",
                m_szDLLorInterfaceName);
        }
        else
        {
            if(!stricmp(m_szFunctionName, "GetProcAddress"))
                m_bLogGetProcAddress = TRUE;
            l_bResult = l_pIntr->bAddFunction(m_szFunctionName, 
                m_iFunctionIndex, &m_pCurFunction);
        }
    }
    else if (m_CurrentFunction == FT_InterfaceFunc)
    {
        l_bResult = m_pCurInterface? m_pCurInterface->bAddFunction(
            m_szFunctionName, m_iFunctionIndex, &m_pCurFunction) : 
            FALSE;
    }
    else
        l_bResult = TRUE;//there is no function

    if(!l_bResult)
        MYTRACE(TF_Error, 
            "ERROR : failed to add function %s for %s",
            m_szFunctionName, m_szDLLorInterfaceName);

    return l_bResult;
}//BOOL Parser::bStartFunctionConstruction


/*-----------------------------------------------------------------

   FUNCTION: Parser::bFinishFunctionConstruction

   PURPOSE:  Finish parsing of a function

   RETURN VALUE:
      BOOL   TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bFinishFunctionConstruction()
{
    m_pCurFunction = NULL;
    m_CurrentFunction = FT_None;
    return TRUE;
}//BOOL Parser::bFinishFunctionConstruction

/*-----------------------------------------------------------------

   FUNCTION: Parser::bAddParam

   PURPOSE:  Add function parameter
      
   PARAMETERS:                   .
                                  
    ParamType   p_Type              type of parameter
    BYTE        p_byFlags           flags from enum ParamFlags
    BOOL        p_bExtraBytePresent TRUE if this parameter has 
                                    the second byte
    BYTE        p_byExtra           the second byte

   RETURN VALUE:
      BOOL        TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bAddParam
(
    ParamType   p_Type, 
    BYTE        p_byFlags,
    BOOL        p_bExtraBytePresent,
    BYTE        p_byExtra
)
{
    if(!m_pCurFunction)
    {
        MYTRACE(TF_Error, "INTERNAL ERROR: cannot add parameter");
        return FALSE;
    }
    return m_pCurFunction->Params().bAddParam(p_Type, 
        p_byFlags, p_bExtraBytePresent, p_byExtra);
}//BOOL Parser::bAddParam

/*-----------------------------------------------------------------

   FUNCTION: Parser::bGetToken

   PURPOSE:  Helper function to extract a token from a line up to
    the specified character ot the end of the line 
      
   PARAMETERS:                   .
                                  
   PCSTR   p_pszFullLine        The whole line 
   PCSTR & p_rpszString         pointer to the current point in the
                                line where to start looking
   char    p_cSeparatorChar     character to terminate the token 
   PSTR    p_pszReturnBuffer    the token will be copied here 
   int     p_iSizeRetBuffer     size of the buffer 
   PCSTR   p_pszTokenNameForDebug 

   RETURN VALUE:
      BOOL        TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bGetToken
(
    PCSTR   p_pszFullLine,
    PCSTR & p_rpszString,
    char    p_cSeparatorChar,
    PSTR    p_pszReturnBuffer,
    int     p_iSizeRetBuffer,
    PCSTR   p_pszTokenNameForDebug
)
{
    int l_iLen;
    if(!p_cSeparatorChar)
        l_iLen = strlen(p_rpszString);
    else
    {
        char * l_pszEnd = 
            strchr(p_rpszString, p_cSeparatorChar);
        if(!l_pszEnd)
        {
            MYTRACE(TF_Error, 
                "ERROR : cannot find a '%c' separator in line %s",
                p_cSeparatorChar, p_pszFullLine);
            return FALSE;
        }
        l_iLen = l_pszEnd - p_rpszString;
    }

    if(p_pszReturnBuffer)
    {
        if(l_iLen + 1 > p_iSizeRetBuffer)
        {
            MYTRACE(TF_Error, "ERROR : %s %s is too long",
                p_pszTokenNameForDebug, p_rpszString);
            return FALSE;
        }//if(l_iLen + 1 > p_iSizeRetBuffer)

        memcpy(p_pszReturnBuffer, p_rpszString, l_iLen);
        p_pszReturnBuffer[l_iLen] = 0;
    }
    
    if(!p_cSeparatorChar)
        p_rpszString += l_iLen;     //stay on the '\0'
    else
        p_rpszString += l_iLen+1;   //skip the separator
    return TRUE;
}//BOOL Parser::bGetToken

/*-----------------------------------------------------------------

   FUNCTION: Parser::bScanFile

   PURPOSE:  Parse the whole file
      
   PARAMETERS:                   .
                                  
   FILE * p_pInputFile            file pointer

   RETURN VALUE:
      BOOL        TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bScanFile
(
    FILE * p_pInputFile
)
{
    char l_szInput[256];
    
    int l_iInterfaceDescription = 0;
    BOOL l_bReadNumInterfaces = FALSE;
    int l_iNextInterfaceIndex = 0;
    m_iLineNumber = 0;
    while ( fgets(l_szInput, sizeof(l_szInput), p_pInputFile) )
    {
        m_iLineNumber++;
        PSTR l_pszNewline, l_pszInput = l_szInput;
        
        while ( isspace(*l_pszInput) )
            l_pszInput++;
        l_pszNewline = strrchr(l_pszInput, '\n');
        if ( l_pszNewline )
            *l_pszNewline = 0;//Cut newline character
        
        if ( !*l_pszInput ) //blank line
            continue;
        if ( *l_pszInput == ';' )//This line is commented out
            continue;

        if(l_iInterfaceDescription == 1)
        {
            //We are between the first line of an interface
            //and the opening bracket
            if(*l_pszInput == '{')
                l_iInterfaceDescription = 2;
            continue;
        }//if(l_iInterfaceDescription == 1)
        if(l_iInterfaceDescription == 2)
        {
            //We are inside the interface declaration brackets
            if(*l_pszInput == '}')
            {
                bFinishFunctionConstruction();
                l_iInterfaceDescription = 0;
                continue;
            }
        }//if(l_iInterfaceDescription == 2)
        
        FunctionType l_NewFunction = 
            IsNewFunctionOrInterface(l_pszInput);

        if(l_NewFunction != FT_None)
        {
            //Finish construction of the previous function (if any)
            bFinishFunctionConstruction();

            if(l_iInterfaceDescription == 2 &&
                l_NewFunction != FT_InterfaceFunc)
            {
                MYTRACE(TF_Error, 
                    "ERROR: only interface functions should be "
                    "declared inside an interface declaration");
            }
            else  if(l_iInterfaceDescription == 0 &&
                l_NewFunction == FT_InterfaceFunc)
            {
                MYTRACE(TF_Error, 
                    "ERROR : interface functions should not be "
                    "declared outside an interface declaration");
            }
            else
            {
                if(bParseLine(l_pszInput,l_NewFunction))
                {
                    if(l_NewFunction == FT_InterfaceName)
                    {
                        l_iInterfaceDescription = 1;
                        
                        bAddInterface();
                    }
                    else if(l_NewFunction == FT_InterfaceFunc)
                    {
                        bStartFunctionConstruction();
                        bAddParam(PARAM_THIS, PF_In, FALSE, 0);
                    }
                    else if(l_NewFunction == FT_Export)
                    {
                        bStartFunctionConstruction();
                    }
                }
            }
        }//if(l_NewFunction != FT_None)
        else    // A parameter line
        {
            if(!bParseAndAddParam(l_pszInput))
            {
                if ( (*l_pszInput != 0) && 
                    (stricmp(l_pszInput, "VOID") != 0) )
                {
                    MYTRACE(TF_Error, 
                        "ERROR in parameters. "
                        "Line %d:\"%s\" DLL=\"%s\" Func=\"%s\"",
                        m_iLineNumber,
                        l_pszInput, 
                        pszGetDLLName(),
                        pszGetFuncName());
                }
            }
        }
    }
    //Finish construction of the previous function (if any)
    bFinishFunctionConstruction();
    return TRUE;
}//BOOL Parser::bScanFile

/*--------------------------------------------------------------

   FUNCTION: Parser::bLoadConfigFile

   PURPOSE:  Load and parse a configuration file describing 
    exported functions and COM interfaces to be intercepted by 
    the tracer
      
   PARAMETERS:                   .
                                  
   char * p_pszConfiFileName      name of the file

   RETURN VALUE:
      BOOL      TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bLoadConfigFile
(
    char * p_pszConfiFileName
)
{
    if(!p_pszConfiFileName)
        return FALSE;
    FILE * l_pFile = fopen(p_pszConfiFileName, "rt");
    if (!l_pFile)
    {
        MYTRACE(TF_Error, 
            "ERROR : cannot open file %s",
            p_pszConfiFileName);
        return FALSE;
    }

    bScanFile(l_pFile);

    fclose(l_pFile);

    m_pAllIntr->iAddIUnknown();
    
    return TRUE;
}//BOOL Parser::bLoadConfigFile

/*-----------------------------------------------------------------

   FUNCTION: Parser::iGetExportedSize

   PURPOSE:   Determine how much space will the configuration
    data take after exporting to a binary file
      
   RETURN VALUE:
      int   size of the data
-----------------------------------------------------------------*/
int Parser::iGetExportedSize()
{
    int l_iSize = m_pAllIntr->iExport(NULL, INT_MAX, 0);
    return l_iSize;
}//int Parser::iGetExportedSize

/*-----------------------------------------------------------------

   FUNCTION: Parser::iExport

   PURPOSE:  Export the whole configuration data to a binary file
      
   PARAMETERS:                   .
   
   PBYTE   p_pMem            beginning of the whole config data         
   int     p_iTotalSize      total size of the whole config data       
   int     p_iCurOffset      number of bytes already filled            

   RETURN VALUE:
      int       return number of bytes added to the exported area
                                  
-----------------------------------------------------------------*/
int Parser::iExport
(
    PBYTE   p_pMem, 
    int     p_iTotalSize, 
    int     p_iCurOffset
)
{
    int l_iSize = m_pAllIntr->iExport
        (p_pMem, p_iTotalSize, p_iCurOffset);
    return l_iSize;
}//int Parser::iExport

/*-----------------------------------------------------------------

   FUNCTION: Parser::IsNewFunctionOrInterface

   PURPOSE: Determine whether an input line describes
    an exported function or an interface method
      
   PARAMETERS:                   .
                                  
   PSTR p_pszInputLine            Input line

   RETURN VALUE:
      FunctionType        Type of the function
-----------------------------------------------------------------*/
FunctionType Parser::IsNewFunctionOrInterface
(
    PSTR p_pszInputLine
)
{
    const char l_szExpKeyWord[] = "EXP:";
    const char l_szInterfaceKeyWord[] = "Interface:";
    const char l_szInterfaceFuncKeyWord[] = "FUNC:";

    if(!memicmp(p_pszInputLine, 
        l_szExpKeyWord, sizeof(l_szExpKeyWord)-1))
        return FT_Export;
    if(!memicmp(p_pszInputLine, 
        l_szInterfaceKeyWord, sizeof(l_szInterfaceKeyWord)-1))
        return FT_InterfaceName;
    if(!memicmp(p_pszInputLine, 
        l_szInterfaceFuncKeyWord, 
        sizeof(l_szInterfaceFuncKeyWord)-1))
        return FT_InterfaceFunc;

    return FT_None;
}//FunctionType Parser::IsNewFunctionOrInterface

/*-----------------------------------------------------------------

   FUNCTION: Parser::bParseLine

   PURPOSE:  Parse function declaration line
      
   PARAMETERS:                   .
                                  
   PSTR p_pszInputLine           Line to parse 
   FunctionType p_Type           Function type 

   RETURN VALUE:
      BOOL       TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bParseLine
(
    PSTR p_pszInputLine,
    FunctionType p_Type
)
{
    //Prepare for an error:
    m_CurrentFunction = FT_None;

    *m_szFunctionName = 0;
    m_iFunctionIndex = -1;

    if(p_Type != FT_InterfaceFunc)
    {
        *m_szDLLorInterfaceName = 0;
        *m_szBaseInterfaceName = 0;
        memset(&m_InterfaceID, 0, sizeof(m_InterfaceID));
        m_iCurInterface = -1;
        m_pCurInterface = NULL;
    }

    BOOL l_bResult = FALSE;
    switch(p_Type)
    {
    case FT_Export:
        l_bResult = bParseExportedFunction(p_pszInputLine);
        break;
    case FT_InterfaceFunc:
        l_bResult = bParseInterfaceFunction(p_pszInputLine);
        break;
    case FT_InterfaceName:
        l_bResult = bParseInterfaceName(p_pszInputLine);
        break;
    }
    if(!l_bResult)
        return FALSE;
    m_CurrentFunction = p_Type;
    return TRUE;
}//BOOL Parser::bParseLine

/*-----------------------------------------------------------------

   FUNCTION: Parser::bParseExportedFunction

   PURPOSE:  Parse declaration of exported function
      
   PARAMETERS:                   .
                                  
   PCSTR p_pszInputLine           Input line

   RETURN VALUE:
      BOOL         TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bParseExportedFunction
(
    PCSTR p_pszInputLine
)
{
    //We have an exported function in the form 
    //EXP:DLLName:FunctionName

    PCSTR l_pszRestInputLine = p_pszInputLine;
    //Try to skip "EXP" keyword
    if(!bGetToken(p_pszInputLine, l_pszRestInputLine, ':', 
        NULL, 0, ""))
        return FALSE;

    //Now get the DLL or Interface name

    if(!bGetToken(p_pszInputLine, l_pszRestInputLine, ':',
        m_szDLLorInterfaceName, sizeof(m_szDLLorInterfaceName),
        "DLL or interface name"))
        return FALSE;
    //OK, DLL/interface name is done. Now - function name

    if(!bGetToken(p_pszInputLine, l_pszRestInputLine, 0,
        m_szFunctionName, sizeof(m_szFunctionName),
        "function name"))
        return FALSE;
    return TRUE;
}//BOOL Parser::bParseExportedFunction

/*-----------------------------------------------------------------

   FUNCTION: Parser::bParseInterfaceName

   PURPOSE:  Parse declaration of an interface
      
   PARAMETERS:                   .
                                  
   PCSTR p_pszInputLine           Input line

   RETURN VALUE:
      BOOL   TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bParseInterfaceName
(
    PCSTR p_pszInputLine
)
{
    //We have an Interface declaration - INTERFACE:Name:IID

    PCSTR l_pszRestInputLine = p_pszInputLine;

    //Try to skip "Interface" keyword
    if(!bGetToken(p_pszInputLine, l_pszRestInputLine, ':', 
        NULL, 0, ""))
        return FALSE;

    //Now get the Interface name

    if(!bGetToken(p_pszInputLine, l_pszRestInputLine, ':',
        m_szDLLorInterfaceName, sizeof(m_szDLLorInterfaceName),
        "DLL or interface name"))
        return FALSE;

    //Read IID
    //{00000000-0000-0000-C000-000000000046}
    //("6d5140c1-7436-11ce-8034-00aa006009fa")) IServiceProvider;
    //We cannot use IIDFromString because we don't want
    //to load OLE32.DLL to some other process's address space.
    DWORD l_dwTemp[10];

    if(11 != sscanf(l_pszRestInputLine, 
        "{%lx-%x-%x-%2x%2x-%2x%2x%2x%2x%2x%2x}",
        &m_InterfaceID.Data1, &l_dwTemp[0], &l_dwTemp[1],
        &l_dwTemp[2], &l_dwTemp[3], &l_dwTemp[4],
        &l_dwTemp[5], &l_dwTemp[6], &l_dwTemp[7],
        &l_dwTemp[8], &l_dwTemp[9]))
    {
        MYTRACE(TF_Error, 
            "ERROR : cannot read an interface IID in the line %s",
            p_pszInputLine);
        return FALSE;
    }

    m_InterfaceID.Data2 = (WORD)l_dwTemp[0];
    m_InterfaceID.Data3 = (WORD)l_dwTemp[1];
    for(int i = 0; i < 8; i++)
        m_InterfaceID.Data4[i] = (BYTE) l_dwTemp[i+2];

    //Now get (optional) base Interface name
    *m_szBaseInterfaceName = 0;
    l_pszRestInputLine = strchr(l_pszRestInputLine, '}');
    if(l_pszRestInputLine)
    {
        l_pszRestInputLine++;
        if(*l_pszRestInputLine == ':')
        {
            l_pszRestInputLine++;
            bGetToken(p_pszInputLine, l_pszRestInputLine, 0,
                m_szBaseInterfaceName, 
                sizeof(m_szBaseInterfaceName),
                "Base interface name");
        }//if(*l_pszRestInputLine == ':')
    }
    return TRUE;
}//BOOL Parser::bParseInterfaceName

/*----------------------------------------------------------------

   FUNCTION: Parser::bParseInterfaceFunction

   PURPOSE:  Parse declaration of an interface function
      
   PARAMETERS:                   .
                                  
   PCSTR p_pszInputLine           Input line

   RETURN VALUE:
      BOOL      TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bParseInterfaceFunction
(
    PCSTR p_pszInputLine
)
{
    //We have an Interface function - FunctionName:FunctionIndex

    PCSTR l_pszRestInputLine = p_pszInputLine;

    if(!bGetToken(p_pszInputLine, l_pszRestInputLine, ':', 
        NULL, 0, ""))
        return FALSE;
    
    if(!bGetToken(p_pszInputLine, l_pszRestInputLine, ':',
        m_szFunctionName, sizeof(m_szFunctionName),
        "function name"))
        return FALSE;
    
    PSTR l_pszEnd = NULL;
    m_iFunctionIndex = strtol(l_pszRestInputLine, &l_pszEnd, 0);
    if(l_pszEnd <= l_pszRestInputLine)
    {
        MYTRACE(TF_Error, 
            "ERROR : cannot find a function index in the line %s",
            p_pszInputLine);
        return FALSE;
    }
    return TRUE;
}//BOOL Parser::bParseInterfaceFunction

struct ParamNameAndType
{
    PSTR        m_pszName;// Param names used in the config file
    ParamType   m_Type;   // from enum ParamType
};

static ParamNameAndType s_Params[] = 
{
{"DWORD", PARAM_DWORD},
{"WORD", PARAM_WORD},
{"BYTE", PARAM_BYTE},
{"FLOAT", PARAM_FLOAT},
{"DOUBLE", PARAM_DOUBLE},
{"DWORD64", PARAM_DWORD64},
{"LPSTR", PARAM_LPSTR},
{"LPWSTR", PARAM_LPWSTR},
{"LPDATA", PARAM_LPDATA},
{"HANDLE", PARAM_HANDLE},
{"HWND", PARAM_HWND},
{"BOOL", PARAM_BOOL},
{"BOOLEAN", PARAM_BOOL},
{"LPCODE", PARAM_LPCODE},
{"LPIUNK", PARAM_LPIUNKNOWN},
{"LPINTERF", PARAM_LPINTERFACE},
{"REFGUID", PARAM_LPGUID},
{"BSTR", PARAM_BSTR},
{"THIS", PARAM_THIS},
{"VARIANT", PARAM_VARIANT},
{"UNICODE_STRING", PARAM_U_STR},
{"ANSI_STRING", PARAM_A_STR},
};//static ParamNameAndType s_Params

/*-----------------------------------------------------------------

   FUNCTION: Parser::bParseAndAddParam

   PURPOSE:  Parse a string which may contain a function parameter
    and fill ParamsUnderConstr & p_rParams structure with the 
    information about the parameter.
      
   PARAMETERS:                   .
                                  
   IN  PSTR                p_pszParam  String with a parameter

   RETURN VALUE:
      BOOL          TRUE on success, FALSE on failure
-----------------------------------------------------------------*/
BOOL Parser::bParseAndAddParam
(
    IN  PSTR pszParam
)
{
    int l_iLenParamType;
    BYTE l_byParamFlags = PF_In;//default is "IN"
    
    if(strchr(pszParam, '*'))
        l_byParamFlags |= PF_Pointer;

    char * p = strchr(pszParam, ' ');
    if(p)
    {
        l_iLenParamType = p - pszParam;

        if(strstr(p, "OUT"))
        {
            l_byParamFlags |= PF_Out;

            if(!strstr(p, "IN"))
                l_byParamFlags &= ~PF_In;
        }//if(strstr(p, "OUT"))
    }
    else
        l_iLenParamType = strlen(pszParam);
        
    for (int i=0; i < (sizeof(s_Params)/sizeof(s_Params[0])); i++ )
    {
        ParamNameAndType & l_rParam = s_Params[i];

        if(memcmp(l_rParam.m_pszName, pszParam, l_iLenParamType) ||
           l_rParam.m_pszName[l_iLenParamType])
           continue;//name does not match
            
        if(l_rParam.m_Type == PARAM_LPIUNKNOWN)
        {
            const char l_szKeyWord[] = "iid_is";
            char * l_pID = p? strstr(p, l_szKeyWord) : NULL;
            if(l_pID)
            {
                char * l_pszEnd = NULL;
                int l_iID = strtol(l_pID + sizeof(l_szKeyWord)-1, 
                    &l_pszEnd,0);
                if(l_pszEnd <= l_pID)
                {
                    MYTRACE(TF_Error, 
                        "ERROR : cannot read ID from line %s",
                        pszParam);
                    return FALSE;
                }
                else if(UINT(l_iID) > 254)
                {
                    MYTRACE(TF_Error, 
                        "ERROR : ID %d > 254 in line %s",
                        l_iID, pszParam);
                    return FALSE;
                }
                else
                {
                    if(m_CurrentFunction == FT_InterfaceFunc)
                        l_iID++;
                    //In the interface all parameters are
                    //offset by THIS parameter

                    return bAddParam(l_rParam.m_Type, 
                        l_byParamFlags, TRUE, BYTE(l_iID));
                }
            }//if(!_pID)
            else
            {
                MYTRACE(TF_Error, "WARNING : LPIUNK does "
                    "not have an \"%s\" keyword",
                    l_szKeyWord);
            }
        }//if(l_rParam.m_Type == PARAM_LPIUNKNOWN)

        if(l_rParam.m_Type == PARAM_LPINTERFACE)
        {
            const char l_szKeyWord[] = "TYPE";
            char * l_pszName = p? strstr(p, l_szKeyWord) : NULL;
            if(l_pszName)
            {
                l_pszName += sizeof(l_szKeyWord)-1;
                while(*l_pszName == ' ')
                    l_pszName++;

                char * l_pszEndType = strchr(l_pszName, ' ');
                if(l_pszEndType)
                    *l_pszEndType = 0;

                int l_iInterface = 
                    m_pAllIntr->iGetInterfaceReference(l_pszName);
                if(l_iInterface < 0)
                {
                    MYTRACE(TF_Error, 
                        "ERROR : cannot declare a reference "
                        "to interface %s",
                        l_pszName);
                    return FALSE;
                }
                else if(l_iInterface > 255)
                {
                    MYTRACE(TF_Error, 
                        "ERROR : we cannot handle more than "
                        "255 interfaces");
                    return FALSE;
                }
                else
                {
                    return bAddParam(l_rParam.m_Type, 
                        l_byParamFlags, TRUE, BYTE(l_iInterface));
                }
            }//if(!l_pszName)
            else
            {
                MYTRACE(TF_Error, 
                    "WARNING : LPINTERF does not have an "
                    "TYPE keyword");
            }
        }//if(l_rParam.m_Type == PARAM_LPINTERFACE)
        return bAddParam(l_rParam.m_Type, l_byParamFlags,
            FALSE, 0);
    }//for (int i=0; i < (sizeof(s_Params)/sizeof(s_Params[0])); i
    return FALSE;
}//BOOL Parser::bParseAndAddParam

