/******************************************************************
Module : Return.cpp   - handles interception of functions returns
             Written 1998,1999 by Dmitri Leman
Purpose: NT kernel mode tracer
******************************************************************/
#include "Headers.h"
#include "Driver.h"
#include "RetStub.h"
#include "return.h"
#include "Trace.h"
#include "log.h"
#include "KrnlTrc.h"
#include "KrnlTrcD.h"
#include "Stat.h"

#define _STATISTICS

/*-----------------------------------------------------------------

   FUNCTION: ReturnMgr::Init
             Allocate an array of return stubs
   PURPOSE:  
      
   RETURN VALUE:
      NTSTATUS 0 on success, error code otherwise
-----------------------------------------------------------------*/
NTSTATUS ReturnMgr::Init
(
    int p_iNumberOfStubs
)
{
    //First - allocate the array of return stubs
    int l_iSize = sizeof(ReturnStub) * p_iNumberOfStubs;
    m_pReturnStubArray = (ReturnStub *)ExAllocatePool(
        NonPagedPool, l_iSize);
    if(!m_pReturnStubArray)
    {
        MYTRACE(TF_Error, "ERROR : ExAllocatePool(%d) failed",
            l_iSize);
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    m_iNumStubs = p_iNumberOfStubs;

    //Second - initialize some members of the stubs
    for(int i = 0; i < m_iNumStubs; i++)
    {
        m_pReturnStubArray[i].m_dwReturnAddress = 0;
        m_pReturnStubArray[i].m_byCallInstruction = 0xe8;
        m_pReturnStubArray[i].m_dwCallOffset = 
            (char*)NakedReturnPoint - 
            (char*)(&m_pReturnStubArray[i].m_dwCallOffset)
            -4;
        //The rest of the members will be filled during the call.
    }//for(int i = 0; i < m_iNumStubs; i++)
    return STATUS_SUCCESS;
}//NTSTATUS ReturnMgr::Init

/*-----------------------------------------------------------------

   FUNCTION: ReturnMgr::Shutdown

   PURPOSE:  Free our array
      
   RETURN VALUE:
      NTSTATUS 0 on success, error code otherwise
-----------------------------------------------------------------*/
NTSTATUS ReturnMgr::Shutdown()
{
    if(m_pReturnStubArray)
    {
        ExFreePool(m_pReturnStubArray);
    }
    m_pReturnStubArray = NULL;
    m_iNumStubs = 0;
    return STATUS_SUCCESS;
}//NTSTATUS ReturnMgr::Shutdown()

/*-----------------------------------------------------------------

   FUNCTION: ReturnMgr::pFindStub

   PURPOSE:  Place our interceptor to the return address 
    of the function
    
   PARAMETERS:
    ULONG p_ulReturnAddress - a value to mark the stub
   
   RETURN VALUE:
      ReturnStub   *    Return a return stub
-----------------------------------------------------------------*/
ReturnStub * ReturnMgr::pFindStub
(
    ULONG p_ulReturnAddress
)
{
    int i;
    
    //First we need to find an unused stub in our array of stubs
    for(i = 0; i < m_iNumStubs; i++)
    {
        ReturnStub & l_rStub = m_pReturnStubArray[i];
        ULONG l_ulPreviousValue = (ULONG)
            InterlockedCompareExchange(
            (void**)&l_rStub.m_dwReturnAddress,
            (void*)p_ulReturnAddress, 0);
        //The previous operator stores the l_ulReturnAddress
        //to the l_rStub.m_dwReturnAddress only
        //if it was equal to 0. The original value is returned.
        //Therefore, if l_ulPreviousValue == 0, we have found
        //an empty slot and already reserved it for us.
        if(!l_ulPreviousValue)
        {
            return &l_rStub;

        }
    }//for(i = 0; i < m_iNumStubs; i++)
    return NULL;
}//ReturnStub * ReturnMgr::InterceptFunctionReturn

/*-----------------------------------------------------------------

   FUNCTION: ReturnMgr::CommonReturnPoint

   PURPOSE:  C routine which will be called by NakedReturnPoint - 
    the real interceptor of function returns
      
   PARAMETERS:                   .
                                  
   PULONG p_pReturnAddress          
   PULONG p_pReturnValue            
-----------------------------------------------------------------*/
void ReturnMgr::CommonReturnPoint
(
    PULONG p_pReturnAddress,
    PULONG p_pReturnValue
)
{
#ifdef _STATISTICS
    GetStat l_Stat(g_StatReturn);
#endif

    __try 
    {
        //The value at *p_pReturnAddress points 5 bytes after the
        //m_byCallInstruction member in our stub.
        
        ReturnStub * l_pStub = (ReturnStub*)
            ((*p_pReturnAddress) - 
            offsetof(ReturnStub, m_byCallInstruction) - 5);

        #ifdef _DEBUG
        int l_iStubIndex = l_pStub - m_pReturnStubArray;
        if(ULONG(l_iStubIndex) >= ULONG(m_iNumStubs) ||
            m_pReturnStubArray + l_iStubIndex != l_pStub)
        {
            MYTRACE(TF_Error, "ERROR: wrong stub pointer %x.",
                l_pStub);
            DbgBreakPoint();
            return;
        }
        #endif

        // Restore the return address to it's original value.
        *p_pReturnAddress = l_pStub->m_dwReturnAddress;
        
        // Emit the information about the function return value 
        // to the logging mechanism.
        Logger::ProcessReturn(l_pStub->m_pStub, 
            *p_pReturnValue, l_pStub);

        //Release the stub by erasing the return address member
        InterlockedExchange((long*)&l_pStub->m_dwReturnAddress, 0);
    }
    __except(1)
    {
        MYTRACE(TF_Error, 
            "ERROR : exception inside CommonReturnPoint");
    }
#ifdef _STATISTICS
    l_Stat.vRecord();
#endif
}//void ReturnMgr::CommonReturnPoint

/*----------------------------------------------------------------

   FUNCTION: ReturnMgr::NakedReturnPoint

   PURPOSE:  Interceptor of function returns. It will be called
             by a call instruction embedded in our stub.
----------------------------------------------------------------*/
__declspec(naked) void ReturnMgr::NakedReturnPoint(void)
{
    PULONG l_pReturnAddress;
    PULONG l_pReturnValue;

    //It would be much easier to write this function entirely on
    //assembly. The code would be shorter and faster but more
    //difficult to read and compile.
    //Therefore, I limited myself to only few assembly operation,
    //writing the rest on C.
    __asm 
    {
        PUSHAD      ; Push EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI
        PUSHFD      ; Store flags
        MOV    EBP, ESP   ; allow access to local variables
        SUB    ESP, __LOCAL_SIZE ;reserve space for local variables
    
        mov l_pReturnAddress, EBP
    }
    l_pReturnAddress += 8+1;
    //adjust for 8 registers pushed by PUSHAD and flags

    l_pReturnValue =l_pReturnAddress-1;//EAX was the first register
                                        //pushed by PUSHAD.
    
    //This function was called by our stub. Therefore, the return
    //address points to our stub.
    //CommonReturnPoint will replace this return address by the
    //original return address, stored in the stub
    if(g_Data.m_pRetMgr)
       g_Data.m_pRetMgr->CommonReturnPoint
        (l_pReturnAddress, l_pReturnValue);

    __asm
    {
        ADD    ESP, __LOCAL_SIZE ;remove local variables from stack
        POPFD                          ;restore flags
        POPAD                          ;restore registers
        RET
    }
}//__declspec(naked) ReturnMgr::NakedReturnPoint(void)
