/*****************************************************************
Module : StubMgr.cpp -manage configuration data in the kernel mode
             Written 1998,1999 by Dmitri Leman
Purpose: NT kernel mode tracer
******************************************************************/
#include "Headers.h"
#include "Driver.h"
#include "Trace.h"
#include "StubMgr.h"
#include "KrnlTrcD.h"
#include "PEHelper.h"
#include "Intrcpt.h"
#include "EnumMods.h"

/*----------------------------------------------------------------

   FUNCTION: StubManager::vDriverInit

   PURPOSE:  After the DLL(driver) is loaded, replace offsets by
    real pointers
      
   PARAMETERS:                   .
                                  
   UCHAR *  p_pMem                 Address where the config data
                                  is mapped
----------------------------------------------------------------*/
void StubManager::vDriverInit
(
    UCHAR *  p_pMem,
    void *  p_pLogFunction
)
{
    //Convert all offsets to pointers
    int i;
    if(m_dwOffsetStubs)
    {
        m_dwOffsetStubs += ULONG(p_pMem);
        ASM_APIFunctionStub * l_pStubs = 
            (ASM_APIFunctionStub * )m_dwOffsetStubs;
        for(i = 0; i < m_iNumStubs; i++)
            l_pStubs[i].vAdjustOnDllLoad(p_pMem, p_pLogFunction);
    }//if(m_dwOffsetStubs)

    for(i = 0; i < m_iTotalNumInts; i++)
        m_ArrayInterfaces[i].vDllInit(p_pMem);
}//void StubManager::vDriverInit

/*----------------------------------------------------------------

   FUNCTION: StubManager::vShortcutAllStubs

   PURPOSE:  Modify all stub to transfer the control directly to
    the original function. In some cases we may decide to keep
    the array of stubs in memory even after the termination
    of our driver.
      
----------------------------------------------------------------*/
void StubManager::vShortcutAllStubs()
{
    //Convert all offsets to pointers
    int i;
    if(m_dwOffsetStubs)
    {
        ASM_APIFunctionStub * l_pStubs = 
            (ASM_APIFunctionStub * )m_dwOffsetStubs;
        for(i = 0; i < m_iNumStubs; i++)
            l_pStubs[i].vShortcut();
    }//if(m_dwOffsetStubs)
}//void StubManager::vShortcutAllStubs

/*----------------------------------------------------------------

   FUNCTION: MyInterface::vDllInit

   PURPOSE:  After the DLL(driver) is loaded, replace offsets by
    real pointers
      
   PARAMETERS:                   .
                                  
   UCHAR *  p_pMem                 Address where the config data
                                  is mapped
-----------------------------------------------------------------*/
void MyInterface::vDllInit(UCHAR *  p_pMem)
{
    if(m_dwOffsetName)
       m_dwOffsetName += ULONG(p_pMem);
    if(m_dwOffsetFunctions)
       m_dwOffsetFunctions += ULONG(p_pMem);
}//void MyInterface::vDllInit()

/*-----------------------------------------------------------------

   FUNCTION: ASM_APIFunctionStub::vAdjustOnDllLoad

   PURPOSE:  After the DLL(driver) is loaded, replace offsets by
    real pointers
      
   PARAMETERS:                   .
                                  
   UCHAR *  p_pMem                 Address where the config data
                                  is mapped
   void *  p_pLogFunction         Common interceptor function
-----------------------------------------------------------------*/
void ASM_APIFunctionStub::vAdjustOnDllLoad
(
    UCHAR *  p_pMem,
    void *  p_pLogFunction
)
{
    m_dwModuleNameOffset = m_dwModuleNameOffset + ULONG(p_pMem);
    m_dwParamsOffset = m_dwParamsOffset + ULONG(p_pMem);
    m_dwFunctionNameOffset = m_dwFunctionNameOffset +ULONG(p_pMem);

    m_Code.offset_LogFunction =
        PUCHAR(p_pLogFunction) - PUCHAR(&m_Code.instr_CALL)-5;
}//void ASM_APIFunctionStub::vAdjustOnDllLoad

/*-----------------------------------------------------------------

   FUNCTION: ASM_APIFunctionStub::vShortcut

   PURPOSE:  Modify the stub to transfer the control directly to
    the original function. In some cases we may decide to keep
    the array of stubs in memory even after the termination
    of our driver.
      
-----------------------------------------------------------------*/
void ASM_APIFunctionStub::vShortcut()
{
    m_Code.instr_CALL = RET_INSTR;//it is "RET" instruction
}//ASM_APIFunctionStub::vShortcut

/*-----------------------------------------------------------------

   FUNCTION: StubManager::pFindStubByOrigAddress

   PURPOSE:  Find a stub correspondent to the specified function.
    It works only for exported functions (not for interfaces).
      
   PARAMETERS:                   .
                                  
    PVOID p_pOrigFunctionAddr      Function address

   RETURN VALUE:
      ASM_APIFunctionStub * found stub or NULL
_________________________________________________________________*/
ASM_APIFunctionStub * StubManager::pFindStubByOrigAddress
(
    PVOID p_pOrigFunctionAddr
)
{
    if(!p_pOrigFunctionAddr)
        return NULL;
    ASM_APIFunctionStub * l_pStubs = 
        (ASM_APIFunctionStub *)m_dwOffsetStubs;
    //Interface stubs are stored first, exported - second
    for(int i = m_iNumInterfStubs; i < m_iNumStubs; i++)
    {
        if(l_pStubs[i].dwGetOriginalFunctionAddress() == 
            ULONG(p_pOrigFunctionAddr))
            return l_pStubs + i;
    }
    return NULL;
}//ASM_APIFunctionStub * StubManager::pFindStubByOrigAddress

MyInterface * StubManager::pFindModuleByName
(
    const char * p_pszName
)
{
    for(int i = m_iNumInterfaces; i < m_iTotalNumInts; i++)
    {
        MyInterface * l_pInt = m_ArrayInterfaces + i;

        if(!_stricmp(p_pszName, (LPCSTR)l_pInt->m_dwOffsetName))
        {
            return l_pInt;
        }
    }//for(i = m_iNumInterfaces; i < m_iTotalNumInts; i++)
    return NULL;
}//MyInterface * StubManager::pFindModuleByName

/*-----------------------------------------------------------------

   FUNCTION: MyInterface::pGetStub

   PURPOSE:  Get a pointer to a stub by index
      
   PARAMETERS:                   .
                                  
   int p_iIndex                  Index of function in the interface 

   RETURN VALUE:
      ASM_APIFunctionStub *      return pointer to the stub or NULL
-----------------------------------------------------------------*/
ASM_APIFunctionStub * MyInterface::pGetStub(int p_iIndex)
{
    return (ULONG(p_iIndex) >= ULONG(m_iNumFunctions))? NULL:
        ((ASM_APIFunctionStub * )m_dwOffsetFunctions) + p_iIndex;
}//ASM_APIFunctionStub * MyInterface::pGetStub

/*-----------------------------------------------------------------

   FUNCTION: StubManager::bProcessAllModulesExports

  PURPOSE: Look for modules whose base addresses are different from
    the previously seen. Update addresses of exported functions
    for all stubs correspondent to this module. Also intercept
    these exports by replacing them by addresses of our stubs.
    This function does not intercept module imports - only exports.

   RETURN VALUE: 
     false in case of a serious error
_________________________________________________________________*/
bool StubManager::bProcessAllModulesExports
(
    bool p_bUnhook
)
{
    //Interfaces are stored first, Modules - second
    EnumModules l_Enum;
    int i;
    //First erase the base address for all modules 
    //(keeping the previous state).
    for(i = m_iNumInterfaces; i < m_iTotalNumInts; i++)
    {
        MyInterface & l_rInt = m_ArrayInterfaces[i];
        l_rInt.m_dwModuleBasePrevious = l_rInt.m_dwModuleBase;
        l_rInt.m_dwModuleBase = 0;
    }

    //Second - enumerate all loaded modules and update our 
    //module information structure by the actual module base 
    //address. 
    //It may require update of all function stubs, belonging 
    //to this module.

    const EnumModules::DriverInfo * l_pInfo = l_Enum.pGetFirst();
    while(l_pInfo)
    {
        //l_pInfo->m_szName contains a full path to the module.
        //We need to compare only the file name.
        const char * l_pFileName = strrchr(l_pInfo->m_szName,'\\');
        if(l_pFileName)
        {
            l_pFileName++;
        }
        else
        {
            l_pFileName = l_pInfo->m_szName;
        }

        //Now search the table of our module information structures
        //for this module:
        MyInterface * l_pModule = pFindModuleByName(l_pFileName);
        if(l_pModule)
        {
            //Module found. Update the address
            l_pModule->m_dwModuleBase = l_pInfo->m_ulAddress;
        }
        l_pInfo = l_Enum.pGetNext();
    }//while(l_pInfo)

    //Third, for all modules, whose addresses have changed, update
    //function stubs:
    bool l_bResult = true;
    for(i = m_iNumInterfaces; i < m_iTotalNumInts; i++)
    {
        MyInterface & l_rInt = m_ArrayInterfaces[i];
        if(!p_bUnhook &&
            l_rInt.m_dwModuleBase == l_rInt.m_dwModuleBasePrevious)
        {
           //The address havn't changed. Our stubs are still valid.
            continue;
        }

        if(!Interceptor::bProcessExportsForOneModule
            (l_rInt, p_bUnhook))
            l_bResult = false;
    }//for(i = m_iNumInterfaces; i < m_iTotalNumInts; i++)
    return l_bResult;
}//bool StubManager::bProcessAllModulesExports()
