/******************************************************************
Module : Trace.cpp Support for debug trace and output file logging
             Written 1998,1999 by Dmitri Leman
Purpose: COM/OLE tracer and NT kernel mode tracer.
        This file is included in the driver and the application projects
******************************************************************/
#include "Headers.h"
#include "tchar.h"
#include "Trace.h"
#include "FileSupp.h"

#ifdef _TRACE //{

#pragma intrinsic(memset,memcpy,strlen,_tcscpy)

#define TRACE_BUFFER_SIGNATURE 0x46425254 //"TRBF"
struct TraceBufferHeader
{
    ULONG m_dwSignature;//TRACE_BUFFER_SIGNATURE
    long  m_lSizeAllocated;
    long  m_lNumBytesWritten;
};//struct TraceBufferHeader

struct TraceImpl
{
    void vClean()
    {
        m_pTraceBuffer = NULL;
        m_dwOutputDebugStringMask = TF_Error;
        m_dwTraceBufferMask = 0;
        #ifdef _DRIVER
        m_dwOutputLogMask = 0;
        m_pDriverObject = NULL;
        #else
        m_dwMessageBoxMask        = TF_Error | TF_MessageBox;
        #endif
        m_dwPrintFileMask         = 0;
        m_dwFlushFileMask         = 0;
        m_dwTotalMask = TF_Error;

        m_File.Cleanup();
        vSetTotalTraceMask();
    }
    void vFree()
    {
        m_File.CloseFile();
        bFreeTraceBuffer();
        vClean();
    }

    bool bAllocateTraceBuffer(int p_iSize);
    bool bFreeTraceBuffer();
    bool bDumpTraceBufferToFile(const char * p_pszFileName);
    bool bAddToTraceBuffer(const char * p_pszBuff, int p_iLen);
    bool bAddToOutputLog(const char * p_pszString, int p_iLen);

    TraceBufferHeader * m_pTraceBuffer;

    ULONG m_dwOutputDebugStringMask;
    ULONG m_dwTraceBufferMask;
    #ifdef _DRIVER
    ULONG m_dwOutputLogMask;
    PDRIVER_OBJECT m_pDriverObject;
    #else
    ULONG m_dwMessageBoxMask       ;
    #endif
    ULONG m_dwFlushFileMask        ;
    ULONG m_dwPrintFileMask   ;

    ULONG m_dwTotalMask;

    MyFileImpl m_File;

    void vSetTotalTraceMask();
};//struct TraceImpl

static TraceImpl s_Impl;

void vTraceInitialize()
{
    s_Impl.vClean();
}

void vTraceUnInitialize()
{
    s_Impl.vFree();
}

void TraceImpl::vSetTotalTraceMask()
{
    m_dwTotalMask = 
        m_dwOutputDebugStringMask 
        | m_dwTraceBufferMask
        #ifdef _DRIVER
        | m_dwOutputLogMask
        #else
        | m_dwMessageBoxMask
        #endif
        | m_dwPrintFileMask         
        ;
    //m_dwPrintFlushFileMask is not included because it affects
    //only flushing of the file - not printing itself
}//void TraceImpl::vSetTotalTraceMask

void vSetOneTraceMask
(
    PULONG   p_pdwMask,
    ULONG    p_dwNewBits,
    ULONG    p_dwBitsToModify
)
{
    *p_pdwMask =  (*p_pdwMask  & ~p_dwBitsToModify) |
                  (p_dwNewBits &  p_dwBitsToModify);
}

/*-----------------------------------------------------------------

   FUNCTION: vSetTraceMasks

   PURPOSE:  This function allows to enable/disable output
    of the specified groups of MYTRACE operators to
    the specified output device.
    Each MYTRACE operator has a bit mask as a first parameter.
    This mask will be matched against a mask for a particular
    device to deside whether to allow output or not.
      
   PARAMETERS:                   .
    
   ULONG    p_dwTraceTypes         flags from TraceTypes.
   ULONG    p_dwNewBits            new bits to set or reset
   ULONG    p_dwBitsToModify       which bits can be modified
-----------------------------------------------------------------*/
void vSetCurTraceMasks
(
    ULONG   p_dwTraceTypes,
    ULONG   p_dwNewBits   ,
    ULONG   p_dwBitsToModify

)
{
    if(p_dwTraceTypes  & TT_DebugMonitor)
        vSetOneTraceMask(&s_Impl.m_dwOutputDebugStringMask,
            p_dwNewBits, p_dwBitsToModify);
    if(p_dwTraceTypes  & TT_MemoryBuffer)
        vSetOneTraceMask(&s_Impl.m_dwTraceBufferMask,
            p_dwNewBits, p_dwBitsToModify);
#ifdef _DRIVER
    if(p_dwTraceTypes  & TT_OutputLog)
        vSetOneTraceMask(&s_Impl.m_dwOutputLogMask,
            p_dwNewBits, p_dwBitsToModify);
#else
    if(p_dwTraceTypes  & TT_MessageBox)
        vSetOneTraceMask(&s_Impl.m_dwMessageBoxMask,
            p_dwNewBits, p_dwBitsToModify);
#endif
    if(p_dwTraceTypes  & TT_File)
        vSetOneTraceMask(&s_Impl.m_dwPrintFileMask,
            p_dwNewBits, p_dwBitsToModify);
    if(p_dwTraceTypes  & TT_FlushFile)
        vSetOneTraceMask(&s_Impl.m_dwFlushFileMask,
            p_dwNewBits, p_dwBitsToModify);

    s_Impl.vSetTotalTraceMask();
}//void vSetCurDebugMasks

#ifndef _DRIVER
void vSetOutputToFile
(
    LPCTSTR p_pszFileName,
    ULONG   p_dwNewBits   ,
    ULONG   p_dwBitsToModify,
    ULONG   p_dwNewFlushBits   ,
    ULONG   p_dwFlushBitsToModify
)
{
    s_Impl.m_File.CloseFile();
    if(s_Impl.m_File.OpenFile(p_pszFileName, true, true, NULL))
    {
        p_dwNewBits = 0;
        p_dwBitsToModify = (ULONG)-1;
    }
    vSetCurTraceMasks(TT_File, p_dwNewBits, p_dwBitsToModify);
    vSetCurTraceMasks(TT_FlushFile, 
        p_dwNewFlushBits, p_dwFlushBitsToModify);
}//void vSetOutputToFile
#endif

bool TraceImpl::bAllocateTraceBuffer
(
    int p_iSize
)
{
    if(!bFreeTraceBuffer())
        return false;

    int l_iTotalSize = sizeof(TraceBufferHeader) + p_iSize;

    TraceBufferHeader * l_pBuffer;
    #ifdef _DRIVER 
        l_pBuffer = (TraceBufferHeader*)
            ExAllocatePool(NonPagedPool, l_iTotalSize);
    #else
        l_pBuffer = (TraceBufferHeader*)
            malloc(l_iTotalSize);
    #endif
    if(!l_pBuffer)
    {
        MYTRACE(TF_Error, 
            "ERROR: cannot allocate trace buffer of size %d",
            l_iTotalSize);
        return false;
    }
    l_pBuffer->m_dwSignature = TRACE_BUFFER_SIGNATURE;
    l_pBuffer->m_lSizeAllocated = p_iSize;
    l_pBuffer->m_lNumBytesWritten = 0;
    m_pTraceBuffer = l_pBuffer;
    return true;
}//bool bAllocateTraceBuffer

bool TraceImpl::bFreeTraceBuffer()
{
    if(m_pTraceBuffer)
    {
        TraceBufferHeader * l_pBuffer = m_pTraceBuffer;
        m_pTraceBuffer = NULL;
        #ifdef _DRIVER 
            ExFreePool(l_pBuffer);
        #else
            free(l_pBuffer);
        #endif
    }//if(m_pTraceBuffer)
    return true;
}//bool bFreeTraceBuffer()

/*-----------------------------------------------------------------

   FUNCTION: bWriteCircularBufferToFile

   PURPOSE:  Used to dump trace memory buffer to a file.
    The memory buffer has a circular structure. The oldest data
    starts right after the most recent data. This position can
    be calculated by dividing the total number of bytes written
    by the size of the buffer. This function starts dumping the
    oldest data up to the end of the buffer and then continues
    from the begining of the buffer.
    This function is available for applications and drivers.
   PARAMETERS:
   const char * p_pszFileName     name of the file
   void       * p_pBuffer         memory buffer to write
   long         p_lNumBytesWritten total number of bytes written
             to the buffer (may be greater than the buffer size)
   long         p_lSizeAllocated   Size of the buffer
   ULONG * p_pdwReturnStatus      in case of error return Win32 
            error (result of GetLastError) or NT kernel status.

   RETURN VALUE:
      bool      true on success, false on failure
-----------------------------------------------------------------*/
bool bWriteCircularBufferToFile
(
    const char * p_pszFileName,
    void       * p_pBuffer,
    long         p_lNumBytesWritten,
    long         p_lSizeAllocated,
    ULONG      * p_pdwReturnStatus
)
{
    MyFileLocal l_File;
    if(l_File.OpenFile(p_pszFileName, true, true, NULL))
        return false;

    ULONG l_dwWritten = 0;
    
    LARGE_INTEGER l_FileOffset;
    l_FileOffset.QuadPart = 0;

    if(!p_pBuffer)
    {
        char * l_pszString = "Debug memory is NULL";
        l_File.ReadWriteFile(true, 
            l_pszString, strlen(l_pszString)+1,
            &l_FileOffset, NULL);
        l_File.CloseFile();
        #if defined(_DRIVER)
        if(p_pdwReturnStatus)
            *p_pdwReturnStatus = STATUS_UNSUCCESSFUL;
        #endif
        return false;
    }
    //Our buffer is circular. 
    //We need to write the tail first, then the begining
    int l_iLastAddedByte = p_lNumBytesWritten % 
                           p_lSizeAllocated;

    char* l_pbyMem = (char*)(p_pBuffer);

    if(p_lNumBytesWritten > p_lSizeAllocated)
        l_File.ReadWriteFile(true, 
            l_pbyMem + l_iLastAddedByte, 
            p_lSizeAllocated - l_iLastAddedByte,
            &l_FileOffset, NULL);

    l_File.ReadWriteFile(true, l_pbyMem, l_iLastAddedByte, 
        &l_FileOffset, NULL);
    
    l_File.CloseFile();
    return true;
}//bool bWriteCircularBufferToFile

/*-----------------------------------------------------------------

   FUNCTION: TraceImpl::bDumpTraceBufferToFile

   PURPOSE:  Print the accumulated messages from the trace memory
             buffer to a file.
           This function is available for applications and drivers.
      
   PARAMETERS:                   .
    const char * p_pszFileName   File name.
-----------------------------------------------------------------*/
bool TraceImpl::bDumpTraceBufferToFile
(
    const char * p_pszFileName
)
{
    if(!bWriteCircularBufferToFile(p_pszFileName,
        m_pTraceBuffer? (char*)(m_pTraceBuffer+1) : NULL,
        m_pTraceBuffer->m_lNumBytesWritten, 
        m_pTraceBuffer->m_lSizeAllocated, NULL))
        return false;
    return true;
}//bool bDumpTraceBufferToFile

/*-----------------------------------------------------------------

   FUNCTION: TraceImpl::bAddToTraceBuffer

   PURPOSE:  Add the given message to the trace memory buffer.
           This function is available for applications and drivers.
      
   PARAMETERS:                   .

    char *  p_pszMessage,        Message - 0-terminated ANSI string
    int     p_iLen               Number of characters in the string
-----------------------------------------------------------------*/
bool TraceImpl::bAddToTraceBuffer
(
    const char * p_pszString, 
    int    p_iLen
)
{
    if(!m_pTraceBuffer)
        return false;

    char * l_pszText = (char * )(m_pTraceBuffer+1);

    if(p_iLen > m_pTraceBuffer->m_lSizeAllocated)
       p_iLen = m_pTraceBuffer->m_lSizeAllocated;

    //To ensure thread safety, we access the m_iWriteTo variable 
    //only from the single atomic operation
    int l_iWriteAt = InterlockedExchangeAdd
       ((long*)&m_pTraceBuffer->m_lNumBytesWritten, p_iLen);

    //Now we reserved a space in the buffer, which other threads will not
    //touch, unless the buffer is filled too fast and wrapped around.
    //In the latter case we will have a meaningless garbadge in the buffer
    //in any case.

    //Wrap around (circular buffer):
    l_iWriteAt %= m_pTraceBuffer->m_lSizeAllocated;

    if(l_iWriteAt + p_iLen <= m_pTraceBuffer->m_lSizeAllocated)
        memcpy(l_pszText + l_iWriteAt, p_pszString, p_iLen);
    else
    {
        int l_iWriteAtTheTail = 
            m_pTraceBuffer->m_lSizeAllocated - l_iWriteAt;

        memcpy(l_pszText + l_iWriteAt, p_pszString, 
            l_iWriteAtTheTail);

        int l_iWriteAtTheBegining = p_iLen - l_iWriteAtTheTail;

        memcpy(l_pszText, p_pszString + l_iWriteAtTheTail,
            l_iWriteAtTheBegining);
    }
    return true;
}//bool bAddToTraceBuffer

bool bAllocateTraceBuffer(int p_iSize,
    ULONG   p_dwNewBits, ULONG   p_dwBitsToModify)
{
    if(!s_Impl.bAllocateTraceBuffer(p_iSize))
    {
        vSetCurTraceMasks(TT_MemoryBuffer, 0, (ULONG)-1);
        return false;
    }
    vSetCurTraceMasks
        (TT_MemoryBuffer, p_dwNewBits, p_dwBitsToModify);
    return true;
}

bool bFreeTraceBuffer()
{
    return s_Impl.bFreeTraceBuffer();
}

bool bDumpTraceBufferToFile(const char * p_pszFileName)
{
    return s_Impl.bDumpTraceBufferToFile(p_pszFileName);
}

#ifdef _DRIVER //{
/*-----------------------------------------------------------------

   FUNCTION: bEnableTraceToOutputLog

   PURPOSE:  To enable printing some groups of trace messages to 
             the system error log.
             This function available for NT drivers only.

   PARAMETERS:                   .

    PDRIVER_OBJECT p_pDriverObject - driver object to be associated
                                     with the all log entries
    ULONG   p_dwNewBits            - which groups of traces should
                                     be added to the log (from 
                                     enum TraceFlags).
    ULONG   p_dwBitsToModify    - which bits to modify. If a bit
                                  set to 1 in p_dwBitsToModify,
                                  the correspondent bit will be
                                  taken from p_dwNewBits. 
                                  Otherwise, it will be preserved.
-----------------------------------------------------------------*/
bool bEnableTraceToOutputLog
(
    PDRIVER_OBJECT p_pDriverObject,
    ULONG   p_dwNewBits, 
    ULONG   p_dwBitsToModify
)
{
    vSetCurTraceMasks
        (TT_OutputLog, p_dwNewBits, p_dwBitsToModify);
    s_Impl.m_pDriverObject = p_pDriverObject;
    return true;
}//bool bEnableTraceToOutputLog

/*-----------------------------------------------------------------

   FUNCTION: TraceImpl::bAddToOutputLog

   PURPOSE:  Add the given message to the system error log.
             This function available for NT drivers only.
      
   PARAMETERS:                   .

    char *  p_pszMessage,        Message - 0-terminated ANSI string
    int     p_iLen               Number of characters in the string
  
-----------------------------------------------------------------*/
bool TraceImpl::bAddToOutputLog
(
    const char * p_pszString,
    int    p_iLen
)
{
    if(!m_pDriverObject ||
        KeGetCurrentIrql() > DISPATCH_LEVEL)
        return false;

    int l_iStringSize = (p_iLen+1)*sizeof(WCHAR);
    UCHAR l_cPacketSize = 
        sizeof( IO_ERROR_LOG_PACKET ) + l_iStringSize;
    if(l_cPacketSize > ERROR_LOG_MAXIMUM_SIZE)
    {
        l_cPacketSize = ERROR_LOG_MAXIMUM_SIZE;
        l_iStringSize = ERROR_LOG_MAXIMUM_SIZE - 
            sizeof( IO_ERROR_LOG_PACKET );
        p_iLen = (l_iStringSize/sizeof(WCHAR))-1;
    }

    PIO_ERROR_LOG_PACKET l_pPacket = (PIO_ERROR_LOG_PACKET)
        IoAllocateErrorLogEntry( 
        m_pDriverObject, l_cPacketSize);

    if(!l_pPacket)
        return false;

    int l_iDumpDataCount = 0;//no data to dump

    //we have only string to report - so set 0 code and value
    l_pPacket->ErrorCode = 0;
    l_pPacket->UniqueErrorValue = 0;
    l_pPacket->MajorFunctionCode = 0;
    l_pPacket->RetryCount = 0;
    l_pPacket->FinalStatus = 0;
    l_pPacket->SequenceNumber = 0;
    l_pPacket->IoControlCode = 0;
    l_pPacket->DumpDataSize = 0;

    l_pPacket->NumberOfStrings = 1;

    l_pPacket->StringOffset = sizeof( IO_ERROR_LOG_PACKET ) +
            ( l_iDumpDataCount - 1 ) * sizeof( ULONG );

    UNICODE_STRING l_ResultString;
    l_ResultString.Length = 0;
    l_ResultString.MaximumLength = l_iStringSize;
    l_ResultString.Buffer = (PWSTR)
        ((PUCHAR)l_pPacket + l_pPacket->StringOffset);

    ANSI_STRING l_AnsiString;
    l_AnsiString.Length = p_iLen;
    l_AnsiString.MaximumLength = p_iLen;
    l_AnsiString.Buffer = (char*)p_pszString;

    RtlAnsiStringToUnicodeString
        (&l_ResultString, &l_AnsiString, FALSE);
    l_ResultString.Buffer[p_iLen] = 0;
    
    IoWriteErrorLogEntry(l_pPacket);// Log the message
    return true;
}//bool TraceImpl::bAddToOutputLog
#endif //#ifdef _DRIVER //}

#ifndef _DRIVER //{
/*-----------------------------------------------------------------

   FUNCTION: bTraceMessageBox

   PURPOSE:  Display the given message in a message box.
             Allow the User to terminate the application, triiger
             a breakpoint or continue the execution.
             This function available for applications only.
      
   PARAMETERS:                   .

    LPCTSTR p_pszMessage,        Message - 0-terminated string
    int     p_iLen               Number of characters in the string
  
-----------------------------------------------------------------*/
bool bTraceMessageBox
(
    LPCTSTR p_pszMessage,
    int     p_iLen
)
{
    static long s_lAssertRecursion = -1;
    if (InterlockedIncrement(&s_lAssertRecursion) > 0)
    {
        InterlockedDecrement(&s_lAssertRecursion);
        __asm int 3
        return false;
    }//if (InterlockedIncrement(&s_lAssertRecursion) > 0)

    TCHAR l_szText[1024];
    TCHAR l_szAddition[] = _T("Press Abort to terminate application,")
            _T("Retry to trigger breakpoint, Ignore to continue");
    int l_iLenBytes = p_iLen*sizeof(TCHAR);
    if(l_iLenBytes + sizeof(l_szAddition) < sizeof(l_szText))
    {
        memcpy(l_szText, p_pszMessage, l_iLenBytes);
        memcpy(l_szText + p_iLen, l_szAddition, sizeof(l_szAddition));
        p_pszMessage = l_szText;
    }

    HWND l_hWndParent = GetActiveWindow();
    if (l_hWndParent != NULL)
        l_hWndParent = GetLastActivePopup(l_hWndParent);

    int l_iRet = MessageBox(l_hWndParent, p_pszMessage,
        _T("Trace Message"),
        MB_TASKMODAL|MB_ABORTRETRYIGNORE|MB_SETFOREGROUND);
    InterlockedDecrement(&s_lAssertRecursion);

    switch(l_iRet) {
    case IDABORT:
        TerminateProcess(GetCurrentProcess(),0);
        return true;
    case IDRETRY:
        __asm int 3
        return false;
    default:
    case IDIGNORE:
        return false;
    }
}//bool bTraceMessageBox
#endif //}

/*-----------------------------------------------------------------

   FUNCTION: OutputTraceString

   PURPOSE:  Print the given 0-terminated ANSI string to the one 
    or several of the trace outputs (such as memory buffer, 
    debug monitor,etc.) 
    The destination is controlled by bit masks for the
    particular output and the given mask for this message.
    Do nothing if the specified mask does not match the
    enabled trace masks.
      
   PARAMETERS:                   .
                        
    ULONG p_dwMask               flags from TraceFlags
    LPCTSTR p_pszMessage,        Message - 0-terminated string
    int     p_iLen               Number of characters in the string
-----------------------------------------------------------------*/
void OutputTraceString
(
    ULONG   p_dwMask,
    LPCTSTR p_pszMessage,
    int     p_iLen
)
{
    if(s_Impl.m_dwOutputDebugStringMask & p_dwMask)
    {
        #ifdef _DRIVER
            DbgPrint((char*)p_pszMessage);
        #else
            OutputDebugString(p_pszMessage);
        #endif
    }
    if(s_Impl.m_dwTraceBufferMask & p_dwMask)
    {
        s_Impl.bAddToTraceBuffer(p_pszMessage, p_iLen);
    }
    #ifdef _DRIVER
    if(s_Impl.m_dwOutputLogMask & p_dwMask)
    {
        s_Impl.bAddToOutputLog(p_pszMessage, p_iLen);
    }
    #else
    if(s_Impl.m_dwMessageBoxMask & p_dwMask)
        bTraceMessageBox(p_pszMessage, p_iLen);
    #endif

    #ifndef _DRIVER
    if(s_Impl.m_dwPrintFileMask & p_dwMask)
    {
        s_Impl.m_File.ReadWriteFile(true, 
            (void*)p_pszMessage, p_iLen, NULL, NULL);

        if(s_Impl.m_dwFlushFileMask & p_dwMask)
            s_Impl.m_File.FlushBuffers();
    }
    #endif
}//void OutputTraceString

/*-----------------------------------------------------------------

   FUNCTION: MyTraceImpl

   PURPOSE:  Perform printf style formatting and tracing.
    Do nothing if the specified mask does not match the
    enabled trace masks.
      
   PARAMETERS:                   .
                                  
   ULONG p_dwMask                 flags from TraceFlags
   LPCTSTR p_pszFormat            printf style format
   ...                            
-----------------------------------------------------------------*/
void MyTraceImpl(ULONG p_dwMask, LPCTSTR p_pszFormat, ...)
{
    //Check whether this group of traces is enabled.
    if(!(s_Impl.m_dwTotalMask & p_dwMask))
        return;//Skip this trace

    va_list l_List;
    va_start(l_List, p_pszFormat);
    TCHAR l_szBuff[512];
    
    *l_szBuff = 0;
    _vsntprintf(l_szBuff, sizeof(l_szBuff)/sizeof(l_szBuff[0]), p_pszFormat, l_List);

    va_end(l_List);

    l_szBuff[sizeof(l_szBuff)/sizeof(l_szBuff[0])-2] = 0;

    int l_iLen = _tcslen(l_szBuff);
    if(l_iLen > 0 && l_szBuff[l_iLen-1] != '\n')
    {
        //Add the new line character if it is not already there
        l_szBuff[l_iLen++] = '\r';
        l_szBuff[l_iLen++] = '\n';
        l_szBuff[l_iLen] = 0;
    }
    OutputTraceString(p_dwMask, l_szBuff, l_iLen);
    
}//void MyTraceImpl

#endif //_TRACE }

#ifndef _DRIVER //{
/*-----------------------------------------------------------------

   FUNCTION: LocalStringHolder strGetErrorExplanation.

   PURPOSE:  Obtain an explanation for the error
      
   PARAMETERS:                   
                                  
   ULONG p_dwErrCode  Win32 error code returned from GetLastError
-----------------------------------------------------------------*/
LocalStringHolder strGetErrorExplanation(ULONG p_dwErrCode)
{
    if(!p_dwErrCode)
        return LocalStringHolder();

    TCHAR * l_pszBuffer = NULL;
    
    FormatMessage(
       FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        0, p_dwErrCode, 0, (TCHAR*)&l_pszBuffer, 0, NULL);
  
    LocalStringHolder l_strReturn;

    if(l_pszBuffer) 
    {
        TCHAR * l_pNewBuffer = (TCHAR * )
            LocalAlloc(LPTR, (_tcslen(l_pszBuffer)+40)*sizeof(TCHAR));
        if(l_pNewBuffer)
        {
            wsprintf(l_pNewBuffer, _T("Error %d(0x%x) - %s"), 
                p_dwErrCode, p_dwErrCode, l_pszBuffer);
            LocalFree(l_pszBuffer);
        }
        else
            l_pNewBuffer = l_pszBuffer;
        l_strReturn.vAdoptStr(l_pNewBuffer);
    }//if(l_pszBuffer) 
  
    return l_strReturn;
}//LocalStringHolder strGetErrorExplanation
#endif //_DRIVER //}


